#include <stdio.h>
#include <apollo/base.h>
#include <apollo/gpr.h>
#include <tml/tml_c_types.h>
#include <tml/tml_dp_graph.hxx>

extern long create_bitmap();
extern void enable_all_keys();

main()
{
    tml_c_model      display_model;
    tml_c_view_port  display_viewport;
    tml_c_format     display_format;
    C_MODEL_POS      start_display;
    long             bitmap_desc;
    char             unob;
    gpr_$event_t     ev_type;
    char             ev_value;
    gpr_$position_t  ret_coord;
    status_$t        status;

    /* initialize TMLib */
    tml_init();

    /* associate all viewport objects with the GPR graphic system */
    tml_c_tml_dp_set_dflt_vpftbl(tml_c_tml_dp_get_gpr_vpftbl());

    /* create a model from the ASCII file test_file */
    display_model = tml_c_read_uasc_path("test_file", 9);
    if (!display_model)
        exit (1);

    /* allocate a viewport */
    display_viewport = view_port_class_alloc();
    /* create a graphics display area, using a routine which depends */
    /*   on the graphics system being used */
    bitmap_desc = create_bitmap();

    /* assign the display area to the viewport */
    view_port_set_handle(display_viewport, &bitmap_desc);

    /* create a format and attach it to the viewport */
    view_port_attach_format(display_viewport, 0, 0, 0);
    /* get a handle to the format created by the last call */
    display_format = view_port_get_format(display_viewport);
    /* attach the model to the format associated with the viewport */
    format_attach(display_format, display_model);

    /* fill viewport with text; first character placed at pixel (0,0) */
    tml_dp_damage_repair(view_port_get_handle(display_viewport));

    /* enable keyboard */
    enable_all_keys();

    /* wait for input, then exit */
    unob = gpr_$event_wait(&ev_type, (unsigned char*) &ev_value,
                           &ret_coord, &status);

    view_port_free( display_viewport);
    format_free( display_format);
    model_free( display_model);
}

