#include <apollo/base.h>
#include <apollo/gpr.h>
#include <apollo/kbd.h>

/***************************************************************************
  In C, enabling the keyset is a manual operation of selecting
  single bits that correspond to predefined key positions
  in the keyset array. See the definition below.

  For example,
  we use the following mechanism in C:
     long int all_keys[8] =
      { 0,  0, 0, 0x00F0F7D0, 0x7FFFFFFF, 0xFFFFFFFF, 0xFFFFFFF9, 0 };
  to correspond to the Pascal definitions:
      mousekeys : gpr_$keyset_t := [KBD_$M1D..KBD_$M3D];
      all_keys : gpr_$keyset_t := ['#'..'~', ' ', KBD_$BS, KBD_$CR, KBD_$TAB,
                                   KBD_$L4, KBD_$L6..KBD_$LA, KBD_$LC..KBD_$LF, KBD_$R5];
*************************************************************************/

unsigned long mousekeys[8] =
{
   0, 0, 0, 0, 0xE, 0 , 0, 0
};

unsigned long all_keys[8] =
{
   0,  0, 0, 0x00F0F7D0, 0x7FFFFFFF, 0xFFFFFFFF, 0xFFFFFFF9, 0
};

status_$t       status;

void enable_mouse ()
{
   gpr_$enable_input(gpr_$buttons, mousekeys, &status);
}



void enable_all_keys ()
{
   gpr_$enable_input(gpr_$keystroke, all_keys, &status);
   gpr_$set_cursor_active(true, &status);
}

/********************************************************************************
   End Bit #                                                         Start Bit #

          _______________________________________________________________
         | keyset(0)     |       |       |       |       |       |       |
    255          |      non-displayable characters       |       |         224
         |_______|_______|_______|_______|_______|_______|_______|_______|
         | keyset(1)     |       |       |       |       |       |       |
    223          |      non-displayable characters       |       |         192
         |_______|_______|_______|_______|_______|_______|_______|_______|
         | keyset(2)     |       |       |       |       |       |       |
    191          |      non-displayable characters       |       |         160
         |_______|_______|_______|_______|_______|_______|_______|_______|
         | keyset(3)     |       |       |       |       |       |       |
    159          |      non-displayable characters       |       |         128
         |_______|_______|_______|_______|_______|_______|_______|_______|
         | keyset(4)     |       |       |       |       |       |       |
    127     ~ } | { z y x|w v u t|s r q p|o n m l|k j u h|g f e d|c b a `   96
         |_______|_______|_______|_______|_______|_______|_______|_______|
         | keyset(5)     |       |       |       |       |       |       |
     95     ^ ] \ [ Z Y X|W V U T S R Q P|O N M L|K J I H|G F E D|C B A @   64
         |_______|_______|_______|_______|_______|_______|_______|_______|
         | keyset(6)     |       |       |       |       |       |       |
     63   ? > = <|; : 9 8|7 6 5 4|3 2 1 0|/ . - `|+ * ) (|' & % $|# " !     32
         |_______|_______|_______|_______|_______|_______|_______|_______|
         | keyset(7)     |       |       |       |       |       |       |
     31          |      non-displayable characters       |       |           0
         |_______|_______|_______|_______|_______|_______|_______|_______|
 char     | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | |
 offset   1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0

          So, if you'd like the following mouse keys
             KBD_$M1D 'a'  left-down
             KBD_$M2D 'b'  middle-down
             KBD_$M3D 'c'  right-down
          to be in your keyset do as follows :

          Find the keyset row that contains the characters 'a','b', and 'c';
          this is  keyset(4).  Then "turn on" the bit corresponding to each letter.
          This will be bits 1,2, and 3.

          In the array we called mousekeys, mousekeys[4] will have bits 1,2 and 3
          set. This corresponds to a hex E. We use the following C expression
          to accomplish this:

          long int mousekeys[8] =
          { 0, 0, 0, 0, 0xE, 0 , 0, 0 };
          gpr_$enable_input(gpr_$buttons, mousekeys, &status);

******************************************************************************/
