{ gpr_example_1c.pas }
PROGRAM gpr_pas_example_1c;

%include '/sys/ins/base.ins.pas';
%include '/sys/ins/gpr.ins.pas';
%include '/sys/ins/error.ins.pas';
%include '/usr/include/tml/tml.ins.pas';

PROCEDURE create_bitmap(OUT return_bitmap : integer32); EXTERN;
PROCEDURE enable_all_keys; EXTERN;
PROCEDURE tml_dp_damage_repair(viewport_handle : tml_c_bitmap_id); EXTERN;

TYPE
   FILE_NAME_TYPE     = ARRAY[1..64] of CHAR;

VAR
    zero              :integer32            :=  0;
    filename          :FILE_NAME_TYPE;
    filename_length   :integer32;
    display_model     :TML_C_MODEL;
    display_viewport  :TML_C_VIEW_PORT;
    display_format    :TML_C_FORMAT;
    start_display     :C_MODEL_POS;
    p_start_display   :PTR_TO_C_MODEL_POS   := addr(start_display);
    bitmap            :integer32;

    dy                :INTEGER32;           { viewport height }
    dx                :INTEGER32;           { viewport width }
    inq_pos           :C_SCREEN_XY;         { any x,y coordinate on the screen }
    p_inq_pos         :PTR_TO_C_SCREEN_XY   := addr(inq_pos);

    last_page         :TML_C_PHEAD;         { used for going to bottom of file }
    last_line         :TML_C_CLINE;         { used for going to bottom of file }

    dummy_integer     :integer32;
    dummy_format      :TML_C_FORMAT;
    unob              :boolean;
    ev_type           :gpr_$event_t;
    ev_value          :char                 := '0';
    ret_coord         :gpr_$position_t;
    status            :status_$t;

{***************************************************************}
FUNCTION example_read( filename: FILE_NAME_TYPE;
                       f_len: INTEGER32): TML_C_MODEL;
{ example read -- construct a model and set properties }

VAR
    return_model      :TML_C_MODEL;         { the model created }
    file_id           :TEXT;
    st                :INTEGER32;
    current_node      :TML_C_NODE;
    current_gnode,
    last_gnode        :TML_C_GNODE;
    font_prop,
    margin_prop,
    wrap_prop         :TML_C_PROPERTY;
    err_stat          :status_$t;
    font_name         :ARRAY[1..25] OF CHAR;
    add_font_prop     :BOOLEAN;
    chars_back,
    chars_attached    :INTEGER32;
    char_buf          :VARYING [500] of CHAR;
    dummy_node        :TML_C_NODE;
    margin_size       :static integer32     :=  40;

BEGIN

{ open file from which model will be made }
OPEN(file_id, filename, 'OLD', st);

IF (st = 0) THEN
   BEGIN
   { pascal requires reset on file }
   RESET(file_id);
   { allocate model }
   return_model := model_class_alloc;
   IF (return_model <> NIL) THEN
      BEGIN
      { make root mode for model }
      current_node := model_make_root(return_model);
      IF (current_node <> NIL) THEN
         BEGIN
         { make a character wrap property }
         { associate it with the root node }
         { enables character wrap for entire document }
         {  except where overridden, for a subtree in }
         {  the model, by a different wrap property }
         {  attached to some other node }
         wrap_prop := tml_c_make_wrap_prop(char_wrap, return_model);
         IF (wrap_prop <> NIL) THEN
            BEGIN
            { connect wrap property to root node }
            tml_c_connect_prop(wrap_prop, current_node);

            { make a margin property }
            { use 40-pixel left and right margin }
            { no top or bottom margin }
            margin_prop :=
              tml_c_make_margin_prop(margin_size, margin_size,
                                     zero, zero, return_model);
            IF (margin_prop <> NIL) THEN
               BEGIN
               { connect margin property to root node }
               tml_c_connect_prop(margin_prop, current_node);

               { make a terminal node }
               current_gnode := gnode_class_alloc(return_model);

               IF (current_gnode <> NIL) THEN
                  BEGIN
                  { connect it as child of root }
                  dummy_node :=
                    node_connect_node(current_node,
                                      current_gnode, n_down);

                  { make a font property }
                  font_name := '/sys/dm/fonts/helvetica12';
                  font_prop :=
                    tml_c_make_font_prop(font_name[1],
                                         INTEGER32(25),
                                         return_model);

                  IF (font_prop <> NIL) THEN
                     BEGIN
                     { connect font property to first terminal node }
                     { overrides, for this node only (because it is a }
                     {  terminal node) the default font property value }
                     {  associated with each model }
                     tml_c_connect_prop(font_prop, current_gnode);

                     { read 256 bytes from file }
                     readln(file_id, char_buf);
                     char_buf.length := char_buf.length + 1;
                     chars_back := char_buf.length;
                     char_buf.body[chars_back] := chr(16#A);

                     IF (chars_back >= 1) THEN
                        { associate them with the newly-created node }
                        BEGIN
                        chars_attached :=
                          node_attach_data(current_gnode,
                                           addr(char_buf.body[1]), chars_back);

                        { check for read error or error in attaching data }
                        IF chars_attached = chars_back THEN
                           BEGIN
                           IF NOT EOF(file_id) THEN
                              BEGIN
                              add_font_prop:= FALSE;

                              { file-reading loop }
                              { read characters from file }
                              { create a new terminal node }
                              { associate characters with newly-created node }
                              { connect new node to last node created }
                              { remember newly-created node as last created }
                              { change font on alternate nodes }
                              REPEAT
                              readln(file_id, char_buf);
                              char_buf.length := char_buf.length + 1;
                              chars_back := char_buf.length ;
                              char_buf.body[chars_back] := chr(16#A);

                              IF chars_back >= 1 THEN
                                 BEGIN

                                 { remember last node created }
                                 last_gnode :=current_gnode;

                                 { create a new terminal node }
                                 current_gnode :=
                                   gnode_class_alloc(return_model);

                                 { quit if allocation fails }
                                 IF (current_gnode <> NIL) THEN
                                    BEGIN
                                 { associate characters read }
                                 {  with newly-created node }
                                    chars_attached :=
                                      node_attach_data(current_gnode,
                                                       addr(char_buf.body[1]),
                                                       chars_back);
                                    IF chars_attached = chars_back THEN
                                       BEGIN
                                       { attach font property }
                                       {  to alternate nodes }
                                       IF add_font_prop THEN
                                          BEGIN
                                          tml_c_connect_prop(font_prop,
                                              current_gnode);
                                          add_font_prop := FALSE;
                                          END
                                       ELSE add_font_prop := TRUE;

                                       { connect new node to last node }
                                       {  created }
                                       dummy_node :=
                                         node_connect_node(last_gnode,
                                                           current_gnode,
                                                           n_right);

                                       END

                                       ELSE BEGIN
                                       writeln('unable to attach all',
                                            ' characters properly');
                                       model_free(return_model);
                                       return_model:=NIL;
                                       EXIT;
                                       END;

                                    END
                                    ELSE BEGIN
                                    writeln('unable to alloc node in',
                                     'loop--returning partial model');
                                    EXIT;
                                    END;

                                 END;

                              UNTIL EOF(file_id);

                              END;

                           END
                           ELSE BEGIN
                           { unable to attach all characters read }
                           writeln('unable to attach all chars to node');
                           model_free(return_model);
                           return_model:=NIL;
                           END;

                        END
                        ELSE writeln('data file empty');

                     END
                     ELSE BEGIN
                     { failed to create font }
                     writeln('failed to create font properly');
                     return_model := NIL;
                     END;

                  END
                  ELSE BEGIN
                  { failed to create gnode }
                  writeln('unable to create gnode');
                  return_model := NIL;
                  END;

               END
               ELSE BEGIN
               { unable to create margin prop }
               writeln('unable to create margin property');
               return_model := NIL;
               END;

            END
            ELSE BEGIN
            { failed to create wrap_prop properly }
            writeln('unable_to_make wrap_prop');
            return_model := NIL;
            END;

         END
         ELSE BEGIN
         { node not created }
         writeln('unable to create current_node');
         return_model:=NIL
         END;

      END
      { cannot allocate model }
      ELSE writeln('unable to create model');

   END
   ELSE BEGIN
   writeln('unable to open file');
   return_model := NIL;
   END;

example_read := return_model;

END;

{***************************************************************}

BEGIN

{ initialize TMLib }
tml_init;

{ associate all viewport objects with the GPR graphic system }
tml_c_tml_dp_set_dflt_vpftbl(tml_c_tml_dp_get_gpr_vpftbl());

{***************************************************************}
{ create a model from the ASCII file test_file }
filename := 'test_file';
filename_length := 9;
display_model := example_read(filename, filename_length);
IF (display_model <> NIL) THEN
{***************************************************************}

   BEGIN
   { allocate a viewport }
   display_viewport := view_port_class_alloc;

   { create a graphics display area }
   { use a routine that depends on the graphics system }
   create_bitmap(bitmap);

   { assign display area to viewport }
   dummy_integer :=
     view_port_set_handle(display_viewport, tml_c_bitmap_id(addr(bitmap)));

   { create a format }
   { attach it to the viewport }
   dummy_format := NIL;
   dummy_integer:=
     view_port_attach_format(display_viewport, dummy_format, zero, zero);

   { get a handle for the format created by the last call }
   display_format := view_port_get_format(display_viewport);
   { attach the model to the format that is associated with the viewport }
   format_attach(display_format, display_model);

   start_display.in_node := model_get_root(display_model);
   { quit if model has no root node }
   IF (start_display.in_node <> NIL) THEN
      BEGIN

      { display top of file }
      tml_dp_damage_repair(view_port_get_handle(display_viewport));

      { enable keyboard }
      enable_all_keys();

      WHILE (ev_value <> 'e') DO
      { typing 'e' exits the program }

         BEGIN

         unob := gpr_$event_wait(ev_type, ev_value, ret_coord, status);

         CASE ev_value OF

         'u' :
            { move viewport up the conceptual page }
            {  half the viewport height }
            { get height of viewport -- it may have changed }
            BEGIN
            dy := view_port_get_hi(display_viewport);
            { find character at top of viewport }
            p_start_display :=
              view_port_xy_2_mpos(display_viewport,
                                  zero_zero, start_display);
            { if a character was found }
            {  display it halfway down the screen }
            IF (start_display.in_node <> NIL) THEN
               BEGIN
               { get true character location }
               p_inq_pos :=
                 view_port_mpos_2_xy(display_viewport, start_display,
                                     inq_pos, NIL);
               inq_pos.at_y := dy div 2;
               view_port_draw_at(display_viewport, start_display,
                                 inq_pos);
               END;
            END;

         'd' :
            { move the viewport down the conceptual page }
            {  half the viewport height }
            { get height of viewport -- it may have changed }
            BEGIN
            { find the character at pixel (0,dy/2) }
            dy := view_port_get_hi(display_viewport);
            inq_pos.at_x := 0;
            inq_pos.at_y := dy div 2;
            p_start_display :=
              view_port_xy_2_mpos(display_viewport,
                                  inq_pos, start_display);
            { if a character was found, display it }
            {  display it at type of screen }
            IF (start_display.in_node <> NIL) THEN
               { get true character location }
               p_inq_pos :=
                 view_port_mpos_2_xy(display_viewport, start_display,
                                     inq_pos, NIL);
               inq_pos.at_y := 0;
               view_port_draw_at(display_viewport, start_display,
                                 inq_pos);
            END;

         'l' :
            { move left on the conceptual page half the viewport width }
            { get width of viewport }
            BEGIN
            dx := view_port_get_width(display_viewport);
            { find the character if any at pixel (0,0) }
            p_start_display :=
              view_port_xy_2_mpos(display_viewport,
                                  zero_zero, start_display);
            { if a character was found, redisplay }
            IF (start_display.in_node <> NIL) THEN
               BEGIN
               { get true character location }
               p_inq_pos :=
                 view_port_mpos_2_xy(display_viewport, start_display,
                                     inq_pos, NIL);
               { move half a viewport left }
               inq_pos.at_x := inq_pos.at_x + (dx div 2);
               view_port_draw_at(display_viewport, start_display, inq_pos);
               END;
            END;

         'r' :
             { move right on conceptual page half the viewport width }
             { get width of viewport }
             BEGIN
             dx := view_port_get_width(display_viewport);
             { find the character if any at pixel (0,0) }
             p_start_display :=
               view_port_xy_2_mpos(display_viewport,
                                   zero_zero, start_display);
             { if a character was found, redisplay }
             IF (start_display.in_node <> NIL) THEN
                BEGIN
                { get true character location }
                p_inq_pos :=
                  view_port_mpos_2_xy(display_viewport,
                                      start_display, inq_pos, NIL);
                { move half a viewport right }
                inq_pos.at_x := inq_pos.at_x - (dx div 2);
                view_port_draw_at(display_viewport, start_display, inq_pos);
                END;
             END;

         't' :
            BEGIN
            { move to display top of file }
            { get root node of model }
            {  and move to first model position }
            start_display.in_node := model_get_root(display_model);
            start_display.char_index := 0;
            p_start_display :=
               model_pos_get_next(start_display, start_display);
            { if a character was found, display it at pixel (0,0) }
            IF (start_display.in_node <> NIL) THEN
                view_port_draw_at(display_viewport, start_display,
                                  zero_zero);
            END;

         'b' :
            { move to display end of file }
            { get root node of model }
            { get last node dominated by root }
            BEGIN
            start_display.in_node :=
              node_get_last(model_get_root(display_model));
            { quit if no character found }
            IF (start_display.in_node <> NIL) THEN
               BEGIN
               { get index of last character for that node }
               start_display.char_index :=
                 node_get_max_index(start_display.in_node);
               { find the conceptual page containing the model position }
               last_page :=
                 format_mpos_2_phead(display_format, start_display);
               { quit if no page found }
               IF (last_page <> NIL) THEN
                  BEGIN
                  { find last cline on last page }
                  last_line :=
                    phead_mpos_2_cline(last_page, start_display);
                  { quit if no line found }
                  IF (last_line <> NIL) THEN
                     BEGIN
                     { set up a screen_xy for bottom of page }
                     {  at height of last line }
                     inq_pos.at_x := 0;
                     inq_pos.at_y :=
                       view_port_get_hi(display_viewport)
                       - cline_get_hi(last_line);
                     view_port_draw(display_viewport, last_line, inq_pos);
                     END;

                 END;

              END;

            END;

         END;

         END;

      END;

   { release dynamically allocated storage }
   view_port_free(display_viewport);
   model_free(display_model);

   END;

END.

