#include <stdio.h>
#include <tml/tml_c_types.h>
#include <tml/tml_os_depends.h>

    /*****************************************/
#include <dlg/Xlib.h>
#include <dlg/Xutil.h>
#include <tml/tml_dp_x_graph.h>
#include <tml/tml_dp_graph.hxx>

#ifdef not_vanilla_C
extern struct window_info* make_window(long mask_selection);
extern char                x_exam_getchar(Display *display,
                                          tml_c_view_port display_viewport);
#else
extern struct window_info* make_window();
extern char                x_exam_getchar();
#endif

main()
{
    tml_c_model      display_model;
    tml_c_view_port  display_viewport;
    tml_c_format     display_format;
    C_MODEL_POS      start_display;
    long             mask_selection =
                       KeyPressMask |       /* Ask XWindows for notice */
                       ExposureMask |       /* of key, expose, and */
                       StructureNotifyMask; /* window size and location */
                                            /* change events */

    /***************************************/
    struct window_info   *window_desc;
    /***************************************/

    /*********************************************/
    int          c;
    long         dx, dy;    /* size of viewport */
    C_SCREEN_XY  inq_pos;   /* any (x,y) coordinate on the screen */
    tml_c_phead  last_page; /* used for going to bottom of file */
    tml_c_cline  last_line; /* used for going to bottom of file */
    /**********************************************/

    /* initialize TMLib */
    tml_init();

    /***************************************/
    /* set up XWindows */
    /* get information about display, etc. */
    window_desc = make_window(mask_selection);
    if (!window_desc) {
        fprintf(stderr, "unable to initialize X\n");
        exit(1);
        }
    /* set font display */
    /* MUST BE DONE BEFORE CREATING A MODEL */
    tml_dp_set_font_display(window_desc->display);
    /***************************************/

    /* create a model from the ASCII file test_file */
    display_model = tml_c_read_uasc_path("test_file", 9);
    if (!display_model)
        exit (1);

    /* allocate a viewport */
    display_viewport = view_port_class_alloc();

    /* assign the display area to the viewport */
    view_port_set_handle(display_viewport, window_desc);

    /* create a format */
    /* attach it to the viewport */
    view_port_attach_format(display_viewport, 0, 0, 0);
    /* get a handle for the format created by the last call */
    display_format = view_port_get_format(display_viewport);
    /* attach the model to the format that is associated with the viewport */
    format_attach(display_format, display_model);

    /*****************************************/
    /* typing 'e' exits the program */
    while ('e' !=
      (c = x_exam_getchar(window_desc->display, display_viewport))) {
        switch (c) {
            case 'u':
                /* move viewport up the conceptual page */
                /*   half the viewport height */
                /* get height of viewport -- it may have changed */
                dy = view_port_get_hi(display_viewport);
                /* find character at top of viewport */
                view_port_xy_2_mpos(display_viewport,
                                    &zero_zero, &start_display);
                /* if a character was found */
                /*   display it halfway down the screen */
                if (start_display.in_node) {
                    inq_pos.at_x = 0;
                    inq_pos.at_y = dy/2;
                    view_port_draw_at(display_viewport, &start_display,
                                      &inq_pos);
                    }
                break;

            case 'd':
                /* move the viewport down the conceptual page */
                /*   half the viewport height */
                /* get height of viewport -- it may have changed */
                dy = view_port_get_hi(display_viewport);
                /* find the character at pixel (0,dy/2) */
                inq_pos.at_x = 0;
                inq_pos.at_y = dy/2;
                view_port_xy_2_mpos(display_viewport,
                                    &inq_pos, &start_display);
                /* if a character was found, display it at pixel (0,0) */
                if (start_display.in_node)
                    view_port_draw_at(display_viewport,
                                      &start_display, &zero_zero);
                break;


            case 'l':
                /* move left on the conceptual page half the viewport width */
                /* get width of viewport */
                dx = view_port_get_width(display_viewport);
                /* find the character if any at pixel (0,0) */
                view_port_xy_2_mpos(display_viewport,
                                    &zero_zero, &start_display);
                /* if a character was found, display it at pixel (dx/2,0) */
                if (start_display.in_node) {
                    /* get true character location */
                    if (0 == view_port_mpos_2_xy(display_viewport,
                                                 &start_display, &inq_pos, 0))
                        break;
                    /* move half a viewport left */
                    inq_pos.at_x += (dx/2);
                    view_port_draw_at(display_viewport, &start_display,
                                      &inq_pos);
                    }
                break;

            case 'r':
                /* move right on conceptual page half the viewport width */
                /* get width of viewport */
                dx = view_port_get_width(display_viewport);
                /* find the character if any at pixel (0,0) */
                view_port_xy_2_mpos(display_viewport, &zero_zero,
                                    &start_display);
                /* if a character was found, display it at pixel (-dx/2,0) */
                if (start_display.in_node) {
                    /* get true character location */
                    if (0 == view_port_mpos_2_xy(display_viewport,
                                                 &start_display, &inq_pos, 0))
                        break;
                    /* move half a viewport right */
                    inq_pos.at_x -= (dx/2);
                    view_port_draw_at(display_viewport, &start_display,
                                      &inq_pos);
                    }
                break;

            case 't':
                /* move to display top of file */
                /* get root node of model */
                /*   and move to first model position */
                start_display.in_node = model_get_root(display_model);
                start_display.char_index = 0;
                model_pos_get_next(&start_display, &start_display);
                /* if a character was found, display it at pixel (0,0) */
                if (start_display.in_node)
                    view_port_draw_at(display_viewport,
                                      &start_display, &zero_zero);
                break;

            case 'b':
                /* move to display end of file */
                /* get root node of model */
                /* get last node dominated by root */
                start_display.in_node =
                  node_get_last( model_get_root(display_model) );
                /* quit if no character found */
                if (!start_display.in_node) break;
                /* get index of last character for that node */
                start_display.char_index =
                  node_get_max_index(start_display.in_node);
                /* find the conceptual page containing the model position */
                last_page = format_mpos_2_phead(display_format, &start_display);
                /* quit if no page found */
                if (!last_page) break;
                /* find the last cline for the last page */
                last_line = phead_mpos_2_cline(last_page, &start_display);
                /* quit if no line found */
                if (!last_line) break;
                /* set up a screen_xy for the bottom of the page */
                /*   at the height of the last line */
                inq_pos.at_x = 0;
                inq_pos.at_y =
                  view_port_get_hi(display_viewport)
                  - cline_get_hi(last_line);
                view_port_draw(display_viewport, last_line, &inq_pos);
                break;

            default:
                break;
        }
    }
    /* release dynamically allocated storage */
    /* freeing the viewport frees the window_info structure as well */
    view_port_free( display_viewport);
    model_free(display_model);
    /*****************************************/
}

