#include <stdio.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>

    /*****************************************/
#include <dlg/Xlib.h>
#include <dlg/Xutil.h>
#include <tml/tml_dp_x_graph.h>
#include <tml/tml_dp_graph.hxx>

extern     window_info* make_window(long mask_selection);

main()
{
    model     *display_model;
    view_port *display_viewport;
    model_pos  start_display;
    long       mask_selection = KeyPressMask |       /* Ask XWindows for */
                                ExposureMask |       /* notice of key, */
                                StructureNotifyMask; /* expose, and window */
                                                     /* size and location */
                                                     /* change events */
    /***************************************/
    window_info    *window_desc;
    XEvent          return_event;  /* records events */
    KeySym          keysym_return;
    XComposeStatus  status_return;
    char            return_char;
    /***************************************/

    int        c;
    long       dx, dy;          /* size of viewport */
    screen_xy  inq_pos;         /* any (x,y) coordinate on the screen */
    phead     *last_page;       /* used for going to bottom of file */
    cline     *last_line;       /* used for going to bottom of file */
    format    *use_format;
    property  *wrap_property,
              *margin_property;
    /***************************************/

    /* set up XWindows */
    /* get information about display, etc. */
    window_desc = make_window(mask_selection);
    if (!window_desc) {
        fprintf(stderr, "unable to initialize X\n");
        exit(1);
        }
    /* set font display */
    /* MUST BE DONE BEFORE CREATING A MODEL */
    tml_dp_set_font_display(window_desc->display);
    /***************************************/

    /* create a model from the ASCII file test_file */
    display_model = read_uasc("test_file", 9);

    /* allocate a viewport */
    display_viewport = new view_port;

    /* assign the display area to the viewport */
    display_viewport->set_handle(window_desc);

    /* create a format */
    use_format = new format;
    if (!use_format) {
        fprintf(stderr, "unable to create format\n");
        exit(1);
        }
    /* attach it to the viewport, setting format page size */
    /*   to 4 inches wide by 5 inches high for displays with */
    /*   resolution of 72 pixels per inch */
    display_viewport->attach(use_format, (4 * 72), (5 * 72));
    /* attach the model to the format that is associated with the viewport */
    use_format->attach(display_model);

    /* get root node of model */
    start_display.in_node = display_model->get_root();
    /* quit on failure */
    if (!start_display.in_node)
        exit(1);

    /* make word wrap property */
    wrap_property = make_wrap_prop(word_wrap, display_model);
    if (!wrap_property) {
        fprintf(stderr, "unable to make wrap property\n");
        exit (1);
        }
    /* attach it to root node of model */
    connect_prop(wrap_property, start_display.in_node);

    /* make margin property */
    /* set top and bottom margins to 10 pixels */
    /*   left and right margins to 20 pixels */
    margin_property = make_margin_prop(20, 20, 10, 10, display_model);
    if (!margin_property) {
        fprintf(stderr, "unable to make the margin property\n");
        exit(1);
        }
    /* attach it to root node of model */
    connect_prop(margin_property, start_display.in_node);

    while (1) {
        XNextEvent(window_desc->display, &return_event);
        switch (return_event.type) {
          case ConfigureNotify:
            tml_dp_record_new_size(display_viewport->get_handle(),
                                   return_event.xconfigure.width,
                                   return_event.xconfigure.height);
            break;
          case Expose:   
            /* expose event */
            /* refresh display */
            tml_dp_add_clip_rect(display_viewport->get_handle(),
                                 return_event.xexpose.x,
                                 return_event.xexpose.y,
                                 return_event.xexpose.width,
                                 return_event.xexpose.height);
            if (return_event.xexpose.count == 0)
                tml_dp_damage_repair(display_viewport->get_handle());
            break;

          case KeyPress:
            c = XLookupString((XKeyEvent*) &return_event, &return_char, 1,
                              &keysym_return, &status_return);
            switch (return_char) {
                case 'e':
                    /* exit the program */
                    delete display_viewport;
                    delete use_format;
                    delete display_model;
                    exit(0);
                    break;

                case 'u':
                    /* move viewport up the conceptual page */
                    /*   half the viewport height */
                    /* get height of viewport -- it may have changed */
                    dy = display_viewport->get_hi();
                    /* find character at top of viewport */
                    start_display = display_viewport->xy_2_mpos(zero_zero);
                    /* if a character was found */
                    /*   display it halfway down the screen */
                    if (start_display.in_node) {
                        /* get true character location */
                        inq_pos = display_viewport->mpos_2_xy(start_display);
                        inq_pos.at_y = dy/2;
                        display_viewport->draw(start_display, inq_pos);
                        }
                    break;

                case 'd':
                    /* move the viewport down the conceptual page */
                    /*   half the viewport height */
                    /* get height of viewport -- it may have changed */
                    dy = display_viewport->get_hi();
                    /* find the character at pixel (0,dy/2) */
                    inq_pos.at_x = 0;
                    inq_pos.at_y = dy/2;
                    start_display = display_viewport->xy_2_mpos(inq_pos);
                    /* if a character was found, display it */
                    /*   at top of screen */
                    if (start_display.in_node) {
                        /* get true character location */
                        inq_pos = display_viewport->mpos_2_xy(start_display);
                        inq_pos.at_y = 0;
                        display_viewport->draw(start_display, inq_pos);
                        }
                    break;

                case 'l':
                    /* move left on conceptual page half the viewport width */
                    /* get width of viewport */
                    dx = display_viewport->get_width();
                    /* find the character if any at pixel (0,0) */
                    start_display = display_viewport->xy_2_mpos(zero_zero);
                    /* if a character was found, redisplay */
                    if (start_display.in_node) {
                        /* get true character location */
                        inq_pos = display_viewport->mpos_2_xy(start_display);
                        /* move half a viewport left */
                        inq_pos.at_x += (dx/2);
                        display_viewport->draw(start_display, inq_pos);
                        }
                    break;

                case 'r':
                    /* move right on conceptual page half the viewport width */
                    /* get width of viewport */
                    dx = display_viewport->get_width();
                    /* find the character if any at pixel (0,0) */
                    start_display = display_viewport->xy_2_mpos(zero_zero);
                    /* if a character was found, redisplay */
                    if (start_display.in_node) {
                        /* get true character location */
                        inq_pos = display_viewport->mpos_2_xy(start_display);
                        /* move half a viewport right */
                        inq_pos.at_x -= (dx/2);
                        display_viewport->draw(start_display, inq_pos);
                        }
                    break;

                case 't':
                    /* move to display top of file */
                    /* get root node of model */
                    /*   and move to first model position */
                    start_display.in_node = display_model->get_root();
                    start_display.char_index = 0;
                    start_display = start_display.get_next();
                    /* if a character was found, display it at pixel (0,0) */
                    if (start_display.in_node)
                        display_viewport->draw(start_display, zero_zero);
                    break;

                case 'b':
                    /* move to display end of file */
                    /* get root node of model */
                    /* get last node dominated by root */
                    start_display.in_node =
                      display_model->get_root()->get_last();
                    /* quit if no character found */
                    if (!start_display.in_node) break;
                    /* get index of last character for that node */
                    start_display.char_index =
                      start_display.in_node->get_max_index();
                    /* find the conceptual page containing the model position */
                    last_page = use_format->mpos_2_phead(start_display);
                    /* quit if no page found */
                    if (!last_page) break;
                    /* find last cline on last page */
                    last_line = last_page->mpos_2_cline(start_display);
                    /* quit if no line found */
                    if (!last_line) break;
                    /* set up a screen_xy for the bottom of the page */
                    /*   at the height of the last line */
                    inq_pos.at_x = 0;
                    inq_pos.at_y =
                      display_viewport->get_hi()
                      - last_line->get_hi();
                    display_viewport->draw(last_line, inq_pos);
                    break;

                default:
                    break;
            }
        default:
            break;
        }
    }
    /* release dynamically allocated storage */
    delete display_viewport;
    delete display_model;
}
