#include <stdio.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml_exam/tml_x_exam_kbd.h>

    /*****************************************/
#include <dlg/Xlib.h>
#include <dlg/Xutil.h>
#include <tml/tml_dp_x_graph.h>
#include <tml/tml_dp_graph.hxx>

extern     window_info* make_window(long mask_selection);

main()
{
    model     *display_model;
    view_port *display_viewport;
    format    *display_format;
    model_pos  start_display;
    long       mask_selection = KeyPressMask |       /* Ask XWindows for */
                                ButtonPressMask |    /* notice of key, */
                                ExposureMask |       /* mouse, expose, and */
                                StructureNotifyMask; /* window size and */
                                                     /* location change */
                                                     /* events */
    /***************************************/
    window_info *window_desc;
    XEvent       return_event;  /* records events */
    KeySym       keysym_return;
    /***************************************/

    long       dx, dy;  /* size of viewport */
    screen_xy  inq_pos; /* any (x,y) coordinate on the screen */

    v_cursor  *the_cursor;
    screen_xy  curs_loc;
    model_pos  curs_pos;
    cline     *old_top_line;
    cline     *new_top_line;
    cline     *first_line;
    phead     *first_page;
    xy_coord   at_coord;
    char       text[10];
    int        textlen;
    phead     *on_phead;
    cline     *in_cline;
    char_set  *in_char_set;
    font_tbl  *use_font;
    long       char_val;

    /***************************************/
    /* set up XWindows */
    /* get information about display, etc. */
    window_desc = make_window(mask_selection);
    if (!window_desc) {
        fprintf(stderr, "unable to initialize X\n");
        exit(1);
        }
    /* set font display */
    /* MUST BE DONE BEFORE CREATING A MODEL */
    tml_dp_set_font_display(window_desc->display);
    /***************************************/

    /* create a model from the ASCII file test_file */
    display_model = read_uasc("test_file", 9);

    /* allocate a viewport */
    display_viewport = new view_port;

    /* assign the display area to the viewport */
    display_viewport->set_handle(window_desc);

    /* create a format */
    /* attach it to the viewport */
    display_viewport->attach((format*) 0);
    /* get a handle for the format created by the last call */
    display_format = display_viewport->get_format();
    /* attach model to format that is associated with viewport */
    display_format->attach(display_model);

    /* allocate cursor */
    /* attach cursor to viewport at first character */
    the_cursor = new v_cursor;
    if (!the_cursor) {
        fprintf(stderr, "unable to allocate a cursor\n");
        exit(1);
        }

    display_viewport->attach(the_cursor, zero_zero);
    the_cursor->enable();

    while (1) {
        XNextEvent(window_desc->display, &return_event);
        switch (return_event.type) {

            case ConfigureNotify:
                /* resize event */
                tml_dp_record_new_size(display_viewport->get_handle(),
                                       return_event.xconfigure.width,
                                       return_event.xconfigure.height);
                break;

            case Expose:
                /* refresh event */
                tml_dp_add_clip_rect(display_viewport->get_handle(),
                                     return_event.xexpose.x,
                                     return_event.xexpose.y,
                                     return_event.xexpose.width,
                                     return_event.xexpose.height);
                if (return_event.xexpose.count == 0)
                    tml_dp_damage_repair(display_viewport->get_handle());
                break;

            case ButtonPress:
                /* mouse event */
                /* ignore all except ButtonPress on Button1 */
                if (Button1 == return_event.xbutton.button) {
                    /* move TMLib cursor to where pointing cursor */
                    /*   was at click */
                    inq_pos.at_x = return_event.xbutton.x;
                    inq_pos.at_y = return_event.xbutton.y;
                    the_cursor->set(inq_pos);
                    }
                break;

            case KeyPress:
                textlen = XLookupString((XKeyEvent*) &return_event,
                                        text, 10, &keysym_return, 0);
                text[textlen] = (char)0;
                switch (keysym_return) {

                    /* The constant values for keyboard return values */
                    /*   in the case statements come from tml_x_exam_kbd.h. */
                    /* Keysym return values are sensitive to the keyboard map */
                    /*   that is in effect at runtime. */
                    /* To use different values, edit tml_x_exam_kbd.h. */
                    /* NOTE: On Apollo systems running Xapollo with the -k */
                    /*   option, these keys are reserved for the Display */
                    /*   Manager (DM), and are never passed to the program. */

                    case tmlScrollUp:
                        /* up-box-arrow key: scroll up half a viewport */
                        /* get (x,y) position of cursor in viewport */
                        curs_loc = the_cursor->get_raw_loc();
                        /* get height of viewport -- it may have changed */
                        dy = display_viewport->get_hi();
                        /* get top line in viewport */
                        old_top_line =
                          display_viewport->get_top_line()->get_cline();
                        /* find the character if any at top of viewport */
                        start_display = display_viewport->xy_2_mpos(zero_zero);
                        /* if a character was found */
                        /*   display it halfway down the screen */
                        if (start_display.in_node) {
                            inq_pos =
                              display_viewport->mpos_2_xy(start_display);
                            inq_pos.at_y = dy/2;
                            display_viewport->draw(start_display, inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            curs_pos = the_cursor->get_model_pos();
                            if (!display_viewport->mpos_2_vpos(curs_pos))
                                the_cursor->set(curs_loc);
                            }
                        /* if no movement and line above exists, */
                        /*   go to top of file */
                        /* get new top line of viewport */
                        new_top_line =
                          display_viewport->get_top_line()->get_cline();
                        if (new_top_line &&
                          (new_top_line == old_top_line) &&
                          (new_top_line->get_prior())) {
                            /* no movement and prior text line exists */
                            /* get first position in model */
                            start_display.in_node =
                              display_format->get_model()->get_root();
                            start_display.char_index = 0;
                            start_display = start_display.get_next();
                            /* get and correct position of first position */
                            /*   on its conceptual page */
                            first_page =
                              display_format->mpos_2_phead(start_display);
                            first_line =
                              first_page->mpos_2_cline(start_display);
                            at_coord = first_line->mpos_2_xy(start_display);
                            inq_pos.at_x = at_coord.x_coord
                              + display_viewport->get_hoffset();
                            inq_pos.at_y = 0;
                            display_viewport->draw(start_display, inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            curs_pos = the_cursor->get_model_pos();
                            if (!display_viewport->mpos_2_vpos(curs_pos))
                                the_cursor->set(curs_loc);
                            }
                        break;

                    case tmlScrollDown:
                        /* down-box arrow key: scroll down half a viewport */
                        /* get (x,y) position of cursor in viewport */
                        curs_loc = the_cursor->get_raw_loc();
                        /* get height of viewport -- it may have changed */
                        dy = display_viewport->get_hi();
                        /* find character halfway down and furthest left */
                        inq_pos.at_x = 0;
                        inq_pos.at_y = dy/2;
                        start_display = display_viewport->xy_2_mpos(inq_pos);
                        /* if character found, display at top of viewport */
                        if (start_display.in_node) {
                            /* get true location of character */
                            inq_pos =
                              display_viewport->mpos_2_xy(start_display);
                            inq_pos.at_y = 0;
                            display_viewport->draw(start_display, inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            curs_pos = the_cursor->get_model_pos();
                            if (!display_viewport->mpos_2_vpos(curs_pos))
                                the_cursor->set(curs_loc);
                            }
                        break;

                    case tmlScrollLeft:
                        /* left-box arrow key: */
                        /*   scroll left 1/4 viewport width */
                        /* get (x,y) position of cursor in viewport */
                        curs_loc = the_cursor->get_raw_loc();
                        /* get width of viewport -- it may have changed */
                        dx = display_viewport->get_width();
                        /* find character at pixel (0,0) */
                        start_display = display_viewport->xy_2_mpos(zero_zero);
                        if (start_display.in_node) {
                            /* get (x,y) of topmost character */
                            inq_pos =
                              display_viewport->mpos_2_xy(start_display);
                            /* increment x coordinate by 1/4 viewport width */
                            inq_pos.at_x += dx/4;
                            /* display at new coordinates */
                            display_viewport->draw(start_display, inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            inq_pos = the_cursor->get_caged_loc();
                            if (inq_pos.at_x > dx)
                                the_cursor->set(curs_loc);
                            }
                        break;

                    case tmlScrollRight:
                        /* right-box arrow key: */
                        /*   scroll right 1/4 viewport width */
                        /* get (x,y) position of cursor */
                        curs_loc = the_cursor->get_raw_loc();
                        /* get width of viewport -- it may have changed */
                        dx = display_viewport->get_width();
                        /* find character at pixel (0,0) */
                        start_display = display_viewport->xy_2_mpos(zero_zero);
                        if (start_display.in_node) {
                            /* get (x,y) of topmost character */
                            inq_pos =
                              display_viewport->mpos_2_xy(start_display);
                            /* decrement x coordinate by 1/4 viewport width */
                            inq_pos.at_x -= dx/4;
                            /* display at new coordinates */
                            display_viewport->draw(start_display, inq_pos);
                            /* if cursor character not on screen, */
                            /*   set cursor at position it had before scroll */
                            inq_pos = the_cursor->get_caged_loc();
                            if (inq_pos.at_x < 0)
                                the_cursor->set(curs_loc);
                            curs_loc = the_cursor->get_raw_loc();
                            curs_pos = the_cursor->get_model_pos();
                            if (curs_loc.at_x < 0) {
                                /* reset cursor to right edge of character */
                                /*   if left edge is off screen */
                                on_phead=display_format->mpos_2_phead(curs_pos);
                                in_cline = on_phead->mpos_2_cline(curs_pos);
                                in_char_set = (char_set*)
                                  curs_pos.in_node->get_model()->inq_state(
                                    curs_pos.in_node, tml_char_set_prop);
                                char_val = in_char_set->as_ulong(
                                  ((unsigned char*) 
                                  curs_pos.in_node->get_contents())
                                  + curs_pos.char_index);
                                use_font = ((tline*) in_cline)->get_font();
                                curs_loc.at_x += use_font->get_width(char_val)
                                  + use_font->get_icspace();
                                the_cursor->set(curs_loc);
                                }
                            }
                        break;

                    case tmlExit:
                        /* exit key: exit the program */
                        /* release dynamically allocated storage */
                        /* freeing viewport frees cursor */
                        delete display_viewport;
                        /* delete the format */
                        delete display_format;
                        /* delete model */
                        delete display_model;
                        /* exit */
                        exit(0);

                    /**********************************************************/
                    case tmlBackspace:
                        /* backspace key: delete character to left */
                        /*   of TMLib cursor */
                        /* try to move cursor left */
                        curs_pos = the_cursor->get_model_pos();
                        curs_loc = the_cursor->move(vp_left);
                        /* quit on failure to move */
                        /*   (cursor at top of file) */
                        if (curs_pos != the_cursor->get_model_pos()) {
                            /* delete character */
                            display_viewport->delete_char(
                              the_cursor->get_model_pos());
                            /* resetting cursor updates its model position */
                            the_cursor->set(curs_loc);
                            }
                        break;

                    case tmlReturn:
                        /* carriage return key: insert a return after cursor */
                        display_viewport->ins_after_cursor("\n", 1);
                        break;

                    default:
                        /* any other key: insert character */
                        if(textlen > 0)
                          display_viewport->ins_after_cursor(text, textlen);
                        break;
                    /**********************************************************/
                }
                break;

            default:
                break;
            }

        }
}


