/*************************************************************************************
    csm_c_interface.cxx
    definitions for functions declared in csm_c_types.h for the C language
      interface of the Character Set Manager (CSM)

    (c) Copyright Apollo Computer Inc., 1990 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/csm_cxx2.0_linkage_decls.hxx>
#include <tml/char_set_mgr.hxx>

/**********************************************************************************
    typedefs for C++ classes

**********************************************************************************/

typedef void *csm_c_csm_character;
typedef void *csm_c_char_range;
typedef void *csm_c_char_range_set;
typedef void *csm_c_collation_map;
typedef void *csm_c_char_set;
typedef void *csm_c_alpha_char_set;
typedef void *csm_c_syll_frag_char_set;
typedef void *csm_c_union_char_set;
typedef void *csm_c_char_set_heap;
typedef void *csm_c_cs_heap;
typedef void *csm_c_csm_string;
typedef void *csm_c_kmp_node;
typedef void *csm_c_xregister;
typedef void *csm_c_char_func_tbl_entry;
typedef void *csm_c_char_func_tbl;
typedef void *tml_c_gc_hash_tbl;
typedef void *tml_c_tml_ext_type;
typedef void *tml_c_endian;
typedef void *tml_c_cs_type;
typedef long  csm_c_char_set_dir;

#ifdef __cplusplus
extern "C" {
#endif

/* FOR CLASS CSM_CHARACTER */

csm_c_csm_character csm_character_class_alloc_1()
{
    return new csm_character;
}

csm_c_csm_character csm_character_class_alloc_2(csm_c_char_set for_char_set)
{
    return (csm_c_csm_character) new csm_character((char_set*) for_char_set);
}

csm_c_csm_character csm_character_class_alloc_3(unsigned char* val, csm_c_char_set for_char_set)
{
    return new csm_character(val, (char_set*) for_char_set);
}

csm_c_csm_character csm_character_class_alloc_4(unsigned char* val, unsigned long len, csm_c_char_set for_char_set)
{
    return new csm_character(val, len, (char_set*) for_char_set);
}

csm_c_csm_character csm_character_class_alloc_5(unsigned char* val, unsigned long len, unsigned long collation_order, csm_c_char_set for_char_set)
{
    return new csm_character(val, len, collation_order, (char_set*) for_char_set);
}

void                csm_character_free(csm_c_csm_character csm_char_p)
{
    delete (char_set*) csm_char_p;
}

void                csm_character_init(csm_c_csm_character csm_char_p, unsigned char* val, unsigned long len, csm_c_char_set for_char_set, unsigned long index)
{
    ((csm_character*) csm_char_p)->init(val, len, (char_set*) for_char_set, index);
}

unsigned char*      csm_character_get_value(csm_c_csm_character csm_char_p)
{
    return ((csm_character*) csm_char_p)->get_value();
}

unsigned long       csm_character_get_num_chars(csm_c_csm_character csm_char_p)
{
    return ((csm_character*) csm_char_p)->get_num_chars();
}

unsigned char*      csm_character_get_byte_at(csm_c_csm_character csm_char_p, long byte_index)
{
    return ((csm_character*) csm_char_p)->get_byte_at(byte_index);
}

unsigned long       csm_character_get_collation(csm_c_csm_character csm_char_p)
{
    return ((csm_character*) csm_char_p)->get_collation();
}

csm_c_char_set      csm_character_get_char_set(csm_c_csm_character csm_char_p)
{
    return ((csm_character*) csm_char_p)->get_char_set();
}

void                csm_character_set_collation(csm_c_csm_character csm_char_p, unsigned long new_order)
{
    ((csm_character*) csm_char_p)->set_collation(new_order);
}


/* FOR CLASS CHAR_RANGE */

csm_c_char_range char_range_class_alloc_1()
{
    return new char_range();
}

csm_c_char_range    char_range_class_alloc_2(unsigned char* lo, unsigned char* hi, csm_c_char_set for_char_set, unsigned long lo_map, unsigned long hi_map)
{
    return new char_range(lo, hi, (char_set*) for_char_set, lo_map, hi_map);
}

csm_c_char_range    char_range_class_alloc_3(unsigned char* lo, unsigned long lo_len, unsigned char* hi, unsigned long hi_len, csm_c_char_set for_char_set, unsigned long lo_map, unsigned long hi_map)
{
    return new char_range(lo, lo_len, hi, hi_len, (char_set*) for_char_set, lo_map, hi_map);
}

void                char_range_free(csm_c_char_range range_p)
{
    delete (char_range*) range_p;
}

csm_c_csm_character char_range_get_high(csm_c_char_range range_p)
{
    return ((char_range*) range_p)->get_high();
}

csm_c_csm_character char_range_get_low(csm_c_char_range range_p)
{
    return ((char_range*) range_p)->get_low();
}

unsigned long       char_range_get_high_index(csm_c_char_range range_p)
{
    return ((char_range*) range_p)->get_high_index();
}

unsigned long       char_range_get_low_index(csm_c_char_range range_p)
{
    return ((char_range*) range_p)->get_low_index();
}

unsigned long       char_range_get_size(csm_c_char_range range_p)
{
    return ((char_range*) range_p)->get_size();
}

long                char_range_contains_char(csm_c_char_range range_p, unsigned char* candidate)
{
    return ((char_range*) range_p)->contains(candidate);
}

long                char_range_contains_coll(csm_c_char_range range_p, unsigned long collation)
{
    return ((char_range*) range_p)->contains(collation);
}

long                char_range_contains_nt_char(csm_c_char_range range_p, unsigned char* candidate, unsigned long len)
{
    return ((char_range*) range_p)->contains(candidate, len);
}


/* FOR CLASS CHAR_RANGE_SET */

csm_c_char_range_set char_range_set_class_alloc()
{
    return new char_range_set;
}

void                 char_range_set_free(csm_c_char_range_set range_set_p)
{
    delete (char_range_set*) range_set_p;
}

long                 char_range_set_contain(csm_c_char_range range_p, unsigned char* candidate)
{
    return ((char_range_set*) range_p)->contain(candidate);
}

csm_c_char_range     char_range_set_get_range(csm_c_char_range range_p, unsigned long i)
{
    return ((char_range_set*) range_p)->get_range(i);
}

csm_c_char_range     char_range_set_get_ranges(csm_c_char_range range_p)
{
    return ((char_range_set*) range_p)->get_ranges();
}

unsigned long        char_range_set_get_num_ranges(csm_c_char_range range_p)
{
    return ((char_range_set*) range_p)->get_num_ranges();
}

void                 char_range_set_make_ranges(csm_c_char_range range_p, long new_num)
{
    ((char_range_set*) range_p)->make_ranges(new_num);
}

void                 char_range_set_init_range(csm_c_char_range range_p, unsigned long slot, unsigned char* lo, unsigned char* hi, csm_c_char_set for_char_set, unsigned long spec_size, unsigned long lo_map, unsigned long hi_map)
{
    ((char_range_set*) range_p)->init_range(slot, lo, hi, (char_set*) for_char_set, spec_size, lo_map, hi_map);
}

void                 char_range_set_set_range_1(csm_c_char_range range_p, csm_c_char_range from_range, csm_c_char_set for_char_set, unsigned long slot)
{
    ((char_range_set*) range_p)->set_range(((char_range*) from_range)->get_high()->get_value(), ((char_range*) from_range)->get_low()->get_value(), (char_set*) for_char_set, slot);
}

void                 char_range_set_set_range_2(csm_c_char_range range_p, unsigned char* lo, unsigned char* hi, csm_c_char_set for_char_set, unsigned long slot, unsigned long lo_map, unsigned long hi_map)
{
    ((char_range_set*) range_p)->set_range(lo, hi, (char_set*) for_char_set, slot, lo_map, hi_map);
}

void                 char_range_set_set_range_3(csm_c_char_range range_p, unsigned char* lo, unsigned long lo_len, unsigned char* hi, unsigned long hi_len, csm_c_char_set for_char_set, unsigned long slot, unsigned long lo_map, unsigned long hi_map)
{
    ((char_range_set*) range_p)->set_range(lo, lo_len, hi, hi_len, (char_set*) for_char_set, slot, lo_map, hi_map);
}

unsigned long        char_range_set_lookup_char(csm_c_char_range range_p, unsigned char* sought)
{
    return ((char_range_set*) range_p)->lookup(sought);
}

unsigned char*       char_range_set_lookup_coll(csm_c_char_range range_p, unsigned long collation, unsigned char* c)
{
    return ((char_range_set*) range_p)->lookup(collation, c);
}

unsigned char*       char_range_set_corresponding(csm_c_char_range range_p, csm_c_char_range_set other_set, unsigned char* c)
{
    return ((char_range_set*) range_p)->corresponding(*((char_range_set*) other_set), c);
}


/* FOR CLASS COLLATION_MAP */

csm_c_collation_map collation_map_class_alloc_1()
{
    return new collation_map;
}

csm_c_collation_map collation_map_class_alloc_2(csm_c_char_range sequences, unsigned long num_sequences, csm_c_char_set for_char_set, unsigned long start_at)
{
    return new collation_map((char_range*) sequences, num_sequences, (char_set*) for_char_set, start_at);
}

void                collation_map_free(csm_c_collation_map map_p)
{
    delete (collation_map*) map_p;
}

csm_c_kmp_node      collation_map_get_many_to_one(csm_c_collation_map map_p)
{
    return ((collation_map*) map_p)->get_many_to_one();
}

void                collation_map_make_secondaries(csm_c_collation_map map_p, csm_c_char_range sequences, unsigned long num_sequences)
{
    ((collation_map*) map_p)->make_secondaries((char_range*) sequences, num_sequences);
}

void                collation_map_make_one_to_many(csm_c_collation_map map_p, csm_c_char_range sequences, unsigned long num_sequences)
{
    ((collation_map*) map_p)->make_one_to_many((char_range*) sequences, num_sequences);
}

void                collation_map_make_substitutes(csm_c_collation_map map_p, char** find, char** sub, unsigned long num_subs)
{
}

unsigned long       collation_map_lookup(csm_c_collation_map map_p, unsigned char** sought)
{
    return ((collation_map*) map_p)->lookup(*sought);
}

long                collation_map_compare(csm_c_collation_map map_p, unsigned char* s1, unsigned char* s2)
{
    return ((collation_map*) map_p)->compare(s1, s2);
}

long                collation_map_ncompare(csm_c_collation_map map_p, unsigned char* s1, unsigned char* s2, unsigned long n)
{
    return ((collation_map*) map_p)->ncompare(s1, s2, n);
}


/* FOR CLASS CHAR_SET */

csm_c_char_set       char_set_class_alloc(unsigned long per_char_size)
{
    return new char_set(per_char_size);
}

void                 char_set_free(csm_c_char_set cs_p)
{
    delete (char_set*) cs_p;
}

csm_c_char_set       char_set_make_another(csm_c_char_set cs_p, unsigned long c_size)
{
    return ((char_set*) cs_p)->make_another(c_size);
}

unsigned long        char_set_is_std(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->is_std();
}

unsigned char*       char_set_cs_strcat (csm_c_char_set cs_p, unsigned char* s1, unsigned char* s2)
{
    return ((char_set*) cs_p)->cs_strcat(s1, s2);
}

unsigned char*       char_set_cs_strncat(csm_c_char_set cs_p, unsigned char* s1, unsigned char* s2, long n)
{
    return ((char_set*) cs_p)->cs_strncat(s1, s2, n);
}

long                 char_set_cs_strcmp (csm_c_char_set cs_p, unsigned char* s1, unsigned char* s2)
{
    return ((char_set*) cs_p)->cs_strcmp (s1, s2);
}

long                 char_set_cs_strncmp(csm_c_char_set cs_p, unsigned char* s1, unsigned char* s2, long n)
{
    return ((char_set*) cs_p)->cs_strncmp(s1, s2, n);
}

unsigned char*       char_set_cs_strcpy (csm_c_char_set cs_p, unsigned char* s1, unsigned char* s2)
{
    return ((char_set*) cs_p)->cs_strcpy (s1, s2);
}

unsigned char*       char_set_cs_strncpy(csm_c_char_set cs_p, unsigned char* s1, unsigned char* s2, long n)
{
    return ((char_set*) cs_p)->cs_strncpy(s1, s2, n);
}

long                 char_set_cs_strlen (csm_c_char_set cs_p, unsigned char* s)
{
    return ((char_set*) cs_p)->cs_strlen (s);
}

unsigned char*       char_set_cs_index  (csm_c_char_set cs_p, unsigned char* s, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_index  (s, c);
}

unsigned char*       char_set_cs_rindex (csm_c_char_set cs_p, unsigned char* s, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_rindex (s, c);
}

tml_c_cs_type        char_set_get_type(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_type();
}

unsigned char*       char_set_get_name(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_name();
}

csm_c_xregister      char_set_get_xreg(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_xreg();
}

unsigned long        char_set_get_char_size(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_char_size();
}

unsigned long        char_set_get_num_chars(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_num_chars();
}

unsigned long        char_set_get_ref_count(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_ref_count();
}

csm_c_char_set_dir   char_set_get_predom_direc(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_predom_direc();
}

long                 char_set_get_dir_to_line(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_dir_to_line();
}

long                 char_set_get_dir_of_line(csm_c_char_set cs_p, long dir)
{
    return ((char_set*) cs_p)->get_dir_of_line((lrud_dir) dir);
}

csm_c_char_set       char_set_get_next(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_next();
}

csm_c_collation_map  char_set_get_collate_map(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_collate_map();
}

csm_c_char_range_set char_set_get_legal_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_legal_ranges();
}

csm_c_char_range_set char_set_get_alpha_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_alpha_ranges();
}

csm_c_char_range_set char_set_get_digit_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_digit_ranges();
}

csm_c_char_range_set char_set_get_xdigit_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_xdigit_ranges();
}

csm_c_char_range_set char_set_get_space_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_space_ranges();
}

csm_c_char_range_set char_set_get_punct_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_punct_ranges();
}

csm_c_char_range_set char_set_get_print_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_print_ranges();
}

csm_c_char_range_set char_set_get_graph_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_graph_ranges();
}

csm_c_char_range_set char_set_get_cntrl_ranges(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->get_cntrl_ranges();
}

void                 char_set_get_terminator(csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_terminator(c);
}

unsigned long        char_set_zero_term(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->zero_term();
}

unsigned long        char_set_hibit_is_clear(csm_c_char_set cs_p)
{
    return ((char_set*) cs_p)->hibit_is_clear();
}

void                 char_set_set_type(csm_c_char_set cs_p, tml_c_cs_type new_cs_type)
{
    ((char_set*) cs_p)->set_type((tml_type*) new_cs_type);
}

void                 char_set_set_name(csm_c_char_set cs_p, unsigned char* new_name)
{
    ((char_set*) cs_p)->set_name(new_name);
}

void                 char_set_set_xreg(csm_c_char_set cs_p, csm_c_xregister new_xreg)
{
    ((char_set*) cs_p)->set_xreg((xregister*) new_xreg);
}

void                 char_set_set_char_size(csm_c_char_set cs_p, unsigned long new_char_size)
{
    ((char_set*) cs_p)->set_char_size(new_char_size);
}

void                 char_set_set_num_chars(csm_c_char_set cs_p, unsigned long new_num_chars)
{
    ((char_set*) cs_p)->set_num_chars(new_num_chars);
}

void                 char_set_set_ref_count(csm_c_char_set cs_p, unsigned long new_ref_count)
{
    ((char_set*) cs_p)->set_ref_count(new_ref_count);
}

void                 char_set_incr_ref_count(csm_c_char_set cs_p)
{
    ((char_set*) cs_p)->incr_ref_count();
}

void                 char_set_decr_ref_count(csm_c_char_set cs_p)
{
    ((char_set*) cs_p)->decr_ref_count();
}

void                 char_set_set_predom_direc(csm_c_char_set cs_p, csm_c_char_set_dir new_direc)
{
    ((char_set*) cs_p)->set_predom_direc((char_set_dir) new_direc);
}

void                 char_set_set_char_direc(csm_c_char_set cs_p, long new_direc)
{
    ((char_set*) cs_p)->set_char_direc((lrud_dir) new_direc);
}

void                 char_set_set_next(csm_c_char_set cs_p, csm_c_char_set new_next)
{
    ((char_set*) cs_p)->set_next((char_set*) new_next);
}

void                 char_set_set_map(csm_c_char_set cs_p, csm_c_collation_map new_map)
{
    ((char_set*) cs_p)->set_map((collation_map*) new_map);
}

void                 char_set_set_null(csm_c_char_set cs_p, unsigned char* new_null)
{
    ((char_set*) cs_p)->set_null(new_null);
}

void                 char_set_init_esc_chars(csm_c_char_set cs_p, unsigned char** esc_chars, long spec_size)
{
    ((char_set*) cs_p)->init_esc_chars(esc_chars, spec_size);
}

csm_c_char_set       char_set_prepend(csm_c_char_set cs_p, csm_c_char_set new_sets)
{
    return ((char_set*) cs_p)->prepend((char_set*) new_sets);
}

csm_c_char_set       char_set_append(csm_c_char_set cs_p, csm_c_char_set new_sets)
{
    return ((char_set*) cs_p)->append((char_set*) new_sets);
}

csm_c_char_set       char_set_detach(csm_c_char_set cs_p, csm_c_char_set from_vec)
{
    return ((char_set*) cs_p)->detach((char_set*) from_vec);
}

long                 char_set_is_null(csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_null(c);
}

long                 char_set_both_null(csm_c_char_set cs_p, unsigned char* c1, unsigned char* c2)
{
    return ((char_set*) cs_p)->both_null(c1, c2);
}

long                 char_set_char_arr_cmp(csm_c_char_set cs_p, unsigned char* s1, unsigned char* s2)
{
    return ((char_set*) cs_p)->char_arr_cmp(s1, s2);
}

long                 char_set_char_arr_ncmp(csm_c_char_set cs_p, unsigned char* s1, unsigned char* s2, unsigned long n)
{
    return ((char_set*) cs_p)->char_arr_ncmp(s1, s2, n);
}

void                 char_set_ignored_prefix_sort(csm_c_char_set cs_p, char** strings, unsigned long num_strings, char** prefixes, unsigned long num_prefixes)
{
    ((char_set*) cs_p)->ignored_prefix_sort(strings, num_strings, prefixes, num_prefixes);
}

unsigned long  csm_c_char_set_as_ulong(csm_c_char_set cs_p, unsigned char* s)
{
    return ((char_set*) cs_p)->as_ulong(s);
}

void                 char_set_advance(csm_c_char_set cs_p, unsigned char** c)
{
    ((char_set*) cs_p)->advance(*c);
}

void                 char_set_retreat(csm_c_char_set cs_p, unsigned char** c)
{
    ((char_set*) cs_p)->retreat(*c);
}

long                 char_set_is_char(csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_char(c);
}

long                 char_set_cs_isalpha (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_isalpha (c);
}

long                 char_set_cs_isdigit (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_isdigit (c);
}

long                 char_set_cs_isxdigit(csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_isxdigit(c);
}

long                 char_set_cs_isalnum (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_isalnum (c);
}

long                 char_set_cs_isspace (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_isspace (c);
}

long                 char_set_cs_ispunct (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_ispunct (c);
}

long                 char_set_cs_isprint (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_isprint (c);
}

long                 char_set_cs_isgraph (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_isgraph (c);
}

long                 char_set_cs_iscntrl (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->cs_iscntrl (c);
}

long                 char_set_is_alert    (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_alert    (c);
}

long                 char_set_is_backspace(csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_backspace(c);
}

long                 char_set_is_formfeed (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_formfeed (c);
}

long                 char_set_is_newline  (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_newline  (c);
}

long                 char_set_is_cr       (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_cr       (c);
}

long                 char_set_is_htab     (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_htab     (c);
}

long                 char_set_is_vtab     (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_vtab     (c);
}

long                 char_set_is_backslash(csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_backslash(c);
}

long                 char_set_is_question (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_question (c);
}

long                 char_set_is_squote   (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_squote   (c);
}

long                 char_set_is_dquote   (csm_c_char_set cs_p, unsigned char* c)
{
    return ((char_set*) cs_p)->is_dquote   (c);
}

void                 char_set_get_alert    (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_alert    (c);
}

void                 char_set_get_backspace(csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_backspace(c);
}

void                 char_set_get_formfeed (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_formfeed (c);
}

void                 char_set_get_newline  (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_newline  (c);
}

void                 char_set_get_cr       (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_cr       (c);
}

void                 char_set_get_htab     (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_htab     (c);
}

void                 char_set_get_vtab     (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_vtab     (c);
}

void                 char_set_get_backslash(csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_backslash(c);
}

void                 char_set_get_question (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_question (c);
}

void                 char_set_get_squote   (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_squote   (c);
}

void                 char_set_get_dquote   (csm_c_char_set cs_p, unsigned char* c)
{
    ((char_set*) cs_p)->get_dquote   (c);
}

csm_c_csm_character char_set_get_regexp_chars(csm_c_char_set cs_p)
{
    return NULL;
}

void                char_set_set_regexp_chars(csm_c_char_set cs_p, csm_c_csm_character new_re_chars)
{
}


/* FOR CLASS ALPHA_CHAR_SET */

csm_c_alpha_char_set alpha_char_set_class_alloc(unsigned long per_char_size)
{
    return new alpha_char_set(per_char_size);
}

void                 alpha_char_set_class_free(csm_c_alpha_char_set acs_p)
{
    delete (alpha_char_set*) acs_p;
}

long           alpha_char_set_cs_isupper(csm_c_alpha_char_set acs_p, unsigned char* c)
{
    return ((alpha_char_set*) acs_p)->cs_isupper(c);
}

long           alpha_char_set_cs_islower(csm_c_alpha_char_set acs_p, unsigned char* c)
{
    return ((alpha_char_set*) acs_p)->cs_islower(c);
}

unsigned char* alpha_char_set_cs_tolower(csm_c_alpha_char_set acs_p, unsigned char* c)
{
    return ((alpha_char_set*) acs_p)->cs_tolower(c);
}

unsigned char* alpha_char_set_cs_toupper(csm_c_alpha_char_set acs_p, unsigned char* c)
{
    return ((alpha_char_set*) acs_p)->cs_toupper(c);
}

void           alpha_char_set_become_iso(csm_c_alpha_char_set acs_p)
{
    ((alpha_char_set*) acs_p)->become_iso_latin_1();
}

void           alpha_char_set_become_ascii(csm_c_alpha_char_set acs_p)
{
    ((alpha_char_set*) acs_p)->become_ascii();
}


/* FOR CLASS SYLL_FRAG_CHAR_SET */

csm_c_syll_frag_char_set syll_frag_char_set_class_alloc(unsigned long per_char_size)
{
    return new syll_frag_char_set(per_char_size);
}

void                     syll_frag_char_set_free(csm_c_syll_frag_char_set sfcs_p)
{
    delete (syll_frag_char_set*) sfcs_p;
}

unsigned long            syll_frag_char_set_is_syll(csm_c_syll_frag_char_set sfcs_p, unsigned char* start)
{
    return ((syll_frag_char_set*) sfcs_p)->is_syll(start);
}

unsigned long            syll_frag_char_set_get_syll(csm_c_syll_frag_char_set sfcs_p, unsigned char** start)
{
    return ((syll_frag_char_set*) sfcs_p)->get_syll(*start);
}

unsigned long            syll_frag_char_set_get_syll_len(csm_c_syll_frag_char_set sfcs_p, unsigned char* start)
{
    return ((syll_frag_char_set*) sfcs_p)->get_syll_len(start);
}


/* FOR CLASS UNION_CHAR_SET */

csm_c_union_char_set union_char_set_class_alloc_1()
{
    return new union_char_set();
}

csm_c_union_char_set union_char_set_class_alloc_2(csm_c_char_set new_char_sets)
{
    return new union_char_set((char_set*) new_char_sets);
}

void                 union_char_set_free(csm_c_union_char_set ucs_p)
{
    delete (union_char_set*) ucs_p;
}

csm_c_char_set       union_char_set_get_sets(csm_c_union_char_set ucs_p)
{
    return ((union_char_set*) ucs_p)->get_sets();
}

unsigned long        union_char_set_get_num_sets(csm_c_union_char_set ucs_p)
{
    return ((union_char_set*) ucs_p)->get_num_sets();
}

csm_c_union_char_set union_char_set_get_next(csm_c_union_char_set ucs_p)
{
    return ((union_char_set*) ucs_p)->get_next();
}

unsigned long        union_char_set_get_ref_count(csm_c_union_char_set ucs_p)
{
    return ((union_char_set*) ucs_p)->get_ref_count();
}

void                 union_char_set_set_sets(csm_c_union_char_set ucs_p, csm_c_char_set new_sets)
{
           ((union_char_set*) ucs_p)->set_sets((char_set*) new_sets);
}

void                 union_char_set_set_num_sets(csm_c_union_char_set ucs_p, unsigned long new_num)
{
           ((union_char_set*) ucs_p)->set_num_sets(new_num);
}

void                 union_char_set_set_next(csm_c_union_char_set ucs_p, csm_c_union_char_set new_next)
{
           ((union_char_set*) ucs_p)->set_next((union_char_set*) new_next);
}

void                 union_char_set_set_ref_count(csm_c_union_char_set ucs_p, unsigned long new_ref_count)
{
           ((union_char_set*) ucs_p)->set_ref_count(new_ref_count);
}

void                 union_char_set_incr_ref_count(csm_c_union_char_set ucs_p)
{
           ((union_char_set*) ucs_p)->incr_ref_count();
}

void                 union_char_set_decr_ref_count(csm_c_union_char_set ucs_p)
{
           ((union_char_set*) ucs_p)->decr_ref_count();
}

csm_c_union_char_set union_char_set_prepend(csm_c_union_char_set ucs_p, csm_c_union_char_set new_unions)
{
    return ((union_char_set*) ucs_p)->prepend((union_char_set*) new_unions);
}

csm_c_union_char_set union_char_set_append(csm_c_union_char_set ucs_p, csm_c_union_char_set new_unions)
{
    return ((union_char_set*) ucs_p)->append((union_char_set*) new_unions);
}

csm_c_union_char_set union_char_set_detach(csm_c_union_char_set ucs_p, csm_c_union_char_set from_vec)
{
    return ((union_char_set*) ucs_p)->detach((union_char_set*) from_vec);
}

csm_c_char_set       union_char_set_prepend_char_set(csm_c_union_char_set ucs_p, csm_c_char_set new_sets)
{
    return ((union_char_set*) ucs_p)->prepend((char_set*) new_sets);
}

csm_c_char_set       union_char_set_append_char_set(csm_c_union_char_set ucs_p, csm_c_char_set new_sets)
{
    return ((union_char_set*) ucs_p)->append((char_set*) new_sets);
}

csm_c_char_set       union_char_set_detach_char_set(csm_c_union_char_set ucs_p, csm_c_char_set old_set)
{
    return ((union_char_set*) ucs_p)->detach((char_set*) old_set);
}

csm_c_char_set       union_char_set_contains_any(csm_c_union_char_set ucs_p, unsigned char* c, unsigned long char_size)
{
    return ((union_char_set*) ucs_p)->contains(c, char_size);
}

long                 union_char_set_contains_ask(csm_c_union_char_set ucs_p, unsigned long char_set_index, unsigned char* c)
{
    return ((union_char_set*) ucs_p)->contains(char_set_index, c);
}

long                 union_char_set_default_only(csm_c_union_char_set ucs_p)
{
    return ((union_char_set*) ucs_p)->default_only();
}

long                 union_char_set_write_ucs_config_file(csm_c_union_char_set ucs_p, void *file_handle)
{
    return ((union_char_set*) ucs_p)->write_ucs_config_file(file_handle);
}

void                 union_char_set_prepare_ucs_binary(csm_c_union_char_set ucs_p, tml_c_gc_hash_tbl** hash_tbl, csm_c_cs_heap** heaps, tml_c_tml_ext_type** types)
{
    ((union_char_set*) ucs_p)->prepare_ucs_binary(*((gc_hash_tbl**) hash_tbl), *((cs_heap**) heaps), *((tml_ext_type**) types));
}

long                 union_char_set_write_ucs_binary_file(csm_c_union_char_set ucs_p, void *file_handle, tml_c_gc_hash_tbl** hash_tbl, csm_c_cs_heap** heaps, tml_c_tml_ext_type** types)
{
    return ((union_char_set*) ucs_p)->write_ucs_binary_file(file_handle, *((gc_hash_tbl**) hash_tbl), *((cs_heap**) heaps), *((tml_ext_type**) types));
}


/* FOR CLASS CS_HEAP */

csm_c_cs_heap cs_heap_class_alloc(tml_c_endian my_endian)
{
    return new cs_heap((endian*) my_endian);
}

void          cs_heap_free(csm_c_cs_heap cs_heap_p)
{
    delete ((cs_heap*) cs_heap_p);
}


/* FOR CLASS KMP_NODE */

void           csm_c_kmp_node_free(csm_c_kmp_node kmp_node_p)
{
           delete (kmp_node*) kmp_node_p;
}

unsigned long  csm_c_kmp_node_lookup(csm_c_kmp_node kmp_node_p, unsigned char** sought)
{
    return ((kmp_node*) kmp_node_p)->lookup(*sought);
}

unsigned long  kmp_node_incr_lookup(csm_c_kmp_node kmp_node_p, csm_c_kmp_node** next_node, unsigned long* num_chars, unsigned char* s)
{
    return ((kmp_node*) kmp_node_p)->incr_lookup((kmp_node*) *next_node, *num_chars, s);
}


/* GENERAL FUNCTIONS */

csm_c_union_char_set csm_c_read_ucs_binary_file(void *file_handle, tml_c_gc_hash_tbl** hash_tbl, tml_c_tml_ext_type** types, tml_c_endian** their_endian)
{
    return read_ucs_binary_file(file_handle, *((gc_hash_tbl**) hash_tbl), *((tml_ext_type**) types), *((endian**) their_endian));
}

csm_c_union_char_set csm_c_read_ucs_config_file(void *file_handle)
{
    return read_ucs_config_file(file_handle);
}

long csm_c_char_arr_cmp(csm_c_csm_character* c1, csm_c_csm_character* c2)
{        
    return     char_arr_cmp((csm_character*) c1, (csm_character*) c2);
}

long csm_c_char_arr_ncmp(csm_c_csm_character* c1, csm_c_csm_character* c2, unsigned long n)
{
    return     char_arr_ncmp((csm_character*) c1, (csm_character*) c2, n);
}

void           csm_c_set_csm_char_set_for_qsort(csm_c_char_set new_char_set)
{
    set_csm_char_set_for_qsort((char_set*) new_char_set);
}

int            csm_c_csm_cmp_for_qsort(char** sp1, char** sp2)
{
    return csm_cmp_for_qsort(sp1, sp2);
}

csm_c_kmp_node csm_c_kmp_fsm(unsigned char** strings, unsigned long num_strings, unsigned long* collations, csm_c_char_set for_char_set)
{
    return kmp_fsm(strings, num_strings, collations, (char_set*) for_char_set);
}

long csm_c_is_char_class_name(char* name)
{
    return NULL;
}

tml_c_endian csm_c_get_their_endian(unsigned long their_magic_number)
{
    return get_their_endian(their_magic_number);
}

tml_c_endian csm_c_get_my_endian()
{
    return get_my_endian();
}

#ifdef __cplusplus
}
#endif

