/**************************************************************************************
    File csm_inlib_entry.cxx
    Initializer entry point for inlib libraries

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <tml/csm_cxx2.0_linkage_decls.hxx>
#include <tml/char_set_mgr.hxx>
#include <tml/tml_os_depends.hxx>

/* Initialize global (non-per-process) objects for inlib'd and global */
/*   libraries (Apollo only). */
/* To cause csm_inlib_entry() to be called when the library is loaded by inlib */
/*   build the library with the "-entry tml_inlib_entry" option to /com/bind. */
/*   (The -entry option must appear AFTER the object module that contains the */
/*   routine it designates as the alternate entry point.) */
/* This routine is NOT called when a library built in that manner is bound into */
/*   a program. */
/* DO NOT specifiy the -entry option when building global libraries. */
/* The work of this module is parallel to the work of char_set_extend.cxx */
/*   in bound-in libraries; variables declared extern here are defined there. */

/* Declare pointers to new global objects */

syll_frag_char_set  *tml_sample_syll_frag_char_set_p;
alpha_char_set      *tml_sample_alpha_char_set_p;
char_set            *tml_sample_char_set_p;

extern tml_ext_type tml_syll_frag_char_set_ext_t;
extern tml_ext_type tml_alpha_char_set_ext_t;
extern tml_ext_type tml_char_set_ext_t;

extern tml_type tml_syll_frag_char_set_t;
extern tml_type tml_alpha_char_set_t;
extern tml_type tml_char_set_t;

extern tml_type* const tml_syll_frag_char_set;
extern tml_type* const tml_alpha_char_set;
extern tml_type* const tml_char_set;

extern type_queue type_q;

#ifdef apollo_global
long csm_initialized;
#endif

void csm_inlib_entry()
{
#ifdef apollo_global
    if (csm_initialized)
        return;
    csm_initialized = -1;
#endif
    void init_csm_ext_types();
    void init_csm_types();
    void init_csm_type_pointers();

    tml_sample_syll_frag_char_set_p = new syll_frag_char_set;
    tml_sample_alpha_char_set_p     = new alpha_char_set;
    tml_sample_char_set_p           = new char_set;

    init_csm_ext_types();
    init_csm_types();
    init_csm_type_pointers();

    type_q.head = (tml_type*) &tml_char_set_t;
    type_q.tail = (tml_type*) &tml_syll_frag_char_set_t;
}

void init_csm_ext_types()
{
    tml_syll_frag_char_set_ext_t.time_high    = 0x41192edb;
    tml_syll_frag_char_set_ext_t.time_low     = 0x4000;
    tml_syll_frag_char_set_ext_t.reserved     = 0x0000;
    tml_syll_frag_char_set_ext_t.family       = 0x0d;
    tml_syll_frag_char_set_ext_t.host.host[0] = 0x00;
    tml_syll_frag_char_set_ext_t.host.host[1] = 0x01;
    tml_syll_frag_char_set_ext_t.host.host[2] = 0x9f;
    tml_syll_frag_char_set_ext_t.host.host[3] = 0xab;
    tml_syll_frag_char_set_ext_t.host.host[4] = 0x00;
    tml_syll_frag_char_set_ext_t.host.host[5] = 0x00;
    tml_syll_frag_char_set_ext_t.host.host[6] = 0x00;
    
    tml_alpha_char_set_ext_t.time_high    = 0x41192e22;
    tml_alpha_char_set_ext_t.time_low     = 0x0000;
    tml_alpha_char_set_ext_t.reserved     = 0x0000;
    tml_alpha_char_set_ext_t.family       = 0x0d;
    tml_alpha_char_set_ext_t.host.host[0] = 0x00;
    tml_alpha_char_set_ext_t.host.host[1] = 0x01;
    tml_alpha_char_set_ext_t.host.host[2] = 0x9f;
    tml_alpha_char_set_ext_t.host.host[3] = 0xab;
    tml_alpha_char_set_ext_t.host.host[4] = 0x00;
    tml_alpha_char_set_ext_t.host.host[5] = 0x00;
    tml_alpha_char_set_ext_t.host.host[6] = 0x00;
    
    tml_char_set_ext_t.time_high    = 0x41192d9d;
    tml_char_set_ext_t.time_low     = 0xc000;
    tml_char_set_ext_t.reserved     = 0x0000;
    tml_char_set_ext_t.family       = 0x0d;
    tml_char_set_ext_t.host.host[0] = 0x00;
    tml_char_set_ext_t.host.host[1] = 0x01;
    tml_char_set_ext_t.host.host[2] = 0x9f;
    tml_char_set_ext_t.host.host[3] = 0xab;
    tml_char_set_ext_t.host.host[4] = 0x00;
    tml_char_set_ext_t.host.host[5] = 0x00;
    tml_char_set_ext_t.host.host[6] = 0x00;
}
    
void init_csm_types()
{
    tml_syll_frag_char_set_t.type      = &tml_syll_frag_char_set_ext_t;
    tml_syll_frag_char_set_t.instance  = tml_sample_syll_frag_char_set_p;
    tml_syll_frag_char_set_t.next      = 0;
    
    tml_alpha_char_set_t.type          = &tml_alpha_char_set_ext_t;
    tml_alpha_char_set_t.instance      = tml_sample_alpha_char_set_p;
    tml_alpha_char_set_t.next          = &tml_syll_frag_char_set_t;
    
    tml_char_set_t.type                = &tml_char_set_ext_t;
    tml_char_set_t.instance            = tml_sample_char_set_p;
    tml_char_set_t.next                = &tml_alpha_char_set_t;
}

void init_csm_type_pointers()
{
    tml_type** ptr_2_tml_syll_frag_char_set;
    tml_type** ptr_2_tml_alpha_char_set;
    tml_type** ptr_2_tml_char_set;

    ptr_2_tml_syll_frag_char_set     = (tml_type**)&tml_syll_frag_char_set;
    *ptr_2_tml_syll_frag_char_set    = &tml_syll_frag_char_set_t;
    ptr_2_tml_alpha_char_set         = (tml_type**)&tml_alpha_char_set;
    *ptr_2_tml_alpha_char_set        = &tml_alpha_char_set_t;
    ptr_2_tml_char_set               = (tml_type**)&tml_char_set;
    *ptr_2_tml_char_set              = &tml_char_set_t;
}

