/**************************************************************************************
    draw_mark_traits.cxx
    functions for the class containing the list of marks associated with a
    particular viewport in the text management library rendering component

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_dp_graph.hxx>

/*draw routine--this will at some point become the default routine, which
    users can change. For now, they will have to stick with this one*/
long def_draw_mark_rtn(view_port *in_viewport, core_m_draw core_func, model_pos start_draw, model_pos end_draw, long erase = 0)
{
    screen_xy       start_xy,            //start coordinates of the draw
                    end_xy,              //end coordinates of the draw
                    first_line_xy;       //start coordinates of line containing start_draw
    model_pos       first_line_mpos,     //model position of head of line containing start_draw
                    line_end_mpos;       //model position of last character on each line of multiple lines
    view_pos       *start_vpos,          //start view_pos
                   *end_vpos;            //end view_pos
    unsigned char*  line_end_char;       //points to last character on each line of multiple lines
    long            last_char_width = 0, //width of last character on last line
                    line_end_width;      //width of last character on each line of multiple lines
    unsigned long   line_end_char_val;   //numeric value of last character on each line of multiple lines
    cline          *last_cline;
    char_set       *in_char_set;
    char           *last_char;
    long            last_char_val;
    long            src_ulx, src_uly,   // coordinates for the blit, corrected by restricting them to 
                    src_lrx, src_lry;   //   viewport real estate to prevent possible illegal calls or
                                        //   infinite loops in some graphics systems

    start_vpos = in_viewport->mpos_2_vpos(start_draw);
    if (!start_vpos)
        return 0;
    start_xy = in_viewport->mpos_2_xy(start_draw, start_vpos);

    end_vpos = in_viewport->mpos_2_vpos(end_draw);
    if (!end_vpos)
        return 0;
    end_xy = in_viewport->mpos_2_xy(end_draw, end_vpos);
    
    /*if the last model pos refers to a character, get the width of that character*/
    last_cline = end_vpos->get_cline()->get_phead()->mpos_2_cline(end_draw);
    if (last_cline && (last_cline->get_type() == cl_text)){
        last_char = (char *)(end_draw.in_node->get_contents());
        last_char += end_draw.char_index;
        in_char_set = (char_set *)end_draw.in_node->get_model()->inq_state(end_draw.in_node, tml_char_set_prop);
        last_char_val = in_char_set->as_ulong((unsigned char*) last_char);
        last_char_width = ((tline *)last_cline)->get_font()->get_width((long)(last_char_val));
        }
                                                                                                 
    /*width of newline if any is in last_char_width*/
    if (start_vpos == end_vpos){
        /* correct for start_x_off to allow marking of tabbed-over space */
        if (0 > (src_ulx = start_xy.at_x - start_vpos->get_cline()->get_start_x_off()))
            src_ulx = 0;
        if (0 > (src_uly = start_vpos->get_y()))
            src_uly = 0;
        if ((in_viewport->get_width() - 1) < (src_lrx = (end_xy.at_x + last_char_width)))
            src_lrx = in_viewport->get_width() - 1;
        if ((in_viewport->get_hi() - 1) < (src_lry = (end_vpos->get_y() + end_vpos->get_cline()->get_hi())))
            src_lry = in_viewport->get_hi() - 1;
        (*core_func)(in_viewport->get_handle(), src_ulx, src_uly, src_lrx, src_lry, src_ulx, src_uly, TML_DP_INV);
        return 1;
        }
                                                                
    /*the highlight is over more than one vpos--so begin by marking the first one*/
    first_line_mpos = start_vpos->get_cline()->get_start();
    first_line_xy = in_viewport->mpos_2_xy(first_line_mpos, start_vpos);

    /* text_width of tline doesn't include the newline - get width of newline if any in line_end_width */
    in_char_set = (char_set*) first_line_mpos.in_node->get_model()->inq_state(first_line_mpos.in_node, tml_char_set_prop);
    line_end_mpos = start_vpos->get_cline()->get_end();
    line_end_width = 0;
    line_end_char = ((unsigned char*) line_end_mpos.in_node->get_contents()) + line_end_mpos.char_index;
    if (in_char_set->is_newline(line_end_char)) {
        line_end_char_val = in_char_set->as_ulong(line_end_char);
        line_end_width = ((tline *) start_vpos->get_cline())->get_font()->get_width((long)(line_end_char_val));
        }

    /* correct for start_x_off to allow marking of tabbed-over space */
    if (0 > (src_ulx = start_xy.at_x - start_vpos->get_cline()->get_start_x_off()))
        src_ulx = 0;
    if (0 > (src_uly = start_vpos->get_y()))
        src_uly = 0;
    if ((in_viewport->get_width() - 1) 
      < (src_lrx = (first_line_xy.at_x + start_vpos->get_cline()->get_contents_width() + line_end_width)))
        src_lrx = in_viewport->get_width() - 1;
    if ((in_viewport->get_hi() - 1) < (src_lry = (start_vpos->get_y() + start_vpos->get_cline()->get_hi())))
        src_lry = in_viewport->get_hi() - 1;
    (*core_func)(in_viewport->get_handle(), src_ulx, src_uly, src_lrx, src_lry, src_ulx, src_uly, TML_DP_INV);
                                                                                                                    
    /*now mark all of the non-blank and non-free lines in between*/
        for (start_vpos = start_vpos->get_next();
          start_vpos && (start_vpos != end_vpos);
          start_vpos = start_vpos->get_next()) {
            if ((start_vpos->get_cline()->get_type() != cl_blank) && (start_vpos->get_cline()->get_type() != cl_free)) {
                /* text_width of tline doesn't include the newline - get width of newline if any in line_end_width */
                in_char_set = (char_set*) start_vpos->get_cline()->get_start().in_node->get_model()->inq_state(first_line_mpos.in_node, tml_char_set_prop);
                line_end_mpos = start_vpos->get_cline()->get_end();
                line_end_width = 0;
                line_end_char = ((unsigned char*) line_end_mpos.in_node->get_contents()) + line_end_mpos.char_index;
                if (in_char_set->is_newline(line_end_char)) {
                    line_end_char_val = in_char_set->as_ulong(line_end_char);
                    line_end_width = ((tline *) start_vpos->get_cline())->get_font()->get_width((long)(line_end_char_val));
                    }
                /* correct for start_x_off to allow marking of tabbed-over space */
                if (0 > (src_ulx = start_vpos->get_x() - start_vpos->get_cline()->get_start_x_off()))
                    src_ulx = 0;
                if (0 > (src_uly = start_vpos->get_y()))
                    src_uly = 0;
                // correct for viewport h_offset on these
                if ((in_viewport->get_width() - 1) 
                  < (src_lrx = (start_vpos->get_x() + start_vpos->get_cline()->get_contents_width() + line_end_width + in_viewport->get_hoffset())))
                    src_lrx = in_viewport->get_width() - 1;
                if ((in_viewport->get_hi() - 1) < (src_lry = (start_vpos->get_y() + start_vpos->get_cline()->get_hi())))
                    src_lry = in_viewport->get_hi() - 1;
                (*core_func)(in_viewport->get_handle(), src_ulx, src_uly, src_lrx, src_lry, src_ulx, src_uly, TML_DP_INV);
                }
            }
            
    /*now mark the last line*/
    /*width of newline if any is in last_char_width*/
    /* correct for start_x_off to allow marking of tabbed-over space */
    if (0 > (src_ulx = end_vpos->get_x() - end_vpos->get_cline()->get_start_x_off()))
        src_ulx = 0;
    if (0 > (src_uly = end_vpos->get_y()))
        src_uly = 0;
    if ((in_viewport->get_width() - 1) < (src_lrx = (end_xy.at_x + last_char_width)))
        src_lrx = in_viewport->get_width() - 1;
    if ((in_viewport->get_hi() - 1) < (src_lry = (end_vpos->get_y() + end_vpos->get_cline()->get_hi())))
        src_lry = in_viewport->get_hi() - 1;
    (*core_func)(in_viewport->get_handle(), src_ulx, src_uly, src_lrx, src_lry, src_ulx, src_uly, TML_DP_INV);
    return 1;

}
    

/*constructor for a draw_mark object*/
draw_mark::draw_mark()
{
    next = prior = 0;
    model_mark = 0;
    is_visible = 0;
    is_enabled = 0;
    in_viewport = 0;
    core_func = 0;
    draw_func = &def_draw_mark_rtn;
}

/* destructor for a draw_mark object */
draw_mark::~draw_mark()
{
    next = prior = 0;
    model_mark = 0;
    in_viewport = 0;
}

/*extend the mark*/
long draw_mark::extend(model_pos new_start, model_pos new_end)
{
    /*begin by checking for trash*/
    if (!this || !new_start.in_node || !new_end.in_node)
        return 0;

    /*extend the mark using the mark_list::extend() call--this
        will draw the mark if it is visible in all viewports*/
    if (!model_mark->extend_mark(new_start, new_end))
        return 0;

    /*if it wasn't redrawn by mark_list::extend, redraw it*/
    if (!is_visible)
        is_visible = draw();

    return 1;
}
                                                       
void draw_mark::set_draw_func(draw_m_draw new_draw)
{
    if (new_draw)
        draw_func = new_draw;
    else
        draw_func = &def_draw_mark_rtn;
}


void draw_mark::set_core_func(core_m_draw new_core)
{
    if (new_core)
        core_func = new_core;
    else
        core_func = &tml_dp_blt;
}


/*draw a mark--returns 1 if the draw actually leaves something on the screen, otherwise
     returns 0*/
long draw_mark::draw()
{
    model_pos     start_draw,
                        end_draw;
    view_pos       *start_vpos,
                        *end_vpos;

    /*check for trash*/
    if (!this || !in_viewport->get_format() || !model_mark ||
        !model_mark->get_range()->get_start().in_node || !model_mark->get_range()->get_end().in_node)
        return 0;
 
    /*if it is already visible, or it is disabled,  just return*/
    if (is_visible || (is_enabled < 1))
        return 0;

    /*get the order of the marks right*/
    start_draw = model_mark->get_range()->get_start();
    end_draw = model_mark->get_range()->get_end();
    if (end_draw < start_draw){
        start_draw = model_mark->get_range()->get_end();
        end_draw = model_mark->get_range()->get_start();
        }

    if ((start_draw.in_node->get_model() != end_draw.in_node->get_model()) ||
         (start_draw.in_node->get_model() != in_viewport->get_format()->get_model()))
        return 0;

    /*see if both are on the screen*/
    start_vpos = in_viewport->mpos_2_vpos(start_draw);
    end_vpos = in_viewport->mpos_2_vpos(end_draw);
                                                                               
    /*if both are on the screen, just draw and get out*/
    if (start_vpos && end_vpos){
        is_visible = (*draw_func)(in_viewport, core_func, start_draw, end_draw);
        return is_visible;
        }

    if (!start_vpos){
        for (start_vpos = in_viewport->get_top_line(); start_vpos; start_vpos = start_vpos->get_next())
            if ((start_vpos->get_cline()->get_type() != cl_free) && (start_vpos->get_cline()->get_type() != cl_blank))
                break;
        if (!start_vpos || (start_vpos->get_cline()->get_start() < start_draw))
            return 0;
        else
            start_draw = start_vpos->get_cline()->get_start();
        }

    if (!end_vpos){
        for (end_vpos = in_viewport->get_bottom_line(); end_vpos; end_vpos = end_vpos->get_prior())
            if ((end_vpos->get_cline()->get_type() != cl_free) && (end_vpos->get_cline()->get_type() != cl_blank))
                break;
        if (end_vpos->get_cline()->get_end() > end_draw)
            return 0;
        else
            end_draw = end_vpos->get_cline()->get_end();
        }

    is_visible = (*draw_func)(in_viewport, core_func, start_draw, end_draw);
    return is_visible;
}

/*erase a mark--returns 1 if the erase actually does something on the screen, otherwise
     returns 0*/
long draw_mark::erase()
{
    model_pos     start_draw,
                        end_draw;
    view_pos       *start_vpos,
                        *end_vpos;
    long                is_erased;

    /*check for trash*/
    if (!this)
        return 0;
 
    /*if it is already invisible, or it isn't enabled, just return*/
    if (!is_visible || (is_enabled < 1))
        return 0;

    /*get the order of the marks right*/
    start_draw = model_mark->get_range()->get_start();
    end_draw = model_mark->get_range()->get_end();
    if (end_draw < start_draw){
        start_draw = model_mark->get_range()->get_end();
        end_draw = model_mark->get_range()->get_start();
        }

    /*see if both are on the screen*/
    start_vpos = in_viewport->mpos_2_vpos(start_draw);
    end_vpos = in_viewport->mpos_2_vpos(end_draw);
                                                                               
    /*if both are on the screen, just draw and get out*/
    if (start_vpos && end_vpos){
        is_erased = (*draw_func)(in_viewport, core_func, start_draw, end_draw, 1);
        is_visible = is_erased ? 0 : 1;
        return is_erased;
        }

    if (!start_vpos){
        for (start_vpos = in_viewport->get_top_line(); start_vpos; start_vpos = start_vpos->get_next())
            if ((start_vpos->get_cline()->get_type() != cl_free) && (start_vpos->get_cline()->get_type() != cl_blank))
                break;
        if (!start_vpos || (start_vpos->get_cline()->get_start() < start_draw))
            return 0;
        else
            start_draw = start_vpos->get_cline()->get_start();
        }

    if (!end_vpos){
        for (end_vpos = in_viewport->get_bottom_line(); end_vpos; end_vpos = end_vpos->get_prior())
            if ((end_vpos->get_cline()->get_type() != cl_free) && (end_vpos->get_cline()->get_type() != cl_blank))
                break;
        if (end_vpos->get_cline()->get_end() > end_draw)
            return 0;
        else
            end_draw = end_vpos->get_cline()->get_end();
        }

    is_erased = (*draw_func)(in_viewport, core_func, start_draw, end_draw, 1);
    is_visible = is_erased ? 0 : 1;
    return is_erased;
}
