/**************************************************************************************
    endian.hxx
    header file for the object definitions for the architecture-neutral
      components of the text management library.

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef endian_hxx
#define endian_hxx

#ifndef arch_neutral_hxx
#include <tml/arch_neutral.hxx>
#endif

/**********************************************************************************
    constants and classes used to handle magic numbers and TML files
      written by heterogeneous architectures
**********************************************************************************/

// Magic numbers for 32-bit architectures
// TML libraries consider their host machine TML_MAGIC_3210

#define TML_MAGIC_0123  0x0A0F080E      /* low order  byte on left;  low order  word on left */
#define TML_MAGIC_1032  0x0F0A0E08      /* high order byte on left;  low order  word on left */
#define TML_MAGIC_2301  0x080E0A0F      /* low order  byte on left;  high order word on left */
#define TML_MAGIC_3210  0x0E080F0A      /* high order byte on left;  high order word on left */

// base class for converters of architecture-specific representations

class endian {
protected:
    unsigned long*   s_mask_arr;
    unsigned long*   s_shift_arr;
    unsigned long*   l_mask_arr;
    unsigned long*   l_shift_arr;
public:
    endian(){s_mask_arr = 0; s_shift_arr = 0; l_mask_arr = 0; l_shift_arr = 0;}

    ~endian(){delete s_mask_arr; delete s_shift_arr; delete l_mask_arr; delete l_shift_arr;}

    virtual long          is_3210(){return(0);}

    // functions "dollar" return a single 8-bit quantity
    //   -- bytes may vary, but there are always 8 bits in a dollar

    virtual unsigned char dollar_s(unsigned short their_u_short, long n)
                                {return((their_u_short & s_mask_arr[n]) >> s_shift_arr[n]);}
             
    virtual unsigned char dollar_l(unsigned long  their_u_long,  long n)
                                {return((their_u_long  & l_mask_arr[n]) >> l_shift_arr[n]);}

    virtual unsigned short my_short(unsigned short their_u_short)
                                   {return((dollar_s(their_u_short, 0)) |
                                           (dollar_s(their_u_short, 1) << 8));}

    virtual unsigned long  my_long (unsigned long  their_u_long)
                                   {return((dollar_l(their_u_long, 0)) |
                                           (dollar_l(their_u_long, 1) << 8) |
                                           (dollar_l(their_u_long, 2) << 16) |
                                           (dollar_l(their_u_long, 3) << 24));}

    virtual HALF_PTR       my_half_ptr(HALF_PTR their_half_ptr);

    // may be used ONLY to unscramble a long-aligned buffer of longs

    virtual unsigned long* long_arr_shuf(unsigned long* inbuf, long n);

    // may be used ONLY to unscramble a short-aligned buffer of shorts

    virtual unsigned short* short_arr_shuf(unsigned short* inbuf, long n);
};

// derived class for converters of architecture-specific representations

class endian_0123 : public endian {
public:
    endian_0123();
};

class endian_1032 : public endian {
public:
    endian_1032();
};

class endian_2301 : public endian {
public:
    endian_2301();
};

class endian_3210 : public endian {
public:
    endian_3210();
    long            is_3210(){return(-1);}
    unsigned short  my_short(unsigned short their_u_short);
    unsigned long   my_long (unsigned long their_u_long);
    HALF_PTR        my_half_ptr(HALF_PTR their_half_ptr);
    unsigned long*  long_arr_shuf(unsigned long* inbuf, long n);
    unsigned short* short_arr_shuf(unsigned short* inbuf, long n);
};

#endif
// endian_hxx

