/**************************************************************************************
    phead_traits.cxx
    member functions for the phead class

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <tml/tml_objects.hxx>

/*constructor for a phead*/
phead::phead(format *to_format)
{
    next = prior = 0;
    dx = dy = 0;
    in_format = to_format;
    page_num = -1;
    start_pos.in_node = end_pos.in_node = 0;
    start_pos.char_index = end_pos.char_index = 0;
    format_ok = 0;
    fline_no = 0;
    line_cnt = 0;
    max_line_width = 0;
    top_used = bottom_used = 0;
    top_line = bottom_line = top_free = bottom_free = new cline(in_format);
    top_line->set_phead(this);
    top_line->set_ul_coord(0,0);
    top_line->set_hi(dy);
    top_line->set_wide(dx);
}

/*destructor for a phead*/
phead::~phead()
{
    register phead *update_phead;
    register cline *this_line,
                   *next_line;

    if (next){
        for (update_phead = next; update_phead; update_phead = update_phead->next){
            update_phead->fline_no -= line_cnt;
            if ((in_format->inq_state(start_pos.in_node, tml_header_prop))
              || (in_format->inq_state(start_pos.in_node, tml_footer_prop)))
                update_phead->format_ok = 0;
            update_phead->page_num--;
            }
        next->prior = prior;
        }
    if (prior)
        prior->next = next;

    in_format->calc_max_line_width();
    in_format->calc_max_used_ht();
    for (this_line = top_line; this_line; this_line = next_line) {
        next_line = this_line->inq_next();
        this_line->release();
        }
}

/******************************************************************************

    traits allowing access to the phead data fields

******************************************************************************/
/*get the first free line with height >= the height requested. If from_top = 1,
    search from the top of the page down, otherwise search from the bottom
    of the page up.*/
cline* phead::get_freeline(short height, short from_top)
{
    cline *return_line;

    if (from_top){
        for (return_line = top_free; return_line; return_line = return_line->inq_next())
            if ((return_line->get_type() == cl_free) && (return_line->get_hi() >= height)) break;
        return return_line;
        }
    else {
        for (return_line = bottom_free; return_line; return_line = return_line->inq_prior())
            if ((return_line->get_type() == cl_free) && (return_line->get_hi() >= height)) break;
        return return_line;
        }
}

cline* phead::get_tline()
{
    if ((!format_ok) && (!is_visible())) end_pos = format_page(start_pos);

    return(top_line);
}

cline* phead::get_bline()
{
    if ((!format_ok) && (!is_visible())) end_pos = format_page(start_pos);

    return(bottom_line);
}
                   
cline* phead::get_tfree()
{
    if ((!format_ok) && (!is_visible())) end_pos = format_page(start_pos);

    return top_free;
}

cline* phead::get_bfree()
{
    if ((!format_ok) && (!is_visible())) end_pos = format_page(start_pos);

    return bottom_free;
}

cline* phead::get_tused()
{
    if ((!format_ok) && (!is_visible())) end_pos = format_page(start_pos);
    
    return top_used;
}

cline* phead::get_bused()
{
    if ((!format_ok) && (!is_visible())) end_pos = format_page(start_pos);
    
    return bottom_used;
}

/*is the phead currently visible in some viewport? return 1 if it is visible, 0 otherwise*/
long phead::is_visible()
{
    register fv_list *view_list;
           
    if (!get_format())
        return 0;
    for (view_list = get_format()->get_fv_list();
          view_list;
          view_list = view_list->inq_next())
        if (view_list->inq_vp() && (view_list->inq_vp()->is_visible(this)))
            return 1;

    return 0;
}

/******************************************************************************

    traits which alter the phead data fields directly

******************************************************************************/
/*set the size of the page. If the page has only one cline associated with it,
    set the size of that line*/
void phead::set_size(short hi, short wide)
{
    dx = wide;
    dy = hi;
    if (format_ok)
        format_ok = ((hi == dy) && (wide == dx));

    if (top_line == bottom_line) {
        top_line->set_wide(dx);
        top_line->set_hi(dy);
        }
}         

void phead::adjust_line_cnt(long adjust_amt)
{
    register phead *next_page;

    line_cnt += (short) adjust_amt;
    for (next_page = next; next_page; next_page = next_page->next)
        next_page->fline_no += adjust_amt;
}
           
void phead::set_max_line_width(long new_width)
{
    register long old_max_width;

    if (max_line_width == in_format->get_max_line_width())
        old_max_width = max_line_width;
    else
        old_max_width = 0;

    max_line_width = new_width;
    if (max_line_width > in_format->get_max_line_width())
        in_format->set_max_line_width(max_line_width);
    else if (old_max_width && (old_max_width > new_width))
        in_format->calc_max_line_width();
}

void phead::calc_max_line_width()
{
    register cline *hold_line;

    max_line_width = 0;
    for (hold_line = top_used; 
      hold_line && (bottom_used != hold_line->inq_prior());
      hold_line = hold_line->inq_next())
        if ((hold_line->get_contents_width() + hold_line->get_start_x_off()) > max_line_width)
            max_line_width = hold_line->get_contents_width() + hold_line->get_start_x_off();
}

void phead::set_max_used_ht(long new_ht)
{
    if (max_used_ht != new_ht) {
        max_used_ht = new_ht;
        in_format->calc_max_used_ht();
        }
}

void phead::calc_max_used_ht()
{
    register cline *hold_line;
    register long   last_y = -1;

    max_used_ht = 0;
    for (hold_line = top_used; 
      hold_line && (bottom_used != hold_line->inq_prior()); 
      hold_line = hold_line->inq_next()) {
        if (last_y != hold_line->get_uly())
            max_used_ht += hold_line->get_hi();

        last_y = hold_line->get_uly();
        }
}



/******************************************************************************

    Navigation and creation functions

******************************************************************************/
/*next_page--return the next page head if there is one; if not and there is more of the model create one*/
phead* phead::next_page(long reformat)
{
    register phead     *return_page;
    register long      *size_matrix;
    register long       new_dx, new_dy;
             model_pos  next_pos;
           
    /*check to make sure things haven't gone completely haywire*/
    if (!start_pos.in_node || !(start_pos.in_node->get_model() == get_format()->get_model()))
        return 0;
    if (!end_pos.in_node || !(end_pos.in_node->get_model() == get_format()->get_model())){
        format_ok = 0;
        end_pos = format_page(start_pos);
        if (!end_pos.in_node) return 0;
        }

    /*see if there is any more of the model left to format*/
    next_pos = end_pos.get_next();
    if (!next_pos.in_node || !next_pos.can_format())
        return 0;

    /*if there is a next page already, update its start position and return it*/
    if (next && (next->page_num == (page_num + 1))){
        if (next->start_pos != next_pos){
            next->start_pos = next_pos;
            next->fline_no = fline_no + line_cnt;
            next->format_ok = 0;
            size_matrix = (long *) in_format->inq_state(next_pos.in_node, tml_pgsize_prop);
            if (size_matrix) {
                new_dx = *size_matrix;
                new_dy = *(size_matrix + 1);
                }
            else {
                new_dx = in_format->get_page_wide();
                new_dy = in_format->get_page_hi();
                }
            if (new_dx < 0)
                new_dx = in_format->get_fv_list() ? in_format->get_fv_list()->inq_vp()->get_width() : 400;
            if (new_dy < 0)
                new_dy = in_format->get_fv_list() ? in_format->get_fv_list()->inq_vp()->get_width() : 400;
            if ((new_dx != next->dx) || (new_dy != next->dy)) {
                next->dx = (short) new_dx;
                next->dy = (short) new_dy;
                }
            }
        if (reformat) next->format_page(next_pos);
        return(next);
        }
         
    /*if no page already there, make a new one and set it up*/
    return_page = insert_page(next_pos);
    return return_page;
}

/*prior_page--return the prior page head if there is one; if not and there is more of the
    model create one and mark it as beginning at the first character after the root...*/
phead* phead::prior_page(long reformat)
{
    register phead     *return_page;
    register long      *size_matrix;
    register long       new_dx, new_dy;
             model_pos  prior_pos;

    /*see if there is any more of the model left to format*/
    prior_pos = start_pos.get_prior();
    if (!prior_pos.can_format())
        return 0;

    /*if there is a prior page already, update its end position and return it*/
    if (prior && (prior->page_num == (page_num - 1))){
        if (prior->end_pos != prior_pos){
            prior->end_pos = prior_pos;
            prior->format_ok = 0;
            size_matrix = (long *) in_format->inq_state(prior_pos.in_node, tml_pgsize_prop);
            if (size_matrix) {
                new_dx = *size_matrix;
                new_dy = *(size_matrix + 1);
                }
            else {
                new_dx = in_format->get_page_wide();
                new_dy = in_format->get_page_hi();
                }
            if (new_dx < 0)
                new_dx = in_format->get_fv_list() ? in_format->get_fv_list()->inq_vp()->get_width() : 400;
            if (new_dy < 0)
                new_dy = in_format->get_fv_list() ? in_format->get_fv_list()->inq_vp()->get_width() : 400;
            if ((new_dx != prior->dx) || (new_dy != prior->dy)) {
                prior->dx = (short) new_dx;
                prior->dy = (short) new_dy;
                }
            }
        if (reformat) prior_pos = prior->format_page(prior->start_pos);
        return(prior);
        }
         
    /*if no page already there, make a new one and set it up*/
    return_page = insert_page(prior_pos, 0);
    return return_page;
}


/* checks for presence of a member of the pgnode class hierarchy */
/* "root" here refers to root of subtree in a header or footer */
/*   property argument */
long check_for_pgnum(node* root)
{
    node*    next_node;
    node_dir dir;

    if (0 == root)
        return 0;
    next_node = root;
    dir = n_down;
    do {
        if (0 == next_node)
            break;
        if (next_node->is_pgnum())
            return 1;
        do {
            next_node = next_node->get_next(dir);
            if ((0 == next_node) || (next_node == root))
                break;
            } while (n_up == dir);
        } while ( ! ((root == next_node) && (n_up == dir)));
    return 0;
}


phead* phead::insert_page(model_pos start, long after)
{
               phead     *return_page,
                         *hold_page;
               long      *size_matrix, 
                          new_dx,
                          new_dy;
               model_pos  next_pos;
               long       offset_page;
    register   long       j;

    return_page = new phead(in_format);
    if (!return_page)
        return 0;
    if (!start.in_node->can_format())
        start = (after ? start.get_next() : start.get_prior());
    if (!start.in_node->can_format()){
        delete return_page;
        return 0;
        }
    size_matrix = (long *) in_format->inq_state(start.in_node, tml_pgsize_prop);
    if (size_matrix) {
        new_dx = *size_matrix;
        new_dy = *(size_matrix + 1);
        }
    else {
        new_dx = in_format->get_page_wide();
        new_dy = in_format->get_page_hi();
        }
    if (new_dx < 0)
        new_dx = in_format->get_fv_list() ? in_format->get_fv_list()->inq_vp()->get_width() : 400;
    if (new_dy < 0)
        new_dy = in_format->get_fv_list() ? in_format->get_fv_list()->inq_vp()->get_width() : 400;
    return_page->dx = (short) new_dx;
    return_page->dy = (short) new_dy;
    return_page->top_line->set_wide(return_page->dx);
    return_page->top_line->set_hi(return_page->dy);
    return_page->top_line->set_ul_coord(0,0);
    return_page->format_ok = 0;
    if (after) {
        return_page->start_pos = start;
        // SET return_page PAGE NUMBER BEFORE FORMATTING
        if (start <= end_pos)                 // return_page starts at a position within or before this page
            return_page->page_num = page_num;
        else if (start == end_pos.get_next())    // return_page starts immediately after this page
            return_page->page_num = page_num + 1;
        else                                // there will be a gap between this page and return_page
            return_page->page_num  = page_num + 2;

        next_pos = return_page->format_page(start);
        return_page->next = next;
        if (next) 
            next->prior = return_page;
        else
            in_format->set_last_page(return_page);
        return_page->prior = this;
        next = return_page;
        if (return_page->next) {
            // CALCULATE PAGE NUMBER OF PAGE FOLLOWING return_page
            offset_page = return_page->page_num;
            //see if there is a gap between the inserted page and the next page
            if (next_pos.get_next() < return_page->next->start_pos)
                offset_page += 2;
            else
                offset_page++;
            // return_page ends within or beyond its next page -- 
            // reset the start position of the next page
            if (next_pos >= return_page->next->start_pos){
                return_page->next->start_pos = next_pos.get_next();
                return_page->next->format_ok = 0;
                }
            }
        // update page numbers and line counts on succeeding pages
        j = return_page->line_cnt;
        return_page->fline_no = return_page->prior->fline_no + return_page->prior->line_cnt;
        for (hold_page = return_page->next; hold_page; hold_page = hold_page->next) {
            hold_page->page_num = offset_page++;
            hold_page->fline_no += j;
            }
       }
    else {
        return_page->end_pos = start;
        //start by assuming that the page number is the same as the page we are inserting before
        if (prior){
            return_page->page_num = page_num++;
            return_page->fline_no = prior->fline_no + prior->line_cnt;
            prior->next = return_page;
            }
        else {
            return_page->page_num = 1;
            in_format->set_first_page(return_page);
            return_page->fline_no = 1;
            }
        next_pos = return_page->tamrof_page(start);
        //now find out the real page number, reformatting if necessary
        if (prior) {
            if (next_pos <= prior->end_pos){
                return_page->start_pos = prior->end_pos.get_next();
                return_page->page_num = prior->page_num + 1;
                return_page->format_ok = 0;
                next_pos = return_page->format_page(return_page->start_pos);
                }
            else if ((next_pos == prior->end_pos.get_next()) &&
                   (return_page->page_num != (prior->page_num + 1))){
                return_page->page_num = prior->page_num + 1;
                return_page->format_ok = 0;
                next_pos = return_page->format_page(return_page->start_pos);
                }
            else if (return_page->page_num != (prior->page_num +2)){
                return_page->page_num = prior->page_num + 2;
                return_page->format_ok = 0;
                next_pos = return_page->format_page(return_page->start_pos);
                }                                                     
            return_page->prior = prior;
            }

        prior = return_page;
        return_page->next = this;
        next_pos = return_page->end_pos.get_next(); //find the start of the next page
        /*calculate the page number of the following page*/
        if (start_pos < next_pos){
            start_pos = next_pos;
            format_ok = 0;
            offset_page = return_page->page_num + 1;
            }
        else if (start_pos == next_pos)
            offset_page = return_page->page_num + 1;
        else
            offset_page = return_page->page_num + 2;

        j = return_page->line_cnt;
        for (hold_page = this; hold_page; hold_page = hold_page->next){
            hold_page->page_num = offset_page++;
            hold_page->fline_no += j;
            }
        }
    return return_page;
}

/*find the conceptual line containing the indicated model position. If the line is not
    on the page, return 0*/
cline* phead::mpos_2_cline(model_pos target)
{
    register cline     *return_line;
             model_pos  find_pos;
 
    if (!target.in_node) return 0;

    if (!target.in_node->can_format())
        find_pos = target.get_next();
    else
        find_pos = target;

    /* nothing to find */
    if (!find_pos.in_node) return 0;

    /* within cline that contains model position sought, */
    /*   find first line containing material from node sought */
    for (return_line = mpos_2_hcline(find_pos); 
      return_line && (find_pos.in_node != return_line->get_start().in_node);
      return_line = return_line->get_next())
        ;

    /* find line containing character index sought */
    while (return_line) {
        if (cl_text == return_line->get_type()) {
            if (find_pos.in_node != ((tline*) return_line)->get_start().in_node) {
                /* offset sought is not formatted */
                return_line = NULL;
                break;
                }
            if ((((tline*) return_line)->get_count())
              > (find_pos.char_index - return_line->get_start().char_index))
                /* offset sought is in return_line */
                break;
            }
        return_line = return_line->get_next();
        }

    return return_line;
}

/*find the head conceptual line containing the indicated model position. */
/*  if the line is not on the page, return 0*/
cline* phead::mpos_2_hcline(model_pos target)
{
    register cline     *return_line;
             model_pos  find_pos;

    if (!target.in_node) return 0;

    if ((!format_ok) && (!is_visible()))
        end_pos = format_page(start_pos);
                                 
    if (!target.in_node->can_format())
        find_pos = target.get_next();
    else
        find_pos = target;

    if (!find_pos.in_node) return 0;

    if (find_pos < start_pos)
        return 0;

    for (return_line = top_used; 
      return_line && (bottom_used != return_line->inq_prior()); 
      return_line = return_line->inq_next()) {
        if (return_line->get_end() >= find_pos)
            break;
        }

    return return_line;
}
                      
/******************************************************************************

    Traits which manipulate the conceptual lines on the page

******************************************************************************/
/*free all of the clines on a page,  and initialize the page to contain one
    free cline with all of the page's space*/
void phead::free_clines(cline *start_line )
{
    register cline *this_line,
                   *next_line;

    /*if we are only freeing up part of the page, we need to do some housekeeping first*/
    if (start_line && (start_line != top_line)){
        /*if we aren't on the right page, just get out*/
        if (start_line->get_phead() != this)
            return;
        /*get the line above, and make it the bottom line*/
        bottom_line = start_line->inq_prior();
        bottom_line->set_next(0);
        if (!bottom_line) {
            free_clines();
            return;
            }
        else if (bottom_line->get_lry() < bottom_used->get_uly())
            bottom_used = bottom_line;

        /*now we can free the lines*/
        for (this_line = start_line; this_line; this_line = next_line) {
            if ((this_line->get_type() != cl_free) &&
               (this_line->get_type() != cl_blank) &&
               (this_line->get_type() != cl_gutter))
                line_cnt--;
            if (this_line == top_free)
                top_free = 0;
            next_line = this_line->inq_next();
            this_line->release();
            }

        if (cl_free == bottom_line->get_type()) {
            for (this_line = bottom_line->inq_prior(); 
              this_line && (this_line->get_type() == cl_free); 
              this_line = this_line->inq_prior())
                ;
            if (this_line) {
                if (this_line->get_uly() < bottom_used->get_lry())
                    bottom_used = this_line;
                /* in the "else" for this "if", the freed lines */
                /* were all lower than bottom_used, i.e. part */
                /* of a footer or the bottom margin, and hence */
                /* bottom_used has not changed */
                }
            else
                bottom_used = top_used = 0;
            if (bottom_line->get_wide() == dx)
                bottom_free = bottom_line;
            else {
                bottom_free = new cline(in_format);
                bottom_free->set_ul_coord(0, bottom_line->get_lry());
                }
            }
        else {
            bottom_free = new cline(in_format);
            bottom_free->set_ul_coord(0, bottom_used->get_uly() + bottom_used->get_hi());
            }
        bottom_free->set_wide(dx);
        bottom_free->set_hi(dy - bottom_free->get_uly());
        bottom_free->set_type(cl_free);
        bottom_free->set_phead(this);
        bottom_free->set_next(0);
        bottom_free->set_prior(bottom_line);
        bottom_line->set_next(bottom_free);
        bottom_line = bottom_free;
        if (!top_free)
            top_free = bottom_free;
        return;
        }
                 
    /*we are freeing all of the lines on the page*/
    start_line = top_line;
    top_line = bottom_line = top_free = bottom_free = new cline(in_format);
    top_used = bottom_used = 0;
    bottom_free->set_ul_coord(0, 0);
    bottom_free->set_wide(dx);
    bottom_free->set_hi(dy);
    line_cnt = 0;

    for(this_line = start_line; this_line; this_line = next_line){
        next_line = this_line->inq_next();
        this_line->release();
        }

    bottom_free->set_type(cl_free);
    bottom_free->set_phead(this);
    format_ok = 0;
}

/*place a line with its upper left corner at the spot indicated. If there is no available free
    space which can hold the line at that coordinate, will return -1; otherwise the line is
    placed and return is 0*/
long phead::line_place(cline *place_line, short at_x, short at_y)
{
    register cline *use_free,
                   *split_line;
                                                                                                    
    /*first, find if there is free space able to contain the line at the place requested*/
    for (use_free = top_free; use_free && (use_free->get_uly() <= at_y); use_free = use_free->inq_next())
        if ((use_free->get_type() == cl_free) &&
            (use_free->get_uly() <= at_y) &&
            (use_free->get_ulx() <= at_x) &&
            (use_free->get_lrx() >= (at_x + place_line->get_wide())) &&
            (use_free->get_lry() >= (at_y + place_line->get_hi())))
                break;

    if (!use_free || (use_free->get_uly() > at_y)) return -1;
                                              
    /*if necessary, split the line so that it is the right size*/
    if (at_y != use_free->get_uly())
        split_line = use_free->split((at_y - use_free->get_uly()));
    if (use_free->get_hi() != place_line->get_hi())
        use_free = use_free->split(place_line->get_hi());
    if (use_free->get_ulx() != at_x)
        split_line = use_free->split((at_x - use_free->get_ulx()), 0, 1);
    if (use_free->get_wide() != place_line->get_wide())
        use_free= use_free->split(place_line->get_wide(), 0, 1);

    place_line->set_ul_coord(at_x, at_y);
    /* replace the correctly sized free line with the one to be placed */
    place_line->replace(use_free);
    return 0;
}
     
/*adjust the ulx, uly coordinates of all of the lines from start_line on down by the indicated amount.
    if any lines go off the page, change the end position of this page, update the start position of the
    next page, and mark the next page as needing reformatting*/
void phead::adjust_xy(cline *start_line, long deltax, long deltay)
{
    register cline *adjust_line;
           
    for(adjust_line = start_line; adjust_line; adjust_line = adjust_line->inq_next())
        adjust_line->adjust_xy(deltax, deltay);

    if ((bottom_line->get_uly() + bottom_line->get_hi()) > dy)
        format_ok = 0;
}

/******************************************************************************

    The formatting function

******************************************************************************/
/*format the page for which this is the header, starting at model position specified*/
model_pos phead::format_page(model_pos start)
{
    register cline     *this_line,
                       *free_space;
    register node      *trace_node;
    register long       lines_set;
    register long      *page_size,
                       *margin_array,
                        last_line_width,
                        open_width,
                        right_margin,
                        last_y = -1, left_margin;
             model_pos  start_line;
             long       old_max_width,
                        new_max_used_ht, 
                        old_line_cnt;
             phead     *adjust_page;
             char_set  *this_char_set;
             lrud_dir   this_dir;

    /*check for incoming trash*/
    if (!this || !start.in_node)
        return start;
                                          
    /*make sure that start.in_node is still connected*/
    for (trace_node = start.in_node; trace_node->get_node(n_up); trace_node = trace_node->get_node(n_up))
        ;
    if (trace_node != in_format->get_model()->get_root())
        if (prior)
            start = prior->end_pos.get_next();
        else {
            start.in_node = in_format->get_model()->get_root();
            start.char_index = 0;
            }

    if (format_ok && (start.in_node == start_pos.in_node) && (start.char_index == start_pos.char_index))
        return end_pos;
                
    if (is_visible())
        return end_pos;

    in_format->make_rp_entry(this);

    if (!start.in_node){
        start_pos = end_pos = start;
        return start;
        }

    start_line = start.in_node->can_format() ? start : start.get_next();
    start_pos = end_pos = start_line;
    if (!start_pos.in_node)
        return end_pos;

    if (!start_line.in_node->can_format())
        return start_line;
                                 
    if (in_format->get_max_line_width() == max_line_width)   
        old_max_width = max_line_width;
    else
        old_max_width = 0;
    max_line_width = 0;

    new_max_used_ht = 0;

    page_size = (long *)(in_format->inq_state(start_line.in_node, tml_pgsize_prop));
    if (page_size){
        dx = (short) ((*page_size > 0) ? *page_size : in_format->get_fv_list()->inq_vp()->get_width());
        dy = (short) ((*(page_size + 1) > 0) ? *(page_size + 1) : in_format->get_fv_list()->inq_vp()->get_hi());
        }
    else {
        dx = (short) ((in_format->get_page_wide()) ? (in_format->get_page_wide()) : in_format->get_fv_list()->inq_vp()->get_width());
        dy = (short) ((in_format->get_page_hi()) ? (in_format->get_page_hi()) : in_format->get_fv_list()->inq_vp()->get_hi());
        }
    top_free->set_hi(dy);
    top_free->set_wide(dx);

    set_margins();
    lines_set = 0;
    if (next)
        old_line_cnt = next->fline_no - fline_no;
    else
        old_line_cnt = 0;

    lines_set += set_header_footer();

    free_space = top_free;

    while (free_space && bottom_free && start_line.in_node && (start_line.in_node->can_format())){
        if (start_line.in_node->get_format_hi(in_format) > free_space->get_hi()){
            free_space = make_blank(free_space);
            continue;
            }
        /*get the left and right margin sizes*/
        if (start_line.in_node->get_model())
            margin_array = (long *)(start_line.in_node->get_model()->inq_state(start_line.in_node, tml_margin_prop));
        else
            margin_array = 0;
        if (margin_array){
            left_margin =  margin_array[left_marg];
            right_margin = margin_array[right_marg];
            }
        else
            right_margin = left_margin = 0;
        /*adjust the width available for formatting by the right and left margin size*/
        open_width = free_space->get_wide() - (right_margin + left_margin);

        this_char_set = (char_set*) start_line.in_node->get_model()->inq_state(start_line.in_node, tml_char_set_prop);
        if (::right == (this_dir = this_char_set->get_dir_to_char(((unsigned char*) start_line.in_node->get_contents() + start_line.char_index))))
            this_line = start_line.in_node->format_line(in_format, start_line.char_index,  free_space->get_hi(), open_width);
        else if (::left == this_dir)
            this_line = start_line.in_node->tamrof_line(in_format, start_line.char_index,  free_space->get_hi(), open_width);

        if (this_line){
            this_line->set_start_x_off(this_line->get_start_x_off() + left_margin);
            this_line->set_wide((short) (this_line->get_wide() + (left_margin + right_margin)));
            last_line_width = this_line->get_contents_width() + left_margin + right_margin;
            if (last_line_width > max_line_width)
                max_line_width = last_line_width;
            free_space = free_space->alloc_space(this_line);

            if (last_y != this_line->get_uly())
                new_max_used_ht += this_line->get_hi();
    
            last_y = this_line->get_uly();

            if (!top_used)
                top_used = this_line;
            bottom_used = this_line;
            lines_set++;
            if (!free_space)
                free_space = top_free;
            }
        else {
            free_space = make_blank(free_space);
            continue;
            }
        end_pos = this_line->get_end();
        start_line = end_pos.get_next();
        }
    
    line_cnt = (short) lines_set;
    old_line_cnt = lines_set - old_line_cnt;
    if (old_line_cnt){
        for (adjust_page = next; adjust_page; adjust_page = adjust_page->next)
            adjust_page->fline_no += old_line_cnt;
        }

    if (max_line_width > in_format->get_max_line_width())
        in_format->set_max_line_width(max_line_width);
    else if (old_max_width)
        in_format->calc_max_line_width();

    if (max_used_ht != new_max_used_ht)
    {
        max_used_ht = new_max_used_ht;
        in_format->calc_max_used_ht();
    }

    format_ok = 1;
    return end_pos;
}
 
/*format the page in reverse for which this is the header, starting at model position specified*/
model_pos phead::tamrof_page(model_pos start)
{
    register cline     *this_line,
                       *free_space;
             model_pos  start_line;
    register node      *trace_node;
    register long       lines_set;
             phead     *adjust_page;
    register long      *page_size,
                       *margin_array,
                        last_line_width,
                        open_width,
                        right_margin,
                        last_y = -1, left_margin;
             long       old_max_width,                
                        new_max_used_ht,
                        old_line_cnt;
             char_set  *this_char_set;
             lrud_dir   this_dir;

    /*check for incoming trash*/
    if (!this || !start.in_node)
        return start;
                                          
    /*make sure that start.in_node is still connected*/
    for (trace_node = start.in_node; trace_node->get_node(n_up); trace_node = trace_node->get_node(n_up))
        ;
    if (trace_node != in_format->get_model()->get_root())
        if (prior)
            start = prior->end_pos.get_next();
        else {
            start.in_node = in_format->get_model()->get_root();
            start.char_index = 0;
            }

    if (format_ok && (start.in_node == end_pos.in_node) && (start.char_index == end_pos.char_index))
        return start_pos;
                
    if (is_visible())
        return start_pos;
                               
    old_line_cnt = line_cnt;

    if ((format_ok) || (top_line != bottom_line) || (!top_free))
        free_clines();
    else
        in_format->make_rp_entry(this);

    if (!start.in_node){
        start_pos = end_pos = start;
        return start;
        }

    start_line = start.in_node->can_format() ? start : start.get_prior();
    start_pos = end_pos = start_line;
    if (!start_pos.in_node)
        return end_pos;

    if (!start_line.in_node->can_format())
        return start_line;
                                        
    if (max_line_width == in_format->get_max_line_width())
        old_max_width = max_line_width;
    else
        old_max_width = 0;
    max_line_width = 0;

    new_max_used_ht = 0;

    page_size = (long *)(in_format->inq_state(start_line.in_node, tml_pgsize_prop));
    if (page_size){
        dx = (short) ((*page_size > 0) ? *page_size : in_format->get_fv_list()->inq_vp()->get_width());
        dy = (short) ((*(page_size + 1) > 0) ? *(page_size + 1) : in_format->get_fv_list()->inq_vp()->get_hi());
        }
    else {
        dx = (short) ((in_format->get_page_wide()) ? (in_format->get_page_wide()) : in_format->get_fv_list()->inq_vp()->get_width());
        dy = (short) ((in_format->get_page_hi()) ? (in_format->get_page_hi()) : in_format->get_fv_list()->inq_vp()->get_hi());
        }
    top_free->set_hi(dy);
    top_free->set_wide(dx);

    set_margins();
    lines_set = 0;
    if (next)
        line_cnt = (short) (next->fline_no - fline_no);
 
    lines_set += set_header_footer();

    free_space = bottom_free;

    while (free_space && top_free && start_line.in_node && (start_line.in_node->can_format())){
        if (start_line.in_node->get_format_hi(in_format) > free_space->get_hi()){
            free_space = make_blank(free_space);
            continue;
            }
        /*get the left and right margin sizes*/
        if (start_line.in_node->get_model())
            margin_array = (long *)(start_line.in_node->get_model()->inq_state(start_line.in_node, tml_margin_prop));
        else
            margin_array = 0;
        if (margin_array){
            right_margin = margin_array[right_marg];
            left_margin = margin_array[left_marg];
            }
        else
            right_margin = left_margin = 0;
        /*adjust the width available for formatting by the right and left margin size*/
        open_width = free_space->get_wide() - (right_margin + left_margin);

        this_char_set = (char_set*) start_line.in_node->get_model()->inq_state(start_line.in_node, tml_char_set_prop);
        if (::right == (this_dir = this_char_set->get_dir_to_char(((unsigned char*) start_line.in_node->get_contents() + start_line.char_index))))
            this_line = start_line.in_node->tamrof_line(in_format, start_line.char_index,  free_space->get_hi(), open_width);
        else if (::left == this_dir)
            this_line = start_line.in_node->format_line(in_format, start_line.char_index,  free_space->get_hi(), open_width);

        if (this_line){
            this_line->set_start_x_off(this_line->get_start_x_off() + left_margin);
            this_line->set_wide((short) (this_line->get_wide() + (left_margin + right_margin)));
            last_line_width = this_line->get_contents_width() + left_margin + right_margin;
            if (last_line_width > max_line_width)
                max_line_width = last_line_width;
            free_space = free_space->alloc_space(this_line, 1);

            if (last_y != this_line->get_uly())
                new_max_used_ht += this_line->get_hi();
    
            last_y = this_line->get_uly();

            if (!bottom_used)
                bottom_used = this_line;
            top_used = this_line;
            lines_set++;
            if (!free_space)
                free_space = bottom_free;
            }
        else {
            free_space = make_blank(free_space);
            continue;
            }
        start_pos = this_line->get_start();
        start_line = start_pos.get_prior();
        }


    old_line_cnt = lines_set - old_line_cnt;
    if (old_line_cnt){
        for (adjust_page = next; adjust_page; adjust_page = adjust_page->next)
            adjust_page->fline_no += old_line_cnt;
        line_cnt = (short) lines_set;
        }

    if (max_line_width > in_format->get_max_line_width())
        in_format->set_max_line_width(max_line_width);
    else if (old_max_width)
        in_format->calc_max_line_width();

    if (max_used_ht != new_max_used_ht) {
        max_used_ht = new_max_used_ht;
        in_format->calc_max_used_ht();
        }

    format_ok = 1;
    return start_pos;
}
 
/*private function used by the formatting component to mark a free space as
    blank when that space is too small to be used by the next line to be formatted--
    returns the next free line after the line to be blanked*/
cline* phead::make_blank(cline *blank_line)
{
    register cline *return_line;
            
    /*first, find the next line below the line to blank which is free*/
    for (return_line = blank_line->inq_next();
           return_line && (return_line->get_type() != cl_free);
           return_line = return_line->inq_next())
           ;
         
    /*if the line is free, mark it as blank and if necessary update the page header*/
    if (cl_free == blank_line->get_type()) {
        if (top_free == blank_line)
            top_free = return_line;
        if (bottom_free == blank_line)
            bottom_free = 0;
        blank_line->set_type(cl_blank);
        }

    return return_line;
}

/*private function used by the formatting function to set the margins*/
void  phead::set_margins()
{
    register long   uly, lly;
             long  *margin_array;
             cline *margin_area;

    /*NOTE: THIS FUNCTION ASSUMES THAT TOP (BEGINNING OF PAGE) MARGIN*/
    /*   AND BOTTOM (END OF PAGE) MARGIN ARE RECORDED IN THE ZEROTH */
    /*   AND ONETH POSITIONS OF THE MARGIN ARRAY RESPECTIVELY */

    margin_array = (long *)(in_format->inq_state(start_pos.in_node, tml_margin_prop));
    if (!margin_array) return;

    /*if there is a margin, allocate the areas*/
    if (margin_array){
        uly = 0;
        lly = dy;
        if (*margin_array){
            margin_area = new cline(in_format);
            margin_area->set_type(cl_gutter);
            margin_area->set_hi((short) *margin_array);
            margin_area->set_wide(dx);
            if (line_place(margin_area, (short) 0, (short) 0))
                delete margin_area;
            else
                top_line = margin_area;
            uly = *margin_array;
            }

        margin_array++;
        if (*margin_array){
            margin_area = new cline(in_format);
            margin_area->set_type(cl_gutter);
            margin_area->set_hi((short) *margin_array);
            margin_area->set_wide(dx);
            if (line_place(margin_area, (short) 0, (short) (dy - *margin_array)))
                delete margin_area;
            else {
                if (!top_line)
                    top_line = margin_area;
                bottom_line = margin_area;
                }
            lly = dy - (*margin_array + uly);
            }
        }
}


/* private function used by the formatting function to set headers and footers */
long phead::set_header_footer()
{
    register node   **hdr, **ftr, **hdr_arr, **ftr_arr, **hf_arr,
                    *hdr_arr_space[3], *ftr_arr_space[3],
                    *sv_prop, *sv_node;

    model           *sv_mod;

    register cline  *alloc_from, *free_space, *this_line, *line_ptr, *cur_line,
                    *split_line;

    /* margins */
    unsigned long    bottom_margin, left_margin, right_margin, *margin_array;

    /* horizontal and vertical measure */
    unsigned long    prev_x,     at_y,   at_x[3],  t_width, x,
                     open_width, offset, which_hf, height;
    model_pos        start_line;

    /* whether header or footer */
    unsigned long    top_loc;

    /* whether left/center/right component */                         
    unsigned long    lcr_hf;

    /* character set and direction */
    char_set         *this_char_set;
    lrud_dir          this_dir;

    /* counters */
    unsigned long    i, total_clines, temp_index, lines_set;

    /* booleans */
    unsigned long    line_placed, 
                     first_footer_component,
                     first_component;

    hdr = (node**) in_format->inq_state(start_pos.in_node, tml_header_prop);
    ftr = (node**) in_format->inq_state(start_pos.in_node, tml_footer_prop);
 
    lines_set = 0;

    if (!hdr && !ftr)
        return lines_set;

    /*where page numbers are involved, copy the subtree: */
    /*  must be able to format and reformat each page number */
    /*  independently */
    if (hdr) {
        for (i = 0; i < 3; i++) {
            if (*(hdr + i)) {
                *(hdr_arr_space + i) = (check_for_pgnum(*(hdr + i)))
                                       ? (*(hdr + i))->copy_all(in_format->get_model())
                                       : *(hdr + i);
                }
            else
                *(hdr_arr_space + i) = 0;
            }
        hdr_arr = hdr_arr_space;
        }
    else
        hdr_arr = 0;

    if (ftr) {
        for (i = 0; i < 3; i++) {
            if (*(ftr + i)) {
                *(ftr_arr_space + i) = (check_for_pgnum(*(ftr + i)))
                                       ? (*(ftr + i))->copy_all(in_format->get_model())
                                       : *(ftr + i);
                }
            else
                *(ftr_arr_space + i) = 0;
            }
        ftr_arr = ftr_arr_space;
        }
    else
        ftr_arr = 0;

    /* note model and get margins */
    if (sv_mod = start_pos.in_node->get_model()) {
        sv_prop = sv_mod->get_prop_pos();
        margin_array = (unsigned long *)(start_pos.in_node->get_model()->inq_state(start_pos.in_node, tml_margin_prop));
        if (margin_array) {
            bottom_margin = margin_array[bottom_marg];
            left_margin   = margin_array[left_marg];
            right_margin  = margin_array[right_marg];
            }
        else
            bottom_margin = left_margin = right_margin = 0;
        }
    else
        bottom_margin = left_margin = right_margin = 0;

    /* set nominal header/footer component positions */
    at_x[0] = left_margin;       // Left header/footer component location
    at_x[1] = dx/2;              // Center header/footer component location before line is centered
    at_x[2] = dx - right_margin; // Right header/footer component location before line length is subtracted

    /* traverse header and footer subtrees */
    for (which_hf = 0; which_hf < 2; which_hf++) {
        t_width = prev_x = line_placed = 0;
        if (0 == which_hf) {     // header
            top_loc = 1;
            hf_arr = hdr_arr;
            alloc_from = top_free;
            first_footer_component = 0;
            }
        else {                   // footer
            top_loc = 0;
            hf_arr = ftr_arr;
            alloc_from = bottom_free;
            first_footer_component = 1;
            }
        if (0 == alloc_from)     // in case header fills page before footer begun
            break;
        at_y = (top_loc) ? (alloc_from->get_uly()) : (dy - bottom_margin);
        open_width = alloc_from->get_wide() - (right_margin + left_margin);
        free_space = alloc_from;

        // Place left, center, right header/footer
        for (lcr_hf = 0; (lcr_hf < 3) && hf_arr; lcr_hf++, hf_arr++) {
            first_component = 1;
            line_ptr = 0;
            offset = total_clines = 0;
            sv_node = 0;
            start_line.in_node = *hf_arr;
            start_line.char_index = 0;
            if (0 != start_line.in_node) {
                if (sv_mod)
                    sv_mod->set_prop_pos(start_line.in_node);
                sv_node = start_line.in_node;
                start_line.in_node->enter_props();
                if (0 == start_line.in_node->can_format()) {
                    start_line = start_line.get_next();
                    if ((0 == start_line.in_node) || (0 == start_line.in_node->can_format())) {
                        if (sv_node)
                            sv_node->exit_props();
                        if (sv_mod)
                            sv_mod->set_prop_pos(sv_prop);
                        continue;
                        }
                    }
                }

            // Build all clines for headers and footers to be adjacent to one another
            // then adjust horizontally.
            while (free_space && start_line.in_node && (start_line.in_node->can_format())) {

                // need more height
                if (start_line.in_node->get_format_hi(in_format) > free_space->get_hi()) {
                    free_space = make_blank(free_space);
                    continue;
                    }

                // pass page number instead of node index to node::format_line()
                //   for nodes in pgnode class hierarchy (page number nodes)
                temp_index = (start_line.in_node->is_pgnum())
                             ? page_num : start_line.char_index;

                this_char_set = (char_set*) start_line.in_node->get_model()->inq_state(start_line.in_node, tml_char_set_prop);
                this_line = start_line.in_node->format_line(in_format, temp_index, free_space->get_hi(), open_width);

                // failure to format -- stop formatting components
                if (0 == this_line)
                    break;
                // component successfully formatted
                else {
                    if (first_footer_component) {
                        at_y -= this_line->get_hi();
                        first_footer_component = 0;
                        }
                    if (first_component) {
                        line_ptr = this_line;
                        first_component = 0;
                        }
                    else
                        prev_x += t_width;
                    t_width = ((tline*) this_line)->get_full_text_width();

                    // cumulative length of header/footer so far
                    offset += t_width;

                    // record width for this component
                    this_line->set_wide((short) t_width);

                    // check for fit; if not, stop formatting components
                    if (-1 == line_place(this_line, (short) (at_x[0] + prev_x), (short) at_y)) {
                        t_width = 0;
                        break;
                        }
                    else {
                        height = this_line->get_hi();
                        ++total_clines;
                        line_placed = 1;
                        }

                    // get next empty space
                    free_space = get_freeline((short) this_line->get_hi(), (short) top_loc);
                    }
                start_line = this_line->get_end().get_next();
                open_width -= t_width;
                }

            // add width of last formatted component
            prev_x += t_width;

            // horizontally adjust center and right header/footer components
            //   (left component always begins at left margin)
            if ((lcr_hf > 0) && (line_ptr)) {
                if (lcr_hf == 1)
                    offset >>= 1;
                x = at_x[lcr_hf] - offset;
                for (i = 0, cur_line = line_ptr; 
                  i < total_clines; 
                  i++, cur_line = cur_line->get_next()) {
                    cur_line->set_ul_coord ((short) x, (short) cur_line->get_uly());
                    x += ((tline*) cur_line)->get_full_text_width();
                    }
                }
            if (sv_node)
                sv_node->exit_props();
            }

        // if header/footer placed, make gutters of remaining portion of line
        if (line_placed) {
            while (this_line = get_freeline((short) height, (short) top_loc)) {
                if (this_line->get_uly () == at_y) {
                    make_blank(this_line);         
                    this_line->set_type(cl_gutter);
                    while ((!bottom_free) && (this_line != top_free)) {
                        this_line = this_line->inq_prior();
                        if (this_line->get_type() == cl_free)
                            bottom_free = this_line;
                        }
                    }
                else {     
                    // Add blank line below headers and above footers for
                    // purely cosmetic reasons.
                    split_line = this_line->split((short) height, (short) which_hf, (short) 0);
                    make_blank(split_line);
                    split_line->set_type(cl_gutter);

                    // make_blank sets bottom_free to null when a line is split
                    //   from the bottom of the bottom free line
                    // this may occur when a blank line is created after a footer
                    if (!bottom_free)
                        bottom_free = split_line->inq_prior();
                    break;
                    }
                }
            ++lines_set;
            }
        }
    if (sv_mod)
        sv_mod->set_prop_pos(sv_prop);
    return lines_set;
}


/*reformat the page. start indicates the last model position which doesn't need to be reformatted
    (this should be the end position of the last line correctly formatted). The model position returned is the
    end of the last line which was reformatted*/
model_pos phead::reformat_page(model_pos start)
{
    register cline *start_line,
                   *this_line,
                   *margin_line,
                   *free_space;
    register long  *margin_array,
                    open_width,
                    old_line_cnt,
                    this_line_width,
                    lines_set,
                    right_margin,
                    left_margin;
    phead          *adjust_page;
    long            old_max_line_width;
    model_pos       return_pos,
                    start_reformat;
    char_set       *this_char_set;
    lrud_dir        this_dir;

    /*first, find the last line correctly formatted*/
    start_line = mpos_2_hcline(start);
    if (!start_line){
        /*wasn't found--something has gone wrong*/
        return_pos.in_node = 0;
        return_pos.char_index = -1;
        return return_pos;
        }

    start_reformat = start_line->get_end().get_next();
    if (!start_reformat.in_node){
        /*nothing left to reformat*/
        end_pos = start;
        return start;
        }

    old_line_cnt = line_cnt;
    if (max_line_width == in_format->get_max_line_width())
        old_max_line_width = max_line_width;
    else
        old_max_line_width = 0;

    start_line = start_line->get_down();
    if (start_line)
        free_clines(start_line);
    
    max_line_width = 0;
    for(margin_line = top_line; margin_line; margin_line = margin_line->inq_next())
        if ((margin_line->get_contents_width() + margin_line->get_start_x_off()) > max_line_width)
            max_line_width = margin_line->get_contents_width() + margin_line->get_start_x_off();
                                              
    /*set the bottom margin if there is one*/
    margin_array = (long *)(start_pos.in_node->get_model()->inq_state(start_reformat.in_node, tml_margin_prop));
    if (margin_array){
        margin_array++;
        if (*margin_array){
            margin_line = new cline(in_format);
            margin_line->set_type(cl_gutter);
            margin_line->set_hi((short) *margin_array);
            margin_line->set_wide(dx);
            if (line_place(margin_line, (short) 0, (short) (dy - *margin_array)))
                delete margin_line;
            else {
                if (!top_used)
                    top_line = margin_line;
                bottom_line = margin_line;
                }
            }
        }
                   
    lines_set = line_cnt;
    if (next)
        line_cnt = (short) (next->fline_no - fline_no);
    free_space = top_free;
    if (!free_space)
        free_space = bottom_free;
    while (bottom_free && start_reformat.in_node && (start_reformat.in_node->can_format())){
        if (start_reformat.in_node->get_format_hi(in_format) > free_space->get_hi()){
            free_space = make_blank(free_space);
            continue;
            }
        /*get the left and right margin sizes*/
        if (start_reformat.in_node->get_model())
            margin_array = (long *)(start_reformat.in_node->get_model()->inq_state(start_reformat.in_node, tml_margin_prop));
        else
            margin_array = 0;
        if (margin_array){
            right_margin = margin_array[right_marg];
            left_margin = margin_array[left_marg];
            }
        else
            right_margin = left_margin = 0;
        /*adjust the width available for formatting by the right and left margin size*/
        open_width = free_space->get_wide() - (right_margin + left_margin);

        this_char_set = (char_set*) start_reformat.in_node->get_model()->inq_state(start_reformat.in_node, tml_char_set_prop);
        if (::right == (this_dir = this_char_set->get_dir_to_char(((unsigned char*) start_reformat.in_node->get_contents() + start_reformat.char_index))))
            this_line = start_reformat.in_node->format_line(in_format, start_reformat.char_index,  free_space->get_hi(), open_width);
        else if (::left == this_dir)
            this_line = start_reformat.in_node->tamrof_line(in_format, start_reformat.char_index,  free_space->get_hi(), open_width);

        if (this_line){
            this_line->set_start_x_off(this_line->get_start_x_off() + left_margin);
            this_line_width = this_line->get_contents_width() + right_margin + left_margin;
            this_line->set_wide((short) (this_line->get_wide() + (left_margin + right_margin)));
            if (this_line_width > max_line_width)
                max_line_width = this_line_width;
            free_space = free_space->alloc_space(this_line);
            if (!top_used)
                top_used = this_line;
            bottom_used = this_line;
            lines_set++;
            if (!free_space)
                free_space = top_free;
            }
        else {
            free_space = make_blank(free_space);
            continue;
            }
        end_pos = this_line->get_end();
        start_reformat = end_pos.get_next();
        }
    
    line_cnt = (short) lines_set;
    old_line_cnt = lines_set - old_line_cnt;
    if (old_line_cnt) {
        for (adjust_page = next; adjust_page; adjust_page = adjust_page->next)
            adjust_page->fline_no += old_line_cnt;
        }

    if (max_line_width > in_format->get_max_line_width())
        in_format->set_max_line_width(max_line_width);
    else if (old_max_line_width)
        in_format->calc_max_line_width();

    format_ok = 1;
    return end_pos;
                               
}

