/**************************************************************************************
    post_filter.cxx
    Module to convert a tml model stream to a postscript stream

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

/*
.............................................................................
//
// Member Functions
//   filter
//
// Private Functions
//   get_iso
//   set_pg_coord
//   write_header
//   build_poststr
//   add_poststr
//   scale_widths
//   float_round
//   f__read
//   f__write
//
.............................................................................

Changes:

    1/28/88  hayes created from tml_post_filter w/o database calls

*/


#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_os_depends.hxx>
#include <tml/post_filter.hxx>

#define FlushPgs "\nshowpage\n"
#define ShowPg "\nrender_pg\n"
#define InitPg "\ninit_pg\n"
#define PtSetPos 1
#define XPos 3
#define YPos 5
#define FontPos 7
#define PgSzPos 5
#define XLatePos 11
#define LandModePos 14
#define MxPgPos 20
#define ColSzPos 23
#define MaxPostStr 9
#define MaxHdStr 26
#define MaxHdxStr 2
#define TabBlanks "        "

#define RoundFactor .4

char *tml_poststring[MaxPostStr] =
{
   ") ", "     ", " pt_set ", "                "," xpos ", "                ", " ypos mv ",
   "                                "," render\n"
};

   // The following has to be broken into three arrays since it causes bugs
   // with the current implementation of C++

static char *tml_postheader[MaxHdStr] =
{
   "% Postscript generated by tml=>pos filter\n\n",
   "/pt ",     "0",       " def\n",
   "/pgsize ", "       ", " def\n",
   "/xoff ",   "0",       " def\n",
   "/xlate ",  "0      ", " def\n",
   "/l_mode ", "0",       " def\n",
   "/curr_pg ","0" ,      " def\n",
   "/max_pg ", "1 " ,     " def\n",
   "/col_sz ", "     ",
" def\n\n/init_pg\n{\n   l_mode 1 eq { \n   curr_pg 0 eq\n   {0 xlate translate 90 neg rotate } if\n} if\n\n   userdict /xoff curr_pg col_sz mul put\n} def\n\n\n/courier 0 def\n/courier-bold 1 def\n/courier-oblique 2 def\n",
"/courier-boldoblique 3 def\n/times-roman 4 def\n/times-bold 5 def\n/times-italic 6 def\n/times-bolditalic 7 def\n/helvetica 8 def\n/helvetica-bold 9 def\n/helvetica-oblique 10 def\n/helvetica-boldoblique 11 def\n/symbol 12 def\n",
};


static char *tml_postheader1[MaxHdxStr] =
{
"/font_array [/Courier /Courier-Bold /Courier-Oblique /Courier-BoldOblique /Times-Roman\n/Times-Bold",
"/Times-Italic /Times-BoldItalic /Helvetica /Helvetica-Bold\n/Helvetica-Oblique /Helvetica-BoldOblique /Symbol ] def\n\n\n"
};

static char *tml_hdr0 = "/pt_set {/p exch def userdict /pt p put} def\n/xpos { xoff add } def\n/ypos { pgsize exch sub } def\n\n/ldfont { findfont pt scalefont setfont } def\n\n/mv { moveto } def\n\n";
static char *tml_hdr1 = "/last_font 999 def\n/encodevec\n[ 127 /Aacute 128 /Acircumflex 129 /Adieresis 130 /Agrave 131 /Aring 132 /Atilde\n133 /Ccedilla 134 /Eacute 135 /Ecircumflex 136 /Edieresis 137 /Egrave 138 /Iacute\n";
static char *tml_hdr2 = "139 /Icircumflex 140 /Idieresis 141 /Igrave 142 /Ntilde 143 /Oacute 144 /Ocircumflex 145 /Odieresis 146 /Ograve 147 /Otilde 148 /Scaron 149 /Uacute 150 /Ucircumflex\n";
static char *tml_hdr3 = "151 /Udieresis 152 /Ugrave 153 /Ydieresis 154 /Zcaron 155 /aacute 156 /acircumflex 157 /adieresis 158 /agrave 159 /aring 160 /atilde 176 /ccedilla 181 /eacute\n";
static char *tml_hdr4 = "190 /ecircumflex 192 /edieresis 201 /egrave 204 /iacute 209 /icircumflex 210 /idieresis 211 /igrave 212 /ntilde 213 /oacute 214 /ocircumflex 215 /odieresis\n";
static char *tml_hdr5 = "216 /ograve 217 /otilde 218 /scaron 219 /uacute 220 /ucircumflex 221 /udieresis 222 /ugrave 223 /ydieresis 224 /zcaron ] def\n/reencodedict 12 dict def\n/Reencode {\nreencodedict begin\n/newcodes exch def";
static char *tml_hdr6 = "/encfont exch def\n/basefont exch def\n/basefontdict basefont findfont def\n/newfont basefontdict maxlength dict def\nbasefontdict\n{ exch dup /FID ne\n{ dup /Encoding eq\n{exch dup length array copy\n";
static char *tml_hdr7 = "newfont 3 1 roll put}\n{exch newfont 3 1 roll put}\nifelse }\n{pop pop}\nifelse } forall\nnewfont /FontName encfont put\nnewcodes aload pop\nnewcodes length 2 idiv{ newfont /Encoding get 3 1 roll put}\n";
static char *tml_hdr8 = "repeat\nencfont newfont definefont pop\nend } def\n/render {\n/curr_font exch def\ncurr_font symbol eq {\nfont_array curr_font get ldfont show}\n{curr_font last_font eq{\n/NewFont ldfont show}\
\n{userdict /last_font curr_font put\n";
static char *tml_hdr9 = "font_array curr_font get /NewFont encodevec Reencode\n/NewFont ldfont show} ifelse\ncurrentpoint\nmv\n} ifelse } def\n/render_pg\n{\n   curr_pg max_pg eq\n   { showpage\n     userdict /curr_pg 0 put }\n";
static char *tml_hdr10 = "{ userdict /curr_pg curr_pg 1 add put } ifelse\n} def\n\n\n% Start Program\n\ninit_pg\n";
                   
const unsigned char tml_StartIso = 128;

typedef struct
{
   char   *iso_str;
   long    lgt;
} tml_iso_chr;

typedef tml_iso_chr tml_iso_map[128];
            
// Postscript rendererer iso character mapping
tml_iso_map tml_post_map =
{
{" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},
{" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},    {" ",1},
{" ",1},    {" ",1},    {" ",1},    {"\\241",4},{"\\242",4},{"\\243",4},{"\\250",4},{"\\245",4},{"|",1},    {"\\247",4},{"\\310",4},{"\\343",4},{"\\343",4},{"\\253",4},{"\\330",4},
{" ",1},    {"\\342",4},{"\\261",4},{"\\312",4},{"\\261",4},{" ",1},    {" ",1},    {"\\302",4},{"\\155",4},{"\\266",4},{"\\267",4},{"\\270",4},{" ",1},    {"\\353",4},{"\\273",4},
{" ",1},    {" ",1},    {" ",1},    {"\\277",4},{"\\202",4},{"\\177",4},{"\\200",4},{"\\204",4},{"\\201",4},{"\\203",4},{"\\341",4},{"\\205",4},{"\\211",4},{"\\206",4},{"\\207",4},
{"\\210",4},{"\\215",4},{"\\212",4},{"\\213",4},{"\\214",4},{"D",1},    {"\\216",4},{"\\222",4},{"\\217",4},{"\\220",4},{"\\223",4},{"\\221",4},{" ",1},    {"\\351",4},{"\\230",4},
{"\\225",4},{"\\226",4},{"\\227",4},{"Y",1},    {"\\122",4},{"\\373",4},{"\\236",4},{"\\233",4},{"\\234",4},{"\\240",4},{"\\235",4},{"\\237",4},{"\\361",4},{"\\260",4},{"\\311",4},
{"\\265",4},{"\\276",4},{"\\300",4},{"\\323",4},{"\\314",4},{"\\321",4},{"\\322",4},{"\\266",4},{"\\324",4},{"\\330",4},{"\\325",4},{"\\326",4},{"\\331",4},{"\\327",4},{" ",1},
{"\\371",4},{"\\336",4},{"\\333",4},{"\\334",4},{"\\335",4},{"y",1},    {"\\162",4},{"\\337",4}
};



//
//  get_iso - Return the postscript mapping of a given iso character
//
//  Input
//    byte  - Iso byte
//    lgt   - Output length
//
//  Result
//    Null if not in iso set (i.e use normal ascii) or device dependent string
//

char* post_filter :: get_iso (unsigned char byte, long* out_lgt)
{
   if (byte < tml_StartIso)
   {
      return NULL;
   }

   *out_lgt = tml_post_map[byte - tml_StartIso].lgt;
   return tml_post_map[byte - tml_StartIso].iso_str;
}



//
//  set_pg_coord - Adjust conceptual page layout dependent on mode
//                   (i.e. portrait,landscape) and number of columns.
//
//  Input
//    ph - Conceptual page head
//    pos - Where we are in the model
//    db  - Print params structure
//
//  Result if success 0
//         else
//             -1
//

long post_filter :: set_pg_coord ( phead* ph, model_pos* pos, print_db* db )
{
   long   *marg_array, *page;

   num_cols = (short) db->columns;
   pg_layout = (short) db->orientation;
       
   if (!pos->in_node)
      return -1;

   if (pg_layout == landscape)
   {
         // Landscape mode coordinate space

      delta_y = ph->get_wide();
      delta_x = ph->get_hi();
   }
   else
   {
         // Portrait mode coordinate space

      delta_y = ph->get_hi();
      delta_x = ph->get_wide();
   }

      // Only reformat if we have changes the makeup of the page size

   if ((num_cols > 1) || (pg_layout == landscape))
   {
      page = (long*)(ph->get_format()->inq_state(pos->in_node, tml_pgsize_prop));

         // Something is wrong, there is no page size property

      if (!page)
         return -1;

      page_size[0] = page[0];
      page_size[1] = page[1];

      delta_x /= num_cols;
      page[0] = delta_x;
      page[1] = delta_y;
      ph->get_format()->set_state(tml_pgsize_prop, (void*)page);
      ph->mark_reformat();
      *pos = ph->format_page(*pos);
   }
   else   // Did not change page size property
      page_size[0] = page_size[1] = -1;

   marg_array = (long*)ph->get_format()->inq_state(pos->in_node, tml_margin_prop);
                      
   max_xcoord = delta_x - ((marg_array) ? marg_array[right_marg] : 0);

   return (pos->in_node) ? 0 : -1;
}



//
//  write_header - Write header describing describing postscript variables functions.
//
//  Input
//    id - Postscript sctream
//    db  - Database object
//
//   Result : if success 0
//            else
//               -1
//
 
long post_filter :: write_header ( long* id, print_db* db )
{
   long       st =0,i ;

   pt_size = db->pointsize;

   if (pg_layout == landscape)
   {
         // We are printing in Landscape so we translate and rotate axis

      strcpy ( tml_postheader[LandModePos], "1");
      sprintf ( tml_postheader[XLatePos], "%ld", delta_x * num_cols );
   }
                                
   sprintf ( tml_postheader[PgSzPos], "%ld", delta_y );
   sprintf ( tml_postheader[MxPgPos], "%ld", num_cols - 1 );
   sprintf ( tml_postheader[ColSzPos], "%ld", delta_x );
   
   io_buff[0] = 0;
                             
      // Hack because of C++ string buf see top header

   for (i=0;i < MaxHdStr;++i)
      strcat ( &io_buff[0], tml_postheader[i]);

   for (i=0;i < MaxHdxStr;++i)
      strcat ( &io_buff[0], tml_postheader1[i]);

   st |= tml_os_write (id, io_buff, strlen (io_buff));
   st |= tml_os_write (id, tml_hdr0, strlen (tml_hdr0));
   st |= tml_os_write (id, tml_hdr1, strlen (tml_hdr1));
   st |= tml_os_write (id, tml_hdr2, strlen (tml_hdr2));
   st |= tml_os_write (id, tml_hdr3, strlen (tml_hdr3));
   st |= tml_os_write (id, tml_hdr4, strlen (tml_hdr4));
   st |= tml_os_write (id, tml_hdr5, strlen (tml_hdr5));
   st |= tml_os_write (id, tml_hdr6, strlen (tml_hdr6));
   st |= tml_os_write (id, tml_hdr7, strlen (tml_hdr7));
   st |= tml_os_write (id, tml_hdr8, strlen (tml_hdr8));
   st |= tml_os_write (id, tml_hdr9, strlen (tml_hdr9));
   st |= tml_os_write (id, tml_hdr10, strlen (tml_hdr10));

   return (st<0) ? -1 : 0;
}



//
//  add_poststr - Add postscript data string to display text
//
//  Input
//    buff - Text buffer
//    lgt - Output buffer length
//
//  Note !! This function assumes that text starts in buff[1]
//          so that it may insert a '(' into buff[0].
//

void post_filter :: add_poststr ( char* buff, long* lgt )
{
   static long unbroken_idx[] = {0,1,2,3,4,5,6,7,8};
   static long broken_idx[]   = {0,1,2,7,8};
          long *idx_ptr,size,i;
                                  
      // If we have a broken physical line do not bother using tml
      // coordinates because of round-off. Assume an absolute move
      // was done after previous show

   idx_ptr = ((broken_line)&&(rel_move))
             ? (size=(long) sizeof(broken_idx),&broken_idx[0])
             : (size=(long) sizeof(unbroken_idx),&unbroken_idx[0]);
              
   size /= (long) sizeof (long);
 
   buff[0] = '(';

   for (i=0;i < size;++i)
      strcat ( buff, tml_poststring[*(idx_ptr++)]);

   *lgt = strlen (buff);
}



//
//  build_poststr - Build a postscript data string to display text
//
//  Input
//    f_tbl - Font table
//    db    - Tml-Print info
//    x    - x coordinate to render text
//    y    - y coordinate to render text
//

void post_filter :: build_poststr ( font_tbl* f_tbl, print_db* db, long x, long y )
{
   char    *base_name,*font_name;
   long     lgt, i;

   if ((base_name = strrchr (font_name=f_tbl->get_name(), '/')) == NULL)
      base_name = font_name;
   else
      ++base_name;

   lgt = strlen (base_name);

   for (i=0;i < lgt;++i)
   {
         // Only copy base name of font w/o point size and extensions
         // ( e.g. /sys/dm/fonts/courier10.a => courier

      if ( isdigit(base_name[i]) || (base_name[i] == '.') )
         break;
      else
         tml_poststring [FontPos][i] = base_name[i];
   }
       
   tml_poststring[FontPos][i] = 0;

      // Insert ascii interpretation of x,y coordinates and pt size setting into postring
                                                       
   sprintf (tml_poststring[PtSetPos], "%d", db->pointsize);
   sprintf (tml_poststring[XPos], "%d", x);
   sprintf (tml_poststring[YPos], "%d", y);
}

//
//  float_round - Round a given integer using float arithmetric
//

long post_filter::float_round (long in_int, long div, long mult)
{
   register float   in_flt;

   in_flt = (float)in_int/(float)div * (float)mult;
   in_int = (long)in_flt;

   if (in_flt - in_int >= RoundFactor)
      ++in_int;
                    
   return in_int;
}


//
//  scale_widths - Scale the widths in width table by desried pt size
//
//  Input
//     mod - Tml model
//     db  - Tml-Print info data structure
//

void post_filter:: scale_widths (model* mod, print_db* db)
{
   register ft_list    *fl;
            long        i,entries,in_point,bl,ht;
            short      *wt;

   if ((fl=mod->get_fonts()) == NULL)
      return;

   do
   {
      fl->inq_font()->set_icspace(0) ;
      in_point = fl->inq_font()->get_point_size() ;

      wt = fl->inq_font()->get_width_table((unsigned long*) &entries)->get_widths();
      if ((in_point != db->pointsize) && wt)
      {
         bl = float_round (fl->inq_font()->get_baseline(), in_point,
                           db->pointsize);
         
         fl->inq_font()->set_baseline (bl);

         ht = float_round (fl->inq_font()->get_hi(), in_point,
                           db->pointsize);
         
         fl->inq_font()->set_hi (ht);

         for (i=0;i<entries;++i)
            wt[i] = (short) float_round (wt[i], in_point, db->pointsize);
      }
   }
   while (fl=fl->inq_next());
}



//
//  f__read - Read in goo from the model
//
//  Input
//    buff - Buffer to load with input data
//    fid  - Goo node to read from
//    lgt -
//
//   Result : Number of goo characters read
//

long post_filter :: f__read (char* buff, model_pos* this_pos, long lgt)
{
   char*   bf;
   char*   iso;
   char*   work_ptr = buff;
   long    spec_chars = 0,iso_lgt, i;
                                        
   if ((bf = (char*)((model_pos*)this_pos)->in_node->get_contents ()) == NULL)
      return 0;

   bf += ((model_pos*)this_pos)->char_index;
     
      // I know this is gross but, Postscript will not handle any of these
      // so-called special characters unless they are escaped.. ARGH !!

   for (i=0;i<lgt;++i)
   {
      if ((iso=get_iso (bf[i], &iso_lgt)) != NULL)
      {
         strcpy (work_ptr, iso);
         spec_chars += iso_lgt - 1;   // Subtract one for actual character
         work_ptr += iso_lgt;
      }
      else
      {
         if ( (bf[i] > 128  || bf[i] == '(' || bf[i] == ')'  || bf[i] == '<' ) ||
            (bf[i] == '>' || bf[i] == '[' || bf[i] == ']' || bf[i] == '{')   ||
            (bf[i] == '}' || bf[i] == '\\' || bf[i] == '/' || bf[i] == '%') )
         {
            *work_ptr = '\\';
            ++spec_chars;
            ++work_ptr;
         }
 
         if (bf[i] == '\t')
         {
            strcpy (work_ptr, TabBlanks);
            work_ptr += strlen (TabBlanks);
            spec_chars += strlen (TabBlanks) - 1;
         }
         else
         {
            *work_ptr = bf[i];
            ++work_ptr;
         }
      }

   }
  
   *work_ptr = 0;
   return lgt + spec_chars;
}



//
//  f__write - Write data to output stream
//
//  Input
//    buff - Buffer to to disk
//    fid  - File identifier
//    lgt -
//
//   Result : if success  0
//            else
//               -1
//

long post_filter :: f__write (char* buff, long* fid, long lgt)
{
      // No need to pay attention to new lines or form feeds

   if ( (buff[lgt - 1] == '\n') || (buff[lgt - 1] == '\f') )
      buff[lgt - 1] = ' ';
          
      // Insert postscript data around input text

   add_poststr ( buff, &lgt );

   return (tml_os_write (fid, buff, lgt) == -1) ? -1 : 0;
}



//
//  filter - Convert input tml stream to a postscript stream
//
//  Input
//    tml_mod - Input Tml model
//    post_id - Output id of postscript
//    db      - Print parameter structure
//
//   Result : Success->0
//            Failure->-1
//

long post_filter :: filter (model* tml_mod, void* post_id, print_db* db)
{
   model_pos    this_pos,next_pos;
   format*      this_format = new format;
   phead*       this_page;
   tline*       tl;
   cline*       cl;
   long         line_cnt, char_cnt, st=0, prev_y, *out_id,
                wrap = 0, *wrap_ptr, next_x;
              
   out_id = (long*)post_id;

   this_format->attach(tml_mod);
   this_page = this_format->get_first_page();
   this_pos.in_node = tml_mod->get_root();
   this_pos.char_index = -1;

   this_page->format_page (this_pos);

   scale_widths (tml_mod, db);

      // Adjust page layout for mode and number of columns

   if ((st = set_pg_coord ( this_page, &this_pos, db )) != 0)
      return st;

      // Write postscript header out to stream

   if ((st = write_header ( out_id, db )) != 0)
      return st;
                                 
   next_pos = this_pos;

   if ((wrap_ptr = (long*)this_page->get_format()->inq_state(this_pos.in_node, tml_wrap_prop)) != NULL)
      wrap = *wrap_ptr;

   do
   {
      cl = this_page->get_tline ();

      line_cnt = this_page->get_line_cnt();
                                  
      // Loop, rendering all conceptual lines that will fit on this page
                                   
      prev_y = -1;

      while ((line_cnt) || ((cl) && (line_cnt == 0) && (prev_y == cl->get_uly()) ))
      {
        if (cl->get_type() == cl_text)
        {
           tl = (tline*)cl;
           this_pos = tl->get_start();

           broken_line = (prev_y == tl->get_uly ());

           if (!broken_line)
              next_x = tl->get_ulx() + tl->get_text_width() + tl->get_start_x_off();

           rel_move = (next_x == tl->get_ulx() + tl->get_start_x_off());

           build_poststr (tl->get_font(), db, tl->get_ulx()+tl->get_start_x_off(), tl->get_uly());
                            
              // Read in the conceptual line goo, use [1] so we can add a '(' for
              // the postscript line in f__write..
           
           char_cnt = (!wrap) ? tl->trunc_line(max_xcoord) : tl->get_count ();

              // If goo runs off the page and wrap is off truncate           

           if (char_cnt)
           {
              char_cnt = f__read (&io_buff[1], &this_pos, char_cnt);
           
                 // Write out goo with postscript data attached

              if ((st=f__write (&io_buff[0], out_id, ++char_cnt)) != 0)
                 return st;
           }

           if (!broken_line)
              --line_cnt;   // Only update line count on text lines !!!
           else
              next_x += tl->get_text_width();

           prev_y = tl->get_uly ();
        }
            
        do
        {
           cl = (cline*)cl->get_next();
        }
        while ((cl) && (cl->get_type() != cl_text));

      }
                        
         // We have reached the end of the conceptual page
 
      if (this_page = this_page->next_page())
      {
         this_pos = next_pos.get_next ();
         next_pos = this_page->format_page(this_pos);
      }

      if ((next_pos.in_node) && (this_page))
      {
         strcpy (&io_buff[0], ShowPg);
         strcat (&io_buff[0], InitPg);
      }
      else
         strcpy (&io_buff[0], FlushPgs);

      if ((st=tml_os_write (out_id, &io_buff[0], strlen(io_buff)) ) == -1)
         return st;
   }
   while ((next_pos.in_node) && (this_page));             

      // Set the page size property back to what it was if it was
      // changed

   if ((page_size[0] != -1) && (page_size[1] != -1))
      this_format->set_state(tml_pgsize_prop, (void*)page_size);

   tml_os_rewind (out_id);   // Rewind output stream

      // Throw away dynamic space

   delete this_format;

   return st;
}
