/**************************************************************************************
    prop_traits.cxx
    traits for the property class of the modelling component
      in the portable version of the text management library

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>
   
#define NULL 0

extern prop_func_ent* get_prop_func_tbl_entry(long);
extern void set_prop_func_tbl_entry(long, prop_func_ent*);

/*basic constructor for a property object*/
property::property(model *to_model, long property_size)
{
    register property *new_prop;
    property *prev_prop;

    if (!property_size)
        property_size = (long) sizeof(property);
    if (!this){
        if (to_model){
            new_prop = to_model->get_current_oheap()->get_tf_prop(property_size, prev_prop);
            if (new_prop) {
                if (NULL == prev_prop)
                    to_model->get_current_oheap()->tf_prop = new_prop->locator.next;
                else
                    prev_prop->locator.next = new_prop->locator.next;
                }
            else {
                new_prop = (property *)(to_model->get_current_oheap()->alloc(property_size));
                if (!new_prop)
                    return;
                }
            this = new_prop;
            }
        else {
            new_prop = (property  *)(new char [(long) sizeof(property)]);
            if (!new_prop)
                return;
            else
                this = new_prop;
            }
        }
    else
        this = this;
    locator.in_model = to_model;
    ref_count = 1;
    index = undefined_prop;
    entry_arg = exit_arg = 0;
    act_value = 0;
}

                                                           
/*destructor for a property object*/
property::~property()
{
    if (locator.in_model){
        locator.in_model->current_oheap->free(this);
        this = 0;
        }
}

/*copy a property to the indicated model*/
property* property::copy(model *to_model)
{
    register property   *return_prop;
    register argument   *new_args;
                                                                  
    if (!to_model)
        to_model = locator.in_model;

    return_prop = to_model? to_model->make_prop() : new property;
    if (!return_prop) return 0;

    return_prop->index = index;
    if (entry_arg){
        new_args = entry_arg->copy_all(to_model);
        return_prop->attach_entry(new_args);
        }
    if (exit_arg){
        new_args = exit_arg->copy_all(to_model);
        return_prop->attach_exit(new_args);
        }
    return(return_prop);
}
                                                                 
/*duplicate a property in its model*/
property* property::dup()
{
    register property *return_prop;
                                                                  
    return_prop = locator.in_model? locator.in_model->make_prop() : new property;
    if (!return_prop) return 0;

    return_prop->index = index;
    if (entry_arg)
        return_prop->attach_entry(entry_arg);
    if (exit_arg)
        return_prop->attach_exit(exit_arg);
    return(return_prop);
}

/*attach an entry argument vector to a property*/
void property::attach_entry(argument *entry_vec)
{
    register argument *attach_vec,
                      *hold_vec;

    history  *past;
    long      has_hist;
    long      is_active;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    // If entry_vec not in same model as "this", copy to this model
    // NOTE: This block must precede block setting history

    if (locator.in_model != entry_vec->in_model)
        attach_vec = entry_vec->copy_all(locator.in_model);
    else
        attach_vec = entry_vec;

    if (entry_arg)
        detach_entry();

    // NOTE: This block must follow block copying entry_vec to this
    //   model so history will be set properly

    if (NULL != locator.in_model) {
        past = locator.in_model->get_past();
        if ((has_hist = past->has_hist()) && (is_active = past->is_active())) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_ptr_arg(past, attach_vec);
            this_atom = new atom(past, (WHOLE_PTR) &property::detach_entry, (WHOLE_PTR) &property::attach_entry,
                                 arg1, NULL, arg1, arg2, NULL);
            past->add_atom(this_atom);
            }
        }

    entry_arg = attach_vec;

    if (is_active || (0 == has_hist)) {
        for(hold_vec = attach_vec; hold_vec != 0; hold_vec = hold_vec->next)
            hold_vec->ref_count++;
        }
}

/*attach an exit argument vector to a property*/
void property::attach_exit(argument *exit_vec)
{
    register argument *attach_vec,
                      *hold_vec;

    history  *past;
    long      has_hist;
    long      is_active;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    // If exit_vec not in same model as "this", copy to this model
    // NOTE: This block must precede block setting history

    if (locator.in_model != exit_vec->in_model)
        attach_vec = exit_vec->copy_all(locator.in_model);
    else
        attach_vec = exit_vec;

    if (exit_arg)
        detach_exit();

    // NOTE: This block must follow block copying exit_vec to this
    //   model so history will be set properly

    if (NULL != locator.in_model) {
        past = locator.in_model->get_past();
        if ((has_hist = past->has_hist()) && (is_active = past->is_active())) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_ptr_arg(past, attach_vec);
            this_atom = new atom(past, (WHOLE_PTR) &property::detach_exit, (WHOLE_PTR) &property::attach_exit,
                                 arg1, NULL, arg1, arg2, NULL);
            past->add_atom(this_atom);
            }
        }

    exit_arg = attach_vec;

    if (is_active || (0 == has_hist)) {
        for(hold_vec = attach_vec; hold_vec != 0; hold_vec = hold_vec->next)
            hold_vec->ref_count++;
        }
}

/*detach an entry argument vector from a property*/
void property::detach_entry()
{
    register argument *current_arg,
                      *next_arg;

    register prop_func_pt   exit_func;
    register void          *exit_val;
    register long           tbl_index;
    register prop_func_ent *this_entry;
    history  *past;
    long      has_hist = 0;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    if (NULL != locator.in_model) {
        past = locator.in_model->get_past();
        if ((has_hist = past->has_hist()) && past->is_active()) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_ptr_arg(past, entry_arg);
            this_atom = new atom(past, (WHOLE_PTR) &property::attach_entry, (WHOLE_PTR) &property::detach_entry,
                                 arg1, arg2, NULL, arg1, NULL);
            past->add_atom(this_atom);
            }
        }

    if (0 == has_hist) {
        for(current_arg = entry_arg; current_arg != 0; current_arg = next_arg){
            next_arg = current_arg->next;
            if (!--current_arg->ref_count)
                delete current_arg;
            }
        }

    /*can't deactivate if it isn't in a model*/
    if (!locator.in_model)
        return;

    tbl_index = ((((unsigned long) index) >> PROP_HASH_SHIFT) & PROP_HASH_MASK);
    for (this_entry = get_prop_func_tbl_entry(tbl_index); this_entry;
            this_entry = this_entry->get_next())
        if (this_entry->get_index() == index)
            break;
            ;
    if (!this_entry) return;
    exit_func = this_entry->get_exit();
    if (exit_func)
        if (exit_arg)
            exit_val = (*exit_func)(exit_arg);
        else if (entry_arg)
            exit_val = (*exit_func)(entry_arg);
    entry_arg = 0;
}

/*detach an exit argument vector from a property*/
void property::detach_exit()
{
    register argument *current_arg,
                      *next_arg;

    history  *past;
    long      has_hist = 0;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    if (NULL != locator.in_model) {
        past = locator.in_model->get_past();
        if ((has_hist = past->has_hist()) && past->is_active()) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_ptr_arg(past, exit_arg);
            this_atom = new atom(past, (WHOLE_PTR) &property::attach_exit, (WHOLE_PTR) &property::detach_exit,
                                 arg1, arg2, NULL, arg1, NULL);
            past->add_atom(this_atom);
            }
        }
    if (0 == has_hist) {
        for(current_arg = exit_arg; current_arg != 0; current_arg = next_arg){
            next_arg = current_arg->next;
            if (!--current_arg->ref_count)
                delete current_arg;
            }
        }

    exit_arg = 0;
}

/*routine to activate a property */
void property::activate()
{
    register prop_func_pt   entry_func;
    register long           tbl_index;
    register prop_func_ent *this_entry;
                                       
    /*can't activate if it isn't in a model*/
    if (!locator.in_model)
        return;

    if (act_value){
        if ((locator.in_model->get_state())->push_prop(index,act_value) < 0)
            (locator.in_model->get_state())->push_prop(index,0);
        return;
        }

    tbl_index = ((((unsigned long) index) >> PROP_HASH_SHIFT) & PROP_HASH_MASK);
    for (this_entry = get_prop_func_tbl_entry(tbl_index); this_entry;
            this_entry = this_entry->get_next())
        if (this_entry->get_index() == index)
            break;
            ;
    if (!this_entry){
        this_entry = new prop_func_ent;
        if (!this_entry) return;
        this_entry->set_next(get_prop_func_tbl_entry(tbl_index));
        set_prop_func_tbl_entry(tbl_index, this_entry);
        this_entry->set_index(index);
        }

    if (!this_entry->get_loaded())
        this_entry->load();

    entry_func = this_entry->get_entry();
    if (entry_func) {
        act_value = (*entry_func)(entry_arg);
        if ((locator.in_model->get_state())->push_prop(index,act_value) < 0)
            (locator.in_model->get_state())->push_prop(index,0);
        }
}

/*routine to deactivate a property*/
void property::deactivate()
{

    /*can't deactivate if it isn't in a model*/
    if (!locator.in_model)
        return;

    (locator.in_model->get_state())->pop_prop(index);

}

/*copy a property to the indicated model*/
property* font_property::copy(model *to_model)
{
    register font_property *return_prop;
    register argument      *new_args;
                                                                  
    if (!to_model)
        to_model = locator.in_model;

    return_prop = (font_property*) (to_model? to_model->make_prop((long) sizeof(font_property)) : new font_property);
    if (!return_prop) return 0;

    return_prop->set_index(index);
    if (entry_arg) {
        new_args = get_entry_av()->copy_all(to_model);
        return_prop->attach_entry(new_args);
        }
    if (exit_arg) {
        new_args = get_exit_av()->copy_all(to_model);
        return_prop->attach_exit(new_args);
        }
    return_prop->set_ftbl(ftbl);
    return(return_prop);
}

property* font_property::dup()
{
    register font_property *return_prop;
                                                                  
    return_prop = (font_property*) (locator.in_model? locator.in_model->make_prop((long) sizeof(font_property)) : new font_property);
    if (!return_prop) return 0;

    return_prop->index = index;
    if (entry_arg)
        return_prop->attach_entry(entry_arg);
    if (exit_arg)
        return_prop->attach_exit(exit_arg);
    return_prop->set_ftbl(ftbl);
    return(return_prop);
}

/*routine to return the graphics system associated with a font property*/
grafsys font_property::get_graphics_system()
{
    argument *next_arg;
    long      i;

    for (i = 2, next_arg = get_entry_av(); 
      i && next_arg; 
      i--, next_arg = next_arg->get_next())
        ;
    if (next_arg)
        return (grafsys) *((long*) ((int_arg*) next_arg)->get_val());
    return 0;
}

/*routine to set the graphics system associated with a font property*/
void font_property::set_graphics_system(grafsys new_grafsys)
{
    argument *next_arg;
    int_arg  *new_arg;
    long      i;

    for (i = 2, next_arg = get_entry_av();
      i && next_arg;
      i--, next_arg = next_arg->get_next())
        ;
    if (next_arg) {
        ((int_arg*) next_arg)->set_val((long) new_grafsys);
        return;
        }
    if (locator.in_model)
        new_arg = new int_arg(locator.in_model);
    else
        new_arg = new int_arg;
    if ( ! new_arg)
        return;
    new_arg->set_val((long) new_grafsys);
    next_arg->append(new_arg);
}

