/**************************************************************************************
    read_model.cxx
    restore a tml model from disk

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_os_depends.hxx>

model *d_read_model(char *pathname, long name_len, long oflag, long mode, void* *file_handle)
{
    void *fdp;
    model *new_model;

    if (NULL == (fdp = tml_os_open(pathname, name_len, oflag, mode))) {
        printf("d_read_model(char *pathname, long name_len, long oflag, long mode, void *fdp): tml_os_open failed\n");
        return(NULL);
        }
    new_model = d_read_model(fdp);
    *file_handle = fdp;
    return(new_model);
}

model *d_read_model(char *pathname, long name_len)
{
    void *fdp;
    model *new_model;

    if (NULL == (fdp = tml_os_open(pathname, name_len, O_RDONLY, 0))) {
        printf("d_read_model(char *pathname, long name_len): tml_os_open failed\n");
        return(NULL);
        }
    new_model = d_read_model(fdp);
    if (-1 == tml_os_close(fdp))
        printf("d_read_model(char *pathname, long name_len): tml_os_close failed\n");
    return(new_model);
}

model *d_read_model(void *file_handle)
{
    long              tml_magic;              // magic number included at beginning of
                                              //   TML Version 2 and later files
    long              version_number;         // local copy of unscrambled version number
    long              to_read;                // number of bytes to read from header
    d_model_info_v4*  model_info_v4;          // basic info on model for files written by TML Versions 2 through 4
    d_model_info_v1*  model_info_v1;          // basic info on model for TML Version 1 files
    d_model_parts*    model_parts;            // basic info on model initialized from disk or,
                                              //   for TML Version 1 files, from d_model_info_v1
                                              //   structure read from disk
    long              max_heap_types,         // number of kinds of heaps in version of TML that
                                              //   wrote file being read
                      heap_type;              // heap type tracker in restore loop
    model            *new_model;              // return value on success
    long              bytes_read;             // return value from tml_os_read()
    gc_hash_tbl       hash_tbl;

    oheap_list *old_oheap;                    // heaps vectors created
    gheap_list *old_gheap;                    //   when new model
    sheap_list *old_sheap;                    //   is allocated

    new_model = new model;
    model_parts = new d_model_parts;
    model_parts->set_model(new_model);

    // disk copy of model includes a struct d_model_info, which becomes part of the
    //   struct d_model_parts model_parts that rules initialization of the new model

    // read magic number (TML Version 2 or later) or version number (TML Version 1)

    if (-1 == (bytes_read = tml_os_read(file_handle, (char*) &tml_magic, (long) sizeof(long))))
        return(NULL);

    if ((0x0100000 == tml_magic) || (0x00010000 == tml_magic)
         || (0x00000100 == tml_magic) || (0x00000001 == tml_magic)) {
            model_parts->version_number = tml_magic;

        // TML Version 1 - determine magic number from byte order of TML version number

        switch (tml_magic) {
            case 0x01000000:
                tml_magic = TML_MAGIC_0123;
                break;
            case 0x0001000:
                tml_magic = TML_MAGIC_1032;
                break;
            case 0x00000100:
                tml_magic = TML_MAGIC_2301;
                break;
            case 0x00000001:
                tml_magic = TML_MAGIC_3210;
                break;
            default:
                printf("d_read_model: not a tml file\n");
                return(NULL);
            }
        }
    else {

        // TML Version 2 and later - read TML version number
        if (-1 == (bytes_read = tml_os_read(file_handle, (char*) model_parts, (long) sizeof(long))))
                return(NULL);
        }

    switch (tml_magic) {
        case TML_MAGIC_0123:
            model_parts->their_endian = new endian_0123;
            break;
        case TML_MAGIC_1032:
            model_parts->their_endian = new endian_1032;
            break;
        case TML_MAGIC_2301:
            model_parts->their_endian = new endian_2301;
            break;
        case TML_MAGIC_3210:
            model_parts->their_endian = new endian_3210;
            break;
        default:
            printf("d_read_model: not a tml file\n");
            return(NULL);
        }

    // unscramble TML version number

    version_number =
        (long) model_parts->their_endian->my_long((unsigned long) model_parts->version_number);

    // check that the TML version of this file is within this program's competence

    if (version_number > tml_version_number) {
        printf("d_read_model: File's version number is greater than library's\n");
        return(NULL);
        }

    // get control structure information from disk ...

    // ... for TML Version 1

    if (1 == version_number) {
        model_info_v1 = new d_model_info_v1;
        to_read = (long) sizeof(d_model_info_v1) - (long) sizeof(long);
        if (-1 == (bytes_read = tml_os_read(file_handle,
                                           ((char*) model_info_v1) + (long) sizeof(long),
                                           to_read)))
            return(NULL);

        model_info_v1->unscramble(model_parts->their_endian);
        model_parts->from_model_info_v1(model_info_v1);
        max_heap_types = NUM_MODEL_HEAP_TYPES;
        }

    // ... for TML Versions 2 through 4

    else if (4 >= version_number) {
        model_info_v4 = new d_model_info_v4;
        to_read = (long) sizeof(d_model_info_v4) - (long) sizeof(long);
        if (-1 == (bytes_read = tml_os_read(file_handle,
                                           ((char*) model_info_v4) + (long) sizeof(long),
                                           to_read)))
            return(NULL);

        model_info_v4->unscramble(model_parts->their_endian);
        model_parts->from_model_info_v4(model_info_v4);
        max_heap_types = SQUEEZE_HEAP + 1;
        }

    // ... for TML Versions 5 and later

    else if (5 <= version_number) {
        model_info_v4 = NULL;
        model_info_v1 = NULL;
        to_read = (long) sizeof(d_model_info) - (long) sizeof(long);
        if (-1 == (bytes_read = tml_os_read(file_handle,
                                           ((char*) model_parts) + (long) sizeof(long),
                                           to_read)))
            return(NULL);
        ((d_model_info*) model_parts)->unscramble(model_parts->their_endian);
        max_heap_types = NUM_HEAP_TYPES;
        }

    // detach heaps allocated on model creation - for goo and string heaps,
    //   heaps read from disk will be used instead. In restoring TML Version 1
    //   files, object heaps read from disk are used; in restoring files
    //   produced by later versions, new object heaps are allocated below.

    old_oheap = new_model->get_first_oheap();
    new_model->detach(old_oheap->get_heap());
    delete old_oheap;
    old_gheap = new_model->get_first_gheap();
    new_model->detach(old_gheap->get_heap());
    delete old_gheap;
    old_sheap = new_model->get_first_sheap();
    new_model->detach(old_sheap->get_heap());
    delete old_sheap;

    // default heap sizes for each heap type

    new_model->set_oheap_size(model_parts->heap_quanta[OBJ_HEAP]);
    new_model->set_gheap_size(model_parts->heap_quanta[GOO_HEAP]);
    new_model->set_sheap_size(model_parts->heap_quanta[ST_HEAP]);

    // read in the array for restoration of virtual function table pointers

    if (1 == model_parts->version_number) {
        if (-1 == (model_parts->read_types_v1(file_handle)))
            return(NULL);
        }
    else {
        if (-1 == (model_parts->read_types(file_handle)))
            return(NULL);
        }

    // read and restore character set information

    if (3 <= model_parts->version_number)
        new_model->set_char_sets(read_ucs_binary_file(file_handle, 
                                                      &hash_tbl, 
                                                      model_parts->types, 
                                                      model_parts->their_endian));

    // for each type of heap, create the necessary arrays and read them in from disk

    model_parts->prepare_for_restore(file_handle, max_heap_types);

    // read in the heaps and attach them to the model or to d_model_parts

    for (heap_type = 0; heap_type < max_heap_types; heap_type++) {
        if (-1 == model_parts->restore_heap_array(file_handle, heap_type))
            return(NULL);
        }

    // restore pointers in the new model and unscramble
    //   two- and four-byte characters

    if (1 == model_parts->version_number)
        model_parts->restore_pointers_v1(hash_tbl);
    else
        model_parts->restore_pointers(hash_tbl);

    // delete free store objects no longer needed

    delete model_parts;
    delete model_info_v1;

    // return a pointer to the restored model

    return(new_model);
}

