/**************************************************************************************
    state_traits.cxx
    member functions for the property state class and the property state entity class

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>

/*************************************************************************************
    traits for the property state entity class
*************************************************************************************/

/*constructor for a property state entity*/
prop_t_ent::prop_t_ent()
{
    prior = next = 0;
    prop_index = undefined_prop;
    value = 0;
    is_default = 0;
}
                                                           
/*destructor for a property state entity*/
prop_t_ent::~prop_t_ent()
{
    if (prior)
        prior->next = next;
    if (next)
        next->prior = prior;
    next = prior = 0;
    prop_index = undefined_prop;
    value = 0;
}


/*************************************************************************************
    traits for the property state  class
*************************************************************************************/

/*constructor for a property state*/
prop_state::prop_state()
{
    register long i;
    
    for (i = 0; i < prop_tbl_size; i++)
        state_tbl[i] = 0;
    tf_entry = 0;
    lock = 0;
}

/*destructor for the property state class*/
prop_state::~prop_state()
{
    register long i;
    register prop_t_ent *this_one,
                                *next_one;

    for(i=0; i<prop_tbl_size; i++){
        for(this_one = state_tbl[i]; this_one; this_one = next_one){
            next_one = this_one->next;
            delete this_one;
            }
        }
    for(this_one = tf_entry; this_one; this_one = next_one){
        next_one = this_one->next;
        delete this_one;
        }
    lock = 0;
}
         

/*get an entry from the free list if there is one; if not add
    some to the free list*/
prop_t_ent* prop_state::get_entry()
{
    register prop_t_ent     *new_entry;

    if (tf_entry){
        new_entry = tf_entry;
        tf_entry = new_entry->next;
        new_entry->next = 0;
        return new_entry;
        }

    return (new prop_t_ent);
}

/*put a prop_t_ent on the free list*/
void prop_state::free_entry(prop_t_ent *to_free)
{
    to_free->prior = 0;
    to_free->next = tf_entry;
    to_free->prop_index = 0;
    to_free->value = 0;
    to_free->is_default = 0;
    tf_entry = to_free;
}

/*push a property on to the current property state*/
long prop_state::push_prop(p_index prop_ind, void *prop_value)
{
    register prop_t_ent *new_entry;
    register long              tbl_index;

    new_entry = get_entry();
    if (!new_entry) return -1;

    new_entry->prop_index = prop_ind;
    new_entry->value = prop_value;
    tbl_index = ((((unsigned long) prop_ind) >> PROP_HASH_SHIFT) & PROP_HASH_MASK);
    new_entry->next = state_tbl[tbl_index];
    if (new_entry->next)
        new_entry->next->prior = new_entry;
    state_tbl[tbl_index] = new_entry;
    return 0;
}
 
/*utility routine--find the topmost property state entry with a value and with property index in_dex*/
prop_t_ent* prop_state::find_entry(p_index in_dex)
{
    register prop_t_ent *return_ent;

    for (return_ent = state_tbl[((((unsigned long) in_dex) >> PROP_HASH_SHIFT) & PROP_HASH_MASK)];
          return_ent;
          return_ent = return_ent->next)
        if (return_ent->prop_index == in_dex)
            break;
          ;
    return return_ent;
}

/*pop a property off of the property state*/
long prop_state::pop_prop(p_index prop_ind)
{
    register prop_t_ent *current_prop_t;
    register long             tbl_index;

    current_prop_t = find_entry(prop_ind);
    if (!current_prop_t || current_prop_t->is_default) return -1;
    tbl_index = ((((unsigned long) prop_ind) >> PROP_HASH_SHIFT) & PROP_HASH_MASK);
    if (state_tbl[tbl_index] == current_prop_t)
        state_tbl[tbl_index] = current_prop_t->next;
    else {
        if (current_prop_t->prior)
            current_prop_t->prior->next = current_prop_t->next;
        if (current_prop_t->next)
            current_prop_t->next->prior = current_prop_t->prior;
        }
    free_entry(current_prop_t);
    return 0;
}

/*returns the pointer to the value of the top property with the indicated index*/
void* prop_state::inq_state(p_index prop_ind)
{
    register prop_t_ent *current_ent;

    current_ent = find_entry(prop_ind);
    return current_ent ? current_ent->value : 0;
}

/*utility routine--find the default property state entry with property index in_dex*/
prop_t_ent* prop_state::find_default(p_index in_dex)
{
    register prop_t_ent *return_default;

    for (return_default = state_tbl[((((unsigned long) in_dex) >> PROP_HASH_SHIFT) & PROP_HASH_MASK)];
          return_default;
          return_default = return_default->next)
        if ((return_default->prop_index == in_dex) && (return_default->is_default))
            break;
          ;
    return return_default;
}

/*return the value of the default setting for the property with index prop_ind*/
void* prop_state::get_default(p_index prop_ind)
{
    register prop_t_ent *default_prop;

    default_prop = find_default(prop_ind);
    return default_prop ? default_prop->value : 0;
}

/*set or reset the default value for a property*/
long prop_state::set_default(p_index prop_ind, void *def_value)
{
    register prop_t_ent *def_prop,
                                *end_prop;

    def_prop = find_default(prop_ind);       /*first, see if there is already a default value*/
    if (!def_prop) {                               /*if none was found, need to make one*/
        def_prop = new prop_t_ent;
        if (!def_prop) return -1;             //couldn't allocate--return an error
        for (end_prop = state_tbl[((((unsigned long) prop_ind) >> PROP_HASH_SHIFT) & PROP_HASH_MASK)];
              end_prop;
              end_prop = end_prop->next)
            if (!end_prop->next)
                break;
              ;                                             //find the last member of this state table entry
        if (!end_prop) state_tbl[((((unsigned long) prop_ind) >> PROP_HASH_SHIFT) & PROP_HASH_MASK)] = def_prop;    //this is the first entry
        else {
            end_prop->next = def_prop;         //add it to the entry at the end
            def_prop->prior = end_prop;
            }
        }
                      
    def_prop->prop_index = prop_ind;
    def_prop->is_default = 1;
    def_prop->value = def_value;
    return 0;
}
