/**************************************************************************************
    tml_c_interface.cxx
    definitions for functions declared in tml_c_types.h for the C language
      interface of the portable version of the text management library.
    functions not found here or in csm_c_interface.cxx are needed by 
      C-language files within the core TMLib library. They can be found 
      in tml_c_internals.cxx. If a function is defined neither here nor 
      in csm_c_interface.cxx nor in tml_c_internals.cxx, it is not exported 
      through the alternate language interfaces.

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_os_depends.hxx>

/* SPECIAL INITIALIZER FUNCTION FOR USERS OF TML AND OTHER LANGUAGES */

#ifdef __cplusplus
extern "C" {
#endif

void tml_init()
{
    /* Performs C++ initialization in library           */
    /*   when main program is in a different language   */
    /* Failure to call this function may cause infinite */
    /*   loop when program exits.                       */
#ifndef hpux
// hpux uses _main() at C++ 1.2
#ifndef __cplusplus
// suitable for Glockenspiel C++ at 1.2 and earlier
    _entry();
#else 
// suitable for AT&T C++
    _main();
#endif
#else
// suitable for AT&T C++
    _main();
#endif
}


/* STRUCTURES AND TYPEDEFS USED IN C CODE FOR COMPATIBILITY WITH C++ CODE */
/* WITH ALLOCATORS AND DEALLOCATORS FOR EACH WHERE NECESSARY */
/* THE STRUCTURES THEMSELVES ARE DEFINED IN tml_c_structs.h */

C_MODEL_RANGE *c_model_range_alloc()
{
    return((C_MODEL_RANGE *) tml_os_alloc(1, (long) sizeof(C_MODEL_RANGE)));
}

void c_model_range_free(C_MODEL_RANGE *this_c_model_range)
{
    tml_os_free((char *) this_c_model_range);
}



C_MARK_LIST *c_mark_list_alloc()
{
    return((C_MARK_LIST *) tml_os_alloc(1, (long) sizeof(C_MARK_LIST)));
}

void c_mark_list_free(C_MARK_LIST *this_c_mark_list)
{
    tml_os_free((char *) this_c_mark_list);
}



C_XY_COORD *c_xy_coord_alloc()
{
    return((C_XY_COORD *) tml_os_alloc(1, (long) sizeof(C_XY_COORD)));
}

void c_xy_coord_free(C_XY_COORD *this_c_xy_coord)
{
    tml_os_free((char*) this_c_xy_coord);
}



C_PRINT_DB *c_print_db_alloc()
{
    return((C_PRINT_DB *) tml_os_alloc(1, (long) sizeof(C_PRINT_DB)));
}

void c_print_db_free(C_PRINT_DB *this_c_print_db)
{
    tml_os_free((char*) this_c_print_db);
}



C_SEARCH_ENV *c_search_env_alloc()
{
    return((C_SEARCH_ENV *) tml_os_alloc(1, (long) sizeof(C_SEARCH_ENV)));
}

void c_search_env_free(C_SEARCH_ENV *this_search_env)
{
    tml_os_free((char *) this_search_env);
}



/* FOR CLASS MODEL_RANGE */

C_MODEL_RANGE *model_range_class_alloc()
{
    return((C_MODEL_RANGE *) new model_range);
}

void model_range_free(C_MODEL_RANGE *model_range_p)
{
    delete ((model_range *) model_range_p);
}

void model_range_set_start(C_MODEL_RANGE *model_range_p, C_MODEL_POS *new_start)
{
    ((model_range *) model_range_p)->set_start(*((model_pos *) new_start));
}

void model_range_set_end(C_MODEL_RANGE *model_range_p, C_MODEL_POS *new_end)
{
    ((model_range *) model_range_p)->set_end(*((model_pos *) new_end));
}

C_MODEL_POS *model_range_get_start(C_MODEL_RANGE *model_range_p, C_MODEL_POS *model_pos_p)
{
    model_pos pos;
    char *new_model_pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("model_range_get_start: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((model_range *) model_range_p)->get_start();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

C_MODEL_POS *model_range_get_end(C_MODEL_RANGE *model_range_p, C_MODEL_POS *model_pos_p)
{
    model_pos pos;
    char *new_model_pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("model_range_get_end: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((model_range *) model_range_p)->get_end();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

void *model_range_isolate(C_MODEL_RANGE *model_range_p)
{
    return((void *) ((model_range *) model_range_p)->isolate());
}

long model_range_contains(C_MODEL_RANGE *model_range_p, C_MODEL_POS *target_pos)
{
    return(((model_range *) model_range_p)->contains(*((model_pos *) target_pos)));
}



/* FOR CLASS MODEL_POS */

C_MODEL_POS *model_pos_class_alloc()
{
    return((C_MODEL_POS *) new model_pos);
}

void model_pos_free(C_MODEL_POS *model_pos_p)
{
    delete ((model_pos *) model_pos_p);
}

long model_pos_eq(C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    return(*((model_pos *) model_pos_p1) == *((model_pos *) model_pos_p2));
}

long model_pos_neq(C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    return(*((model_pos *) model_pos_p1) != *((model_pos *) model_pos_p2));
}

long model_pos_lt(C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    return(*((model_pos *) model_pos_p1) < *((model_pos *) model_pos_p2));
}

long model_pos_le(C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    return(*((model_pos *) model_pos_p1) <= *((model_pos *) model_pos_p2));
}

long model_pos_gt(C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    return(*((model_pos *) model_pos_p1) > *((model_pos *) model_pos_p2));
}

long model_pos_ge(C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    return(*((model_pos *) model_pos_p1) >= *((model_pos *) model_pos_p2));
}

C_MODEL_POS* model_pos_get_next(C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    model_pos pos;
    char *new_model_pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("model_pos_get_next: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((model_pos *) model_pos_p1)->get_next();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p2, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

C_MODEL_POS* model_pos_get_prior(C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    model_pos pos;
    char *new_model_pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("model_pos_get_prior: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((model_pos *) model_pos_p1)->get_prior();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p2, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

long model_pos_can_format(C_MODEL_POS *model_pos_p)
{
    return(((model_pos *) model_pos_p)->can_format());
}

void model_pos_mark_changed(C_MODEL_POS *model_pos_p)
{
    ((model_pos *) model_pos_p)->mark_changed();
}
       
void model_pos_in_place_redraw(C_MODEL_POS *model_pos_p)
{
    ((model_pos *) model_pos_p)->in_place_redraw();
}



/* FOR CLASS NODE */

void *node_class_alloc(long size, void *model_p)
{
    return((void*) new node(size, (model *) model_p));
}

void node_free(void *dead_node)
{
    delete (node*) dead_node;
}

void node_set_model(void *node_p, void *model_p)
{
    ((node *) node_p)->set_model((model *) model_p);
}

void *node_get_type(void *node_p)
{
    return(((node *) node_p)->get_type());
}

void node_set_type(void *node_p, void *type)
{
    ((node *) node_p)->set_type((node_type) type);

}

long node_get_level(void *node_p)
{
    return(((node *) node_p)->get_level());
}

long node_get_sib_order(void *node_p)
{
    return(((node *) node_p)->get_sib_order());
}

void *node_get_props(void *node_p)
{
    return(((node *) node_p)->get_props());
}

C_MODEL_RANGE *node_get_span(void *node_p, C_MODEL_RANGE *model_range_p)
{
    model_range range;
    char *new_model_range;

    if ((long) sizeof(model_range) != (long) sizeof(C_MODEL_RANGE)) {
        printf("node_get_span: sizes of class model_range and type C_MODEL_RANGE are incompatible\n");
        return(NULL);
        }
    else {
        range = ((node *) node_p)->get_span();
        new_model_range = tml_os_byte_copy((char *) &range, (char *) model_range_p, (long) sizeof(model_range));
        return((C_MODEL_RANGE *) new_model_range);
        }
}

void *node_connect_node(void *node_p1, void *node_p2, long dir)
{
    return((void *) ((node *) node_p1)->connect_node((node *) node_p2, (node_dir) dir));
}

void node_detach(void *node_p)
{
    ((node *) node_p)->detach();
}

void *node_move(void *node_p1, void *node_p2, long dir)
{
    return((void *) ((node *) node_p1)->move((node *) node_p2, (node_dir) dir));
}

void node_prop_attach(void *node_p, void *prop_vec_p)
{
    ((node *) node_p)->prop_attach((prop_vec *) prop_vec_p);
}

void node_prop_detach(void *node_p)
{
    ((node *) node_p)->prop_detach();
}

void node_exit_props(void *node_p)
{
    ((node *) node_p)->exit_props();
}

void node_enter_props(void *node_p)
{
    ((node *) node_p)->enter_props();
}

long node_can_format(void *node_p)
{
    return(((node *) node_p)->can_format());
}

long node_get_max_index(void *node_p)
{
    return(((node *) node_p)->get_max_index());
}

void *node_copy(void *node_p, void *model_p)
{
    return((void *) ((node *) node_p)->copy((model *) model_p));
}

void *node_copy_all(void *node_p, void *model_p)
{
    return((void *) ((node *) node_p)->copy_all((model *) model_p));
}

void *node_dup(void *node_p)
{
    return((void *) ((node *) node_p)->dup());
}

void *node_dup_all(void *node_p)
{
    return((void *) ((node *) node_p)->dup_all());
}

void *node_split(void *node_p, long split_point)
{
    return((void *) ((node *) node_p)->split(split_point));
}

void *node_get_next(void *node_p, long *dir)
{
    node_dir short_dir = (node_dir) *dir;
    node*    next_node;

    next_node = ((node *) node_p)->get_next(short_dir);  /* the C++ function requires a short reference argument */
    *dir = (long) short_dir;
    return next_node;
}

void *node_get_prev(void *node_p, long *dir)
{
    node_dir short_dir = (node_dir) *dir;
    node*    prev_node;

    prev_node = ((node *) node_p)->get_prev(short_dir);  /* the C++ function requires a short reference argument */
    *dir = (long) short_dir;
    return prev_node;
}
         
void *node_get_node(void *node_p, long dir)
{
    return((void *) ((node *) node_p)->get_node((node_dir) dir));
}

void node_set_node(void *node_p, long dir)
{
    ((node *) node_p)->get_node((node_dir) dir);
}

void *node_get_last(void *node_p)
{
    return((void *) ((node *) node_p)->get_last());
}

long node_attach_data(void *node_p, void *data_start, long data_count)
{
    return(((node*) node_p)->attach_data(data_start, data_count));
}

void *node_insert(void *node_p, long after_index, char *add_data, long add_count, C_MODEL_POS *model_pos_p)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("node_insert: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((node *) node_p)->insert(after_index, add_data, add_count);
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((void *) new_model_pos);
        }
}

void *node_del_cont(void *node_p, long delete_index, long delete_num, C_MODEL_POS *model_pos_p)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("node_del_cont: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((node *) node_p)->del_cont(delete_index, delete_num);
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((void *) new_model_pos);
        }
}

C_MODEL_POS *node_overstrike(void *node_p, long begin_index, char *rep_with, long over_num, C_MODEL_POS *model_pos_p)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("node_overstrike: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((node *) node_p)->overstrike(begin_index, rep_with, over_num);
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

long node_get_format_hi(void *node_p, void *format_p)
{
    return(((node *) node_p)->get_format_hi((format *) format_p));
}

void *node_format_line(void *node_p, void *format_p, long start_index, long space_hi, long space_wide, long first_concep_pixel)
{
    return((void *) ((node *) node_p)->format_line((format *) format_p, start_index, space_hi, space_wide, first_concep_pixel));
}

void *node_tamrof_line(void *node_p, void *format_p, long start_index, long space_hi, long space_wide, long first_concep_pixel)
{
    return((void *) ((node *) node_p)->tamrof_line((format *) format_p, start_index, space_hi, space_wide, first_concep_pixel));
}



/* FOR CLASS BNODE */

void *bnode_class_alloc(void *model_p)
{
    return((void *) new bnode((model *) model_p));
}

void bnode_free(void *bnode_p)
{
    delete ((bnode *) bnode_p);
}



/* FOR CLASS GNODE */

void *gnode_class_alloc(void *model_p)
{
    return((void *) new gnode((model *) model_p));
}

void gnode_free(void *gnode_p)
{
    delete ((gnode *) gnode_p);
}

void gnode_gets(void *gnode_p, char *s)
{
    *((gnode *) gnode_p) = s;
}

void *gnode_get_contents(void *gnode_p)
{
    return((void *) ((gnode *) gnode_p)->get_contents());
}

long  gnode_get_char_count(void *gnode_p)
{
    return(((gnode *) gnode_p)->get_char_count());
}

char *gnode_get_0t_contents(void *gnode_p)
{
    return(((gnode *) gnode_p)->get_0t_contents());
}

void gnode_set_char_count(void *gnode_p, long new_count)
{
    ((gnode *) gnode_p)->set_char_count(new_count);
}



/* FOR CLASS PROP_VEC */

void *prop_vec_class_alloc(void *model_p)
{
    return((void *) new prop_vec((model *) model_p));
}

void prop_vec_free(void *prop_vec_p)
{
    delete (prop_vec *) prop_vec_p;
}

void *prop_vec_get_next(void *prop_vec_p)
{
    return((void *) ((prop_vec *) prop_vec_p)->get_next());
}

void *prop_vec_detach_pvec(void *prop_vec_p1, void *prop_vec_p2)
{
    return((void *) ((prop_vec*) prop_vec_p1)->detach((prop_vec*) prop_vec_p2));
}

void *prop_vec_prepend(void *prop_vec_p1, void *prop_vec_p2)
{
    return((void *) ((prop_vec*) prop_vec_p1)->prepend((prop_vec *) prop_vec_p2));
}

void prop_vec_append(void *prop_vec_p1, void *prop_vec_p2)
{
    ((prop_vec*) prop_vec_p1)->append((prop_vec *) prop_vec_p2);
}

void *prop_vec_copy(void *prop_vec_p, void *model_p)
{
    return((void *) ((prop_vec *) prop_vec_p)->copy((model *) model_p));
}

void *prop_vec_copy_all(void *prop_vec_p, void *model_p)
{
    return((void *) ((prop_vec *) prop_vec_p)->copy_all((model *) model_p));
}

void *prop_vec_dup(void *prop_vec_p)
{
    return((void *) ((prop_vec *) prop_vec_p)->dup());
}

void *prop_vec_dup_all(void *prop_vec_p)
{
    return((void *) ((prop_vec *) prop_vec_p)->dup_all());
}

void prop_vec_attach_prop(void *prop_vec_p, void *property_p)
{
    ((prop_vec *) prop_vec_p)->attach((property *) property_p);
}

void prop_vec_detach_prop(void *prop_vec_p)
{
    ((prop_vec *) prop_vec_p)->detach();
}

void prop_vec_activate(void *prop_vec_p)
{
    ((prop_vec *) prop_vec_p)->activate();
}

void prop_vec_deactivate(void *prop_vec_p)
{
    ((prop_vec *) prop_vec_p)->deactivate();
}

long prop_vec_get_ref_count(void *prop_vec_p)
{
    return(((prop_vec *) prop_vec_p)->get_ref_count());
}

void *prop_vec_get_prop(void *prop_vec_p)
{
    return((void *) ((prop_vec *) prop_vec_p)->get_prop());
}



/* FOR CLASS PROPERTY */

void *property_class_alloc(void *model_p)
{
    return((void *) new property((model *) model_p));
}

void property_free(void *property_p)
{
    delete (property *) property_p;
}

void *property_get_index(void *property_p)
{
    return(((property *) property_p)->get_index());
}

void property_set_index(void *property_p, void *i)
{
    ((property *) property_p)->set_index((p_index) i);
}

void *property_copy(void *property_p, void *model_p)
{
    return((void *) ((property *) property_p)->copy((model *) model_p));
}

void *property_dup(void *property_p)
{
    return((void *) ((property *) property_p)->dup());
}

void *property_get_entry_av(void *property_p)
{
    return((void *) ((property *) property_p)->get_entry_av());
}

void *property_get_exit_av(void *property_p)
{
    return((void *) ((property *) property_p)->get_exit_av());
}

void property_attach_entry(void *property_p, void *argument_p)
{
    ((property *) property_p)->attach_entry((argument *) argument_p);
}

void property_attach_exit(void *property_p, void *argument_p)
{
    ((property *) property_p)->attach_exit((argument *) argument_p);
}

void property_detach_entry(void *property_p)
{
    ((property *) property_p)->detach_entry();
}

void property_detach_exit(void *property_p)
{
    ((property *) property_p)->detach_exit();
}

void property_activate(void *property_p)
{
    ((property *) property_p)->activate();
}

void property_deactivate(void *property_p)
{
    ((property *) property_p)->deactivate();
}

long property_get_ref_count(void *property_p)
{
    return(((property *) property_p)->get_ref_count());
}



/* FOR CLASS ARGUMENT */

void *argument_class_alloc(void *model_p)
{
    return((void *) new property((model *) model_p));
}

void argument_free(void *argument_p)
{
    delete (argument *) argument_p;
}

void *argument_get_model(void *argument_p)
{
    return((void *) ((argument *) argument_p)->get_model());
}

void argument_set_ival(void *argument_p, long i)
{
    ((int_arg*) argument_p)->set_val(i);
}

void argument_set_fval(void *argument_p, double f)
{
    ((float_arg*) argument_p)->set_val(f);
}

void argument_set_sval(void *argument_p, char *s, long count)
{
    ((string_arg*) argument_p)->set_val(s, count);
}

void *argument_get_next(void *argument_p)
{
    return((void *) ((argument*) argument_p)->get_next());
}

arg_type argument_get_vtype(void *argument_p)
{
    return(((argument*) argument_p)->get_vtype());
}

long argument_get_ival(void *argument_p)
{
    return(((argument*) argument_p)->get_ival());
}

double argument_get_fval(void *argument_p)
{
    return(((argument*) argument_p)->get_fval());
}

char *argument_get_sval(void *argument_p)
{
    return(((argument*) argument_p)->get_sval());
}

long argument_get_scount(void *argument_p)
{
    return(((argument*) argument_p)->get_scount());
}

void *argument_copy(void *argument_p, void *model_p)
{
    return((void *) ((argument *) argument_p)->copy((model *) model_p));
}

void *argument_copy_all(void *argument_p, void *model_p)
{
    return((void *) ((argument *) argument_p)->copy_all((model *) model_p));
}

void *argument_detach(void *argument_p1, void *argument_p2)
{
    return((void *) ((argument *) argument_p1)->detach((argument *) argument_p2));
}

void argument_append(void *argument_p1, void *argument_p2)
{
    ((argument *) argument_p1)->append((argument *) argument_p2);
}

void *argument_prepend(void *argument_p1, void *argument_p2)
{
    return((void *) ((argument *) argument_p1)->prepend((argument *) argument_p2));
}

long argument_get_ref_count(void *argument_p)
{
    return(((argument *) argument_p)->get_ref_count());
}



/* FOR CLASS HISTORY */

long history_set_hist_size(void *history_p, long size)
{
    return(((history *) history_p)->set_hist_size(size));
}

void history_end_molecule(void *history_p)
{
    ((history *) history_p)->end_molecule();
}

void history_undo_molecule(void *history_p)
{
    ((history *) history_p)->undo_molecule();
}

void history_redo_molecule(void *history_p)
{
    ((history *) history_p)->redo_molecule();
}



/* FOR CLASS MARK_LIST */

C_MARK_LIST *mark_list_class_alloc()
{
    return((C_MARK_LIST *) new mark_list);
}

void mark_list_free(C_MARK_LIST *this_mark_list)
{
    delete (mark_list *) this_mark_list;
}

C_MARK_LIST *mark_list_get_next(C_MARK_LIST *mark_list_p)
{
    return((C_MARK_LIST *) ((mark_list *) mark_list_p)->get_next());
}

C_MARK_LIST *mark_list_get_prior(C_MARK_LIST *mark_list_p)
{
    return((C_MARK_LIST *) ((mark_list *) mark_list_p)->get_prior());
}

C_MODEL_RANGE *mark_list_get_range(C_MARK_LIST *mark_list_p)
{
    return((C_MODEL_RANGE *) ((mark_list *) mark_list_p)->get_range());
}

void mark_list_set_next(C_MARK_LIST *mark_list_p, C_MARK_LIST *next_ml)
{
    ((mark_list *) mark_list_p)->set_next((mark_list *) next_ml);
}

void mark_list_set_prior(C_MARK_LIST *mark_list_p, C_MARK_LIST *prior_ml)
{
    ((mark_list *) mark_list_p)->set_prior((mark_list *) prior_ml);
}

void mark_list_set_range(C_MARK_LIST *mark_list_p, C_MODEL_RANGE *new_range)
{
    ((mark_list *) mark_list_p)->set_range(*((model_range *) new_range));
}

long mark_list_extend_mark(C_MARK_LIST *mark_list_p, C_MODEL_POS *new_start, C_MODEL_POS *new_end)
{
    return(((mark_list *) mark_list_p)->extend_mark(*((model_pos *) new_start), *((model_pos *) new_end)));
}

long mark_list_draw(C_MARK_LIST *mark_list_p)
{
    return(((mark_list *) mark_list_p)->draw());
}

long mark_list_erase(C_MARK_LIST *mark_list_p)
{
    return(((mark_list *) mark_list_p)->erase());
}



/* FOR CLASS SEARCH_ENV */

C_SEARCH_ENV *search_env_class_alloc(C_SEARCH_FUNC use_func, C_COMPILE_FUNC use_compile, C_SEARCH_NAV_FUNC use_for_next, C_SEARCH_NAV_FUNC use_for_prior, C_SEARCH_NAV_FUNC use_for_this)
{
    return((C_SEARCH_ENV *) new search_env((search_func) use_func, (compile_func) use_compile, (search_nav_func) use_for_next, (search_nav_func) use_for_prior, (search_nav_func) use_for_this));
}

void search_env_free(C_SEARCH_ENV *this_search_env)
{
    delete (search_env *) this_search_env;
}

C_SEARCH_ENV *search_env_get_next(C_SEARCH_ENV *search_env_p)
{
    return((C_SEARCH_ENV *) ((search_env *) search_env_p)->get_next());
}

void search_env_set_next(C_SEARCH_ENV *search_env_p, C_SEARCH_ENV *new_next)
{
    ((search_env *) search_env_p)->set_next((search_env *) new_next);
}

   

/* FOR CLASS MODEL */

void *model_class_alloc()
{
    return((void *) new model);
}

void model_free(void *model_p)
{
    delete (model *) model_p;
}

void *model_get_root(void *model_p)
{
    return((void *) ((model *) model_p)->get_root());
}

void *model_get_prop_pos(void *model_p)
{
    return((void *) ((model *) model_p)->get_prop_pos());
}

void *model_make_root(void *model_p)
{
    return((void *) ((model *) model_p)->make_root());
}

void *model_set_root(void *model_p, node *node_p)
{
    return((void *) ((model *) model_p)->set_root((node *) node_p));
}

    /* inquire the state of the property associated with the p_index argument */

void *model_inq_state(void *model_p, void *i)
{
    return((void *) ((model *) model_p)->inq_state((p_index) i));
}

void *model_move_prop(void *model_p, void *node_p)
{
    return((void *) ((model *) model_p)->move_prop((node *) node_p));
}

void model_set_prop_pos(void *model_p, void *node_p)
{
    ((model *) model_p)->set_prop_pos((node *) node_p);
}

void *model_make_pvec(void *model_p)
{
    return((void *) ((model *) model_p)->make_pvec());
}

void *model_make_prop(void *model_p)
{
    return((void *) ((model *) model_p)->make_prop());
}

void *model_make_int_arg(void *model_p, long i)
{
    return((void *) ((model *) model_p)->make_arg(i));
}

void *model_make_float_arg(void *model_p, double f)
{
    return((void *) ((model *) model_p)->make_arg((float) f));
}

void *model_make_string_arg(void *model_p, char *string, long i)
{
    return((void *) ((model *) model_p)->make_arg(string, i));
}

char *model_make_goo(void *model_p, long i, char *string)
{
    return((char *) ((model *) model_p)->make_goo(i, string));
}

char *model_make_obj(void *model_p, long i)
{
    return(((model *) model_p)->make_obj(i));
}

long model_d_write_model_path(void *model_p, char *pathname, long name_len)
{
    return(((model *) model_p)->d_write_model(pathname, name_len));
}

long model_d_write_model_handle(void *model_p, void *file_handle)
{
    return(((model *) model_p)->d_write_model(file_handle));
}

void *model_get_past(void *this_model)
{
    return(((model *) this_model)->get_past());
}

void *model_cut(void *this_model, void *cut_node)
{
    return((void *) ((model *) this_model)->cut((node *) cut_node));
}

void *model_paste(void *this_model, void *paste_node, void *paste_to, long paste_dir)
{
    return((void *) ((model *) this_model)->paste((node *) paste_node, (node *) paste_to, (node_dir) paste_dir));
}

C_MARK_LIST *model_get_first_mark(void *this_model)
{
    return((C_MARK_LIST *) ((model *) this_model)->get_first_mark());
}

C_MARK_LIST *model_make_mark(void *this_model, C_MODEL_RANGE *init_range)
{
    return((C_MARK_LIST *) ((model *) this_model)->make_mark(*((model_range *) init_range)));
}

void model_delete_mark(void *this_model, C_MARK_LIST *to_delete)
{
    ((model *) this_model)->delete_mark((mark_list *) to_delete);
}

long model_mark_draw(void *this_model, C_MARK_LIST *to_draw)
{
    return(((model *) this_model)->mark_draw((mark_list *) to_draw));
}

long model_mark_erase(void *this_model, C_MARK_LIST *to_erase)
{
    return(((model *) this_model)->mark_erase((mark_list *) to_erase));
}

void model_update_display(void *this_model)
{
    ((model *)(this_model))->update_display();
}

void model_push_search_env(void *this_model, C_SEARCH_ENV *new_envir)
{
    ((model *) this_model)->push_search_env((search_env *) new_envir);
}

void model_pop_search_env(void *this_model)
{
    ((model *) this_model)->pop_search_env();
}

C_MODEL_RANGE *model_search_exp(void *this_model, void *pattern, C_MODEL_RANGE *search_range, long forward, C_MODEL_RANGE *model_range_p)
{
    model_range range;
    char *new_model_range;

    if ((long) sizeof(model_range) != (long) sizeof(C_MODEL_RANGE)) {
        printf("model_search_exp: sizes of class model_range and type C_MODEL_RANGE are incompatible\n");
        return(NULL);
        }
    else {
        range = ((model *) this_model)->search_exp(pattern, *((model_range *) search_range), forward);
        new_model_range = tml_os_byte_copy((char *) &range, (char *) model_range_p, (long) sizeof(model_range));
        return((C_MODEL_RANGE *) new_model_range);
        }
}

long model_compile_exp(void *this_model, char *search_exp, long search_exp_len, void *compiled_exp, long max_comp_len, long case_sens)
{
    return(((model *) this_model)->compile_exp(search_exp, search_exp_len, compiled_exp, max_comp_len, case_sens));
}

C_MODEL_RANGE *model_search(void *this_model, char *search_exp, long search_exp_len, C_MODEL_RANGE *search_range, long forward, long case_sens, C_MODEL_RANGE *model_range_p)
{
    model_range range;
    char *new_model_range;

    if ((long) sizeof(model_range) != (long) sizeof(C_MODEL_RANGE)) {
        printf("model_search: sizes of class model_range and type C_MODEL_RANGE are incompatible\n");
        return(NULL);
        }
    else {
        range = ((model *) this_model)->search(search_exp, search_exp_len, *((model_range *) search_range), forward, case_sens);
        new_model_range = tml_os_byte_copy((char *) &range, (char *) model_range_p, (long) sizeof(model_range));
        return((C_MODEL_RANGE *) new_model_range);
        }
}



    
/* FOR CLASS DRAW_MARK */

void *draw_mark_alloc()
{
    return((void *) new draw_mark);
}

void draw_mark_free(void *this_draw_mark)
{
    delete (draw_mark *) this_draw_mark;
}

void *draw_mark_get_next(void *this_draw_mark)
{
    return((void *) ((draw_mark *) this_draw_mark)->get_next());
}

void *draw_mark_get_prior(void *this_draw_mark)
{
    return((void *) ((draw_mark *) this_draw_mark)->get_prior());
}

void *draw_mark_get_model_mark(void *this_draw_mark)
{
    return((void *) ((draw_mark *) this_draw_mark)->get_model_mark());
}
   
void draw_mark_set_draw_func(void *this_draw_mark, C_DRAW_M_DRAW new_draw)
{
    ((draw_mark *) this_draw_mark)->set_draw_func((draw_m_draw) new_draw);
}

void draw_mark_set_core_func(void *this_draw_mark, C_CORE_M_DRAW new_core)
{
    ((draw_mark *) this_draw_mark)->set_core_func((core_m_draw) new_core);
}

void draw_mark_extend(void *this_draw_mark, C_MODEL_POS *new_start, C_MODEL_POS *new_end)
{
    ((draw_mark *) this_draw_mark)->extend(*((model_pos *) new_start), *((model_pos *) new_end));
}

void draw_mark_draw(void *this_draw_mark)
{
    ((draw_mark *) this_draw_mark)->draw();
}

void draw_mark_erase(void *this_draw_mark)
{
    ((draw_mark *) this_draw_mark)->erase();
}



/* FOR CLASS PROP_FUNC_ENT */

void *prop_func_ent_class_alloc()
{
    return((void *) new prop_func_ent);
}

void prop_func_ent_free(void *prop_func_ent_p)
{
    delete (prop_func_ent *) prop_func_ent_p;
}

void prop_func_ent_set_next(void *prop_func_ent_p1, void *prop_func_ent_p2)
{
    ((prop_func_ent *) prop_func_ent_p1)->set_next((prop_func_ent *) prop_func_ent_p2);
}

void *prop_func_ent_get_next(void *prop_func_ent_p)
{
    return((void *) ((prop_func_ent *) prop_func_ent_p)->get_next());
}

void prop_func_ent_set_index(void *prop_func_ent_p, void *i)
{
    ((prop_func_ent *) prop_func_ent_p)->set_index((p_index) i);
}

void *prop_func_ent_get_index(void *prop_func_ent_p)
{
    return(((prop_func_ent *) prop_func_ent_p)->get_index());
}

void prop_func_ent_load_entry(void *prop_func_ent_p, C_PROP_FUNC_PT pfp)
{
    ((prop_func_ent *) prop_func_ent_p)->load_entry((prop_func_pt) pfp);
}

void prop_func_ent_load_exit(void *prop_func_ent_p, C_PROP_FUNC_PT pfp)
{
    ((prop_func_ent *) prop_func_ent_p)->load_exit((prop_func_pt) pfp);
}

C_PROP_FUNC_PT prop_func_ent_get_entry(void *prop_func_ent_p)
{
    return((C_PROP_FUNC_PT) ((prop_func_ent *) prop_func_ent_p)->get_entry());
}

C_PROP_FUNC_PT prop_func_ent_get_exit(void *prop_func_ent_p)
{
    return((C_PROP_FUNC_PT) ((prop_func_ent *) prop_func_ent_p)->get_exit());
}

long prop_func_ent_get_loaded(void *prop_func_ent_p)
{
    return(((prop_func_ent *) prop_func_ent_p)->get_loaded());
}

void prop_func_ent_set_loaded(void *prop_func_ent_p, long sint)
{
    ((prop_func_ent *) prop_func_ent_p)->set_loaded((short) sint);
}

void prop_func_ent_load(void *prop_func_ent_p)
{
    ((prop_func_ent *) prop_func_ent_p)->load();
}



/* FOR CLASS CLINE */

void cline_free(void *cline_p)
{
    delete (cline *) cline_p;
}

long cline_get_type(void *cline_p)
{
    return((long) ((cline *) cline_p)->get_type());
}

long cline_get_ulx(void *cline_p)
{
    return(((cline *) cline_p)->get_ulx());
}

long cline_get_uly(void *cline_p)
{
    return(((cline *) cline_p)->get_uly());
}

long cline_get_lrx(void *cline_p)
{
    return(((cline *) cline_p)->get_lrx());
}

long cline_get_lry(void *cline_p)
{
    return(((cline *) cline_p)->get_lry());
}

long cline_get_hi(void *cline_p)
{
    return(((cline *) cline_p)->get_hi());
}

long cline_get_wide(void *cline_p)
{
    return(((cline *) cline_p)->get_wide());
}

void *cline_get_phead(void *cline_p)
{
    return((void *) ((cline *) cline_p)->get_phead());
}

void *cline_get_format(void *cline_p)
{
        return((void *) ((cline *) cline_p)->get_format());
}

void *cline_inq_next(void *cline_p)
{
    return((void *) ((cline *) cline_p)->inq_next());
}

void *cline_inq_prior(void *cline_p)
{
    return((void *) ((cline *) cline_p)->inq_prior());
}

void *cline_get_down(void *cline_p)
{
    return((void *) ((cline *) cline_p)->get_down());
}

void *cline_get_up(void *cline_p)
{
    return((void *) ((cline *) cline_p)->get_up());
}

void cline_set_phead(void *cline_p, void *phead_p)
{
    ((cline *) cline_p)->set_phead((phead *) phead_p);
}

void cline_set_type(void *cline_p, long type)
{
    ((cline *) cline_p)->set_type((cl_type) type);
}

void cline_set_ul_coord(void *cline_p, long upper_x, long upper_y)
{
   ((cline *) cline_p)->set_ul_coord((short) upper_x, (short) upper_y);
}

void cline_set_wide(void *cline_p, long new_width)
{
   ((cline *) cline_p)->set_wide((short) new_width);
}

void cline_set_hi(void *cline_p, long new_hi)
{
   ((cline *) cline_p)->set_hi((short) new_hi);
}

void *cline_split(void *cline_p, long new_size, long from_top, long direction)
{
    return((void *) ((cline *) cline_p)->split((short) new_size, (short) from_top, (short) direction));
}

long cline_resize(void *cline_p, long hi_delta)
{
    return(((cline *) cline_p)->resize((short) hi_delta));
}
                                       
void cline_attach(void *cline_p1, void *cline_p2, long before)
{
    ((cline *) cline_p1)->attach(((cline *) cline_p2), before);
}

void cline_detach(void *cline_p)
{
    ((cline *) cline_p)->detach();
}

void *cline_get_next(void *cline_p)
{
    return((void *) ((cline *) cline_p)->get_next());
}

void cline_set_next(void *cline_p1, void *cline_p2)
{
    ((cline *) cline_p1)->set_next((cline *) cline_p2);
}

void cline_set_prior(void *cline_p1, void *cline_p2)
{
    ((cline *) cline_p1)->set_prior((cline *) cline_p2);
}

void *cline_get_prior(void *cline_p)
{
    return((void *) ((cline *) cline_p)->get_prior());
}

C_MODEL_POS *cline_get_end(void *cline_p, C_MODEL_POS *model_pos_p)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("cline_get_end: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((cline *) cline_p)->get_end();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

long cline_get_start_x_off(void* this_cline)
{
    return(((cline*) this_cline)->get_start_x_off());
}

void cline_replace(void* this_cline, void* replace_line)
{
    ((cline *) this_cline)->replace((cline *) replace_line);
}

void cline_release(void *this_cline)
{
    ((cline *) this_cline)->release();
}

void *cline_alloc_space(void *cline_p1, void *cline_p2, long at_bottom)
{
    return((void *) ((cline *) cline_p1)->alloc_space((cline *) cline_p2, at_bottom));
}

C_XY_COORD *cline_mpos_2_xy(void *cline_p, C_MODEL_POS *model_pos_p, C_XY_COORD *xy_coord_p)
{
    char *new_xy_coord;
    xy_coord coord;

    if ((long) sizeof(xy_coord) != (long) sizeof(C_XY_COORD)) {
        printf("cline_mpos_2_xy: sizes of xy_coord and type C_XY_COORD are incompatible\n");
        return(NULL);
        }
    else {
        coord = ((cline *) cline_p)->mpos_2_xy(*((model_pos *) model_pos_p));
        new_xy_coord = tml_os_byte_copy((char *) &coord, (char *) xy_coord_p, (long) sizeof(xy_coord));
        return((C_XY_COORD *) new_xy_coord);
        }
}

C_MODEL_POS *cline_xy_2_mpos(void *cline_p, C_XY_COORD *xy_coord_p, C_MODEL_POS *model_pos_p)
{
    char *new_mod_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("cline_xy_2_mpos: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((cline *) cline_p)->xy_2_mpos(*((xy_coord *) xy_coord_p));
        new_mod_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_mod_pos);
        }
}

void *cline_draw_line(void *cline_p, view_port *view_port_p, C_SCREEN_XY *screen_xy_p, long start_index)
{
    return((void *) ((cline *) cline_p)->draw_line((view_port *) view_port_p, *((screen_xy *) screen_xy_p), start_index));
}

void *cline_reformat(void *cline_p, C_MODEL_POS *new_start, long *next_ok, long force_start)
{
    return((void *) ((cline *) cline_p)->reformat(*((model_pos *) new_start), *next_ok, force_start));
}

   

/* FOR CLASS TLINE */

void *tline_class_alloc(void *format_p)
{
    return((void *) new tline((format *) format_p));
}

void tline_free(void* tline_p)
{
    delete ((tline *) tline_p);
}

void *tline_get_right(void *tline_p)
{
    return((void *) ((tline *) tline_p)->get_right());
}

void *tline_get_left(void *tline_p)
{
    return((void *) ((tline *) tline_p)->get_left());
}

void *tline_get_font(void* tline_p)
{
    return((void *) ((tline *) tline_p)->get_font());
}

long tline_get_baseline(void *tline_p)
{
    return(((tline *) tline_p)->get_baseline());
}

long tline_get_count(void *tline_p)
{
    return(((tline *) tline_p)->get_count());
}

long tline_adjust_baseline(void *tline_p, long new_baseline_y)
{
    return(((tline *) tline_p)->adjust_baseline(new_baseline_y));
}

void tline_set_baseline(void *tline_p, long new_baseline)
{
    ((tline *) tline_p)->set_baseline(new_baseline);
}

void tline_set_font(void *tline_p, void *font_tbl_p)
{
    ((tline *) tline_p)->set_font((font_tbl *) font_tbl_p);
}

void tline_set_start(void *tline_p, C_MODEL_POS *pos)
{
    ((tline *) tline_p)->set_start(*((model_pos *) pos));
}

void tline_set_count(void *tline_p, long new_count)
{
    ((tline *) tline_p)->set_count(new_count);
}

void tline_set_text_width(void *tline_p, long new_width)
{
    ((tline *) tline_p)->set_text_width(new_width);
}

long tline_get_text_width(void *tline_p)
{
    return(((tline *) tline_p)->get_text_width());
}

long tline_get_full_text_width(void *tline_p)
{
    return(((tline *) tline_p)->get_full_text_width());
}

void tline_set_right(void *tline_p, void* new_right)
{
    ((tline *) tline_p)->set_right((tline *) new_right);
}

void tline_set_left(void *tline_p, void* new_left)
{
    ((tline *) tline_p)->set_left((tline *) new_left);
}



/* FOR CLASS PHEAD */

void *phead_class_alloc(void *format_p)
{
    return((void *) new phead((format *) format_p));
}

void phead_free(void *phead_p)
{
    delete (phead *) phead_p;
}

void *phead_get_next(void *phead_p)
{
    return((void *) ((phead *) phead_p)->get_next());
}

void *phead_get_prior(void *phead_p)
{
    return((char *) ((phead *) phead_p)->get_prior());
}

long phead_get_page_no(void *phead_p)
{
    return(((phead *) phead_p)->get_page_no());
}

long phead_get_fline_no(void *phead_p)
{
    return(((phead *) phead_p)->get_fline_no());
}

long phead_get_max_line_width(void *phead_p)
{
    return(((phead *) phead_p)->get_max_line_width());
}

long phead_get_max_used_ht(void *phead_p)
{
    return(((phead *) phead_p)->get_max_used_ht());
}

long phead_get_line_cnt(void *phead_p)
{
    return(((phead *) phead_p)->get_line_cnt());
}

C_MODEL_POS *phead_get_end(void *phead_p, C_MODEL_POS *pos_p)
{
    char *new_mod_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("phead_get_end: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((phead *) phead_p)->get_end();
        new_mod_pos = tml_os_byte_copy((char *) &pos, (char *) pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_mod_pos);
        }
}

void *phead_get_format(void *phead_p)
{
    return((void *) ((phead *) phead_p)->get_format());
}

long phead_get_format_ok(void *phead_p)
{
    return(((phead *) phead_p)->get_format_ok());
}

long phead_get_hi(void *phead_p)
{
    return(((phead *) phead_p)->get_hi());
}

long phead_get_wide(void *phead_p)
{
    return(((phead *) phead_p)->get_wide());
}

void *phead_get_freeline(void *phead_p, long height, long from_top)
{
    return((void *) ((phead *) phead_p)->get_freeline((short) height, (short) from_top));
}

void *phead_get_tline(void *phead_p)
{
    return((void *) ((phead *) phead_p)->get_tline());
}

void *phead_get_bline(void *phead_p)
{
    return((void *) ((phead *) phead_p)->get_bline());
}

long phead_is_visible(void *phead_p)
{
    return(((phead *) phead_p)->is_visible());
}

void phead_set_format(void *phead_p, void *format_p)
{
    ((phead *) phead_p)->set_format((format *) format_p);
}

void phead_set_size(void *phead_p, long hi, long wide)
{
    ((phead *) phead_p)->set_size((short) hi, (short) wide);
}

void phead_set_start_pos(void *phead_p, C_MODEL_POS *pos)
{
    ((phead *) phead_p)->set_start_pos(*((model_pos *) pos));
}

void phead_set_end_pos(void *phead_p, C_MODEL_POS *pos)
{
    ((phead *) phead_p)->set_end_pos(*((model_pos *) pos));
}

void phead_mark_reformat(void *phead_p)
{
    ((phead *) phead_p)->mark_reformat();
}

void *phead_next_page(void *phead_p, long reformat)
{
    return((void *) ((phead *) phead_p)->next_page(reformat));
}

void *phead_prior_page(void *phead_p, long reformat)
{
    return((void *) ((phead *) phead_p)->prior_page(reformat));
}

void *phead_insert_page(void *phead_p, C_MODEL_POS *start, long after)
{
    return((void *) ((phead *) phead_p)->insert_page(*((model_pos *) start), after));
}

void *phead_mpos_2_cline(void *phead_p, C_MODEL_POS *model_pos_p)
{
    return((void *) ((phead *) phead_p)->mpos_2_cline(*((model_pos *) model_pos_p)));
}

void *phead_mpos_2_hcline(void *phead_p, C_MODEL_POS *model_pos_p)
{
    return((void *) ((phead *) phead_p)->mpos_2_hcline(*((model_pos *) model_pos_p)));
}

void phead_free_clines(void *phead_p)
{
    ((phead *) phead_p)->free_clines();
}

long phead_line_place(void *phead_p, void *place_line, long at_x, long at_y)
{
    return(((phead *) phead_p)->line_place((cline *) place_line, (short) at_x, (short) at_y));
}

C_MODEL_POS *phead_format_page(void *phead_p, C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("phead_format_page: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((phead *) phead_p)->format_page(*((model_pos *) model_pos_p1));
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p2, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

C_MODEL_POS *phead_tamrof_page(void *phead_p, C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("phead_tamrof_page: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((phead *) phead_p)->tamrof_page(*((model_pos *) model_pos_p1));
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p2, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

C_MODEL_POS *phead_reformat_page(void *phead_p, C_MODEL_POS *model_pos_p1, C_MODEL_POS *model_pos_p2)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("phead_reformat_page: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((phead *) phead_p)->reformat_page(*((model_pos *) model_pos_p1));
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) model_pos_p2, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}



/* FOR CLASS FORMAT */

void *format_class_alloc()
{
    return (void *) new format;
}

void format_free(void *format_p)
{
    delete (format *) format_p;
}

long format_get_ref_count(void *format_p)
{
    return(((format *) format_p)->get_ref_count());
}

void *format_get_model(void *format_p)
{
    return((void *) ((format *) format_p)->get_model());
}

void *format_get_last_page(void *format_p)
{
    return((void *) ((format *) format_p)->get_last_page());
}

void* format_inq_state(void *format_p, void *prop_ind)
{
    return((void *) ((format *) format_p)->inq_state((p_index) prop_ind));
}

void* format_inq_state_at_node(void *format_p, void *at_node, void *prop_ind)
{
    return((void *) ((format *) format_p)->inq_state((node *) at_node, (p_index) prop_ind));

}

long format_get_page_hi(void* format_p)
{
    return(((format *) format_p)->get_page_hi());
}

long format_get_page_wide(void* format_p)
{
    return(((format *) format_p)->get_page_wide());
}

long format_get_max_line_width(void* format_p)
{
    return(((format *) format_p)->get_max_line_width());
}

long format_get_max_used_ht(void* format_p)
{
    return(((format *) format_p)->get_max_used_ht());
}

void format_set_page_size(void *format_p, long hi, long wide)
{
    ((format *) format_p)->set_page_size((short) hi, (short) wide);
}

void format_set_pages_saved(void *format_p, long save_num)
{
    ((format *) format_p)->set_pages_saved((short) save_num);
}

void format_set_first_page(void *format_p, void *phead_p)
{
    ((format *) format_p)->set_first_page((phead *) phead_p);
}

void format_set_last_page(void *format_p, void *phead_p)
{
    ((format *) format_p)->set_last_page((phead *) phead_p);
}

long format_set_state(void *format_p, void *prop_ind, void *value)
{
    return(((format *) format_p)->set_state((p_index) prop_ind, (void *) value));
}

long format_unset_state(void *format_p, void *prop_ind)
{
    return(((format *) format_p)->unset_state((p_index) prop_ind));
}

void format_attach(void *format_p, void *model_p)
{
    ((format *) format_p)->attach((model *) model_p);
}

void format_detach(void *format_p)
{
    ((format *) format_p)->detach();
}

void *format_mpos_2_phead(void *format_p, C_MODEL_POS *pos)
{
    return((void *) ((format *) format_p)->mpos_2_phead(*((model_pos *) pos)));
}

void *format_get_page(void *format_p, long page_no)
{
    return((void *) ((format *) format_p)->get_page(page_no));
}



/* FOR CLASS FONT_TBL */

void *font_tbl_class_alloc()
{
    return (void *) new font_tbl;
}

void *font_tbl_class_alloc_table(void* ftbl)
{
    return (void *) new font_tbl((vpftbl*) ftbl);
}

void font_tbl_free(void *font_tbl_p)
{
    delete (font_tbl *) font_tbl_p;
}

void *font_tbl_get_handle(void *font_tbl_p)
{
    return((void *) ((font_tbl *) font_tbl_p)->get_handle());
}

long font_tbl_get_hi(void *font_tbl_p)
{
    return(((font_tbl *) font_tbl_p)->get_hi());
}

long font_tbl_get_baseline(void *font_tbl_p)
{
    return(((font_tbl *) font_tbl_p)->get_baseline());
}

long font_tbl_get_width(void *font_tbl_p, long voidac)
{
    return(((font_tbl *) font_tbl_p)->get_width(voidac));
}

long font_tbl_get_char_size(void *font_tbl_p)
{
    return(((font_tbl *) font_tbl_p)->get_char_size());
}

long font_tbl_get_icspace(void *font_tbl_p)
{
    return(((font_tbl *) font_tbl_p)->get_icspace());
}

long font_tbl_get_leading(void *font_tbl_p)
{
    return(((font_tbl *) font_tbl_p)->get_leading());
}
            
long font_tbl_set_info(void *font_tbl_p, char *fname, long flen)
{
    return(((font_tbl *) font_tbl_p)->set_info(fname, flen));
}

long font_tbl_load_font(void *font_tbl_p)
{
    return(((font_tbl *) font_tbl_p)->load_font());
}

long font_tbl_set_leading(void *font_tbl_p, long lead_pixels)
{
    return(((font_tbl *) font_tbl_p)->set_leading(lead_pixels));
}

void font_tbl_set_icspace(void *font_tbl_p, long new_icspace)
{
    ((font_tbl *) font_tbl_p)->set_icspace(new_icspace);
}
   


/* FOR CLASS V_CURSOR */

void *v_cursor_class_alloc()
{
    return((void *) new v_cursor);
}

void v_cursor_free(void *v_cursor_p)
{
    delete ((v_cursor *) v_cursor_p);
}

void *v_cursor_get_viewport(void *v_cursor_p)
{
    return((void *) ((v_cursor *) v_cursor_p)->get_viewport());
}

C_SCREEN_XY *v_cursor_get_raw_location(void *v_cursor_p, C_SCREEN_XY* xy_p)
{
    char *new_screen_xy;
    screen_xy xy;

    if ((long) sizeof(screen_xy) != (long) sizeof(C_SCREEN_XY)) {
        printf("v_cursor_get_location: sizes of screen_xy and type C_SCREEN_XY not compatible\n");
        return(NULL);
        }
    else {
        xy = ((v_cursor *) v_cursor_p)->get_raw_loc();
        new_screen_xy = tml_os_byte_copy((char *) &xy, (char *) xy_p, (long) sizeof(screen_xy));
        return((C_SCREEN_XY *) new_screen_xy);
        }
}

C_SCREEN_XY *v_cursor_get_caged_location(void *v_cursor_p, C_SCREEN_XY* xy_p)
{
    char *new_screen_xy;
    screen_xy xy;

    if ((long) sizeof(screen_xy) != (long) sizeof(C_SCREEN_XY)) {
        printf("v_cursor_get_location: sizes of screen_xy and type C_SCREEN_XY not compatible\n");
        return(NULL);
        }
    else {
        xy = ((v_cursor *) v_cursor_p)->get_caged_loc();
        new_screen_xy = tml_os_byte_copy((char *) &xy, (char *) xy_p, (long) sizeof(screen_xy));
        return((C_SCREEN_XY *) new_screen_xy);
        }
}

void *v_cursor_get_view_pos(void *v_cursor_p)
{
    return((void *) ((v_cursor *) v_cursor_p)->get_view_pos());
}

C_MODEL_POS *v_cursor_get_model_pos(void *v_cursor_p, C_MODEL_POS *pos_p)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("v_cursor_get_model_pos: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else {
        pos = ((v_cursor *) v_cursor_p)->get_model_pos();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

C_CURSOR_DRAW v_cursor_get_draw_func(void *v_cursor_p)
{
    return((C_CURSOR_DRAW) ((v_cursor *) v_cursor_p)->get_draw_func());
}

long v_cursor_get_height(void* v_cursor_p)
{
    return ((v_cursor*) v_cursor_p)->get_height();
}

long v_cursor_get_width(void* v_cursor_p)
{
    return ((v_cursor*) v_cursor_p)->get_width();
}

void v_cursor_attach(void *v_cursor_p, void *view_port_p)
{
    ((v_cursor *) v_cursor_p)->attach((view_port *) view_port_p);
}

void v_cursor_detach(void *v_cursor_p)
{
    ((v_cursor *) v_cursor_p)->detach();
}

void v_cursor_set_model_pos(void* v_cursor_p, void *new_pos_p)
{
    ((v_cursor *) v_cursor_p)->set_model_pos(*((model_pos *) new_pos_p));
}

long v_cursor_enable(void *v_cursor_p, long force)
{
    return(((v_cursor *) v_cursor_p)->enable(force));
}

long v_cursor_disable(void *v_cursor_p, long force)
{
    return(((v_cursor *) v_cursor_p)->disable(force));
}
                                                                     
void v_cursor_set_visible(void *v_cursor_p, long is_visible)
{
    ((v_cursor *) v_cursor_p)->set_visible(is_visible);
}
                                                               
void v_cursor_set_draw_func(void *v_cursor_p, C_CURSOR_DRAW draw_func)
{
    ((v_cursor *) v_cursor_p)->set_draw_func((cursor_draw) draw_func);
}

C_SCREEN_XY *v_cursor_move(void *v_cursor_p, long dir, C_SCREEN_XY *xy_p)
{
    char *new_screen_xy;
    screen_xy xy;

    if ((long) sizeof(screen_xy) != (long) sizeof(C_SCREEN_XY)) {
        printf("v_cursor_move: sizes of screen_xy and type C_SCREEN_XY not compatible\n");
        return(NULL);
        }
    else {
        xy = ((v_cursor *) v_cursor_p)->move((vp_dir) dir);
        new_screen_xy = tml_os_byte_copy((char *) &xy, (char *) xy_p, (long) sizeof(screen_xy));
        return((C_SCREEN_XY *) new_screen_xy);
        }
}

long v_cursor_xy_set(void *v_cursor_p, C_SCREEN_XY *where)
{
    return(((v_cursor *) v_cursor_p)->set(*((screen_xy *) where)));
}

long v_cursor_mpos_set(void *v_cursor_p, C_MODEL_POS* pos_p)
{
    return (((v_cursor *)v_cursor_p)->set(*((model_pos*)pos_p)));
}



/* FOR CLASS VIEW_PORT */

void *view_port_class_alloc()
{
    return((void *) new view_port);
}

void view_port_free(void *view_port_p)
{
    delete (view_port *) view_port_p;
}


void* view_port_get_handle(void *view_port_p)
{
    return((void *) ((view_port*) view_port_p)->get_handle());
}

long view_port_get_width(void *view_port_p)
{
    return(((view_port*) view_port_p)->get_width());
}

long view_port_get_hoffset(void *view_port_p)
{
    return(((view_port*) view_port_p)->get_hoffset());
}

long view_port_get_hi(void *view_port_p)
{
    return(((view_port*) view_port_p)->get_hi());
}

void *view_port_get_bottom_line(void *view_port_p)
{
    return((void *) ((view_port*) view_port_p)->get_bottom_line());
}

long view_port_get_foreg_id(void *view_port_p)
{
    return(((view_port*) view_port_p)->get_foreg_id());
}

long view_port_get_backg_id(void *view_port_p)
{
    return(((view_port*) view_port_p)->get_backg_id());
}

void *view_port_inq_state_at_node(void *view_port_p, node *node_p, void *prop_ind)
{
    return((void *) ((view_port*) view_port_p)->inq_state((node *) node_p,   (p_index) prop_ind));
}

void *view_port_get_marks(void *view_port_p)
{
    return((draw_mark *) ((view_port *) view_port_p)->get_marks());
}

void view_port_set_handle(void *view_port_p, void *bitmap_id)
{
    ((view_port *) view_port_p)->set_handle((void *) bitmap_id);
}

long view_port_attach_format(void *view_port_p, format* format_p, long fx_size, long fy_size)
{
    return(((view_port*) view_port_p)->attach((format *) format_p, fx_size, fy_size));
}

void view_port_detach_format(void *view_port_p)
{
    ((view_port *) view_port_p)->detach_format();
}

long view_port_attach_cursor(void *view_port_p, void *v_cursor_p, C_SCREEN_XY* where)
{
    return(((view_port*) view_port_p)->attach((v_cursor *) v_cursor_p, *((screen_xy *) where)));
}

void view_port_detach_cursor(void *view_port_p)
{
    ((view_port *) view_port_p)->detach_cursor();
}

void view_port_set_top_line(void *view_port_p, void *view_pos_p)
{
    ((view_port *) view_port_p)->set_top_line((view_pos *) view_pos_p);
}

void view_port_set_bottom_line(void *view_port_p, void *view_pos_p)
{
    ((view_port *) view_port_p)->set_bottom_line((view_pos *) view_pos_p);
}

void view_port_set_foreg_id(void *view_port_p, long new_id)
{
    ((view_port *) view_port_p)->set_foreg_id(new_id);
}

void view_port_set_backg_id(void *view_port_p, long new_id)
{
    ((view_port *) view_port_p)->set_backg_id(new_id);
}
 
long view_port_set_state(void *view_port_p, void *prop_ind, void *value)
{
    return(((view_port*) view_port_p)->set_state((p_index) prop_ind, (void *) value));
}

long view_port_unset_state(void *view_port_p, void *prop_ind)
{
    return(((view_port*) view_port_p)->unset_state((p_index) prop_ind));
}

void view_port_add_mark(void *view_port_p, void *new_mark)
{
    ((view_port *) view_port_p)->add_mark((draw_mark *) new_mark);
}

void view_port_detach_mark(void *view_port_p, void *del_mark)
{
    ((view_port *) view_port_p)->detach_mark((draw_mark *) del_mark);
}


void* view_port_get_pos(void *view_port_p)
{
    return((void *) ((view_port*) view_port_p)->get_pos());
}

void view_port_free_pos(void *view_port_p, void *view_pos_p)
{
    ((view_port *) view_port_p)->free_pos((view_pos *) view_pos_p);
}

void view_port_free_all_pos(void *view_port_p)
{
    ((view_port *) view_port_p)->free_all_pos();
}

void view_port_clear(void *view_port_p, C_SCREEN_XY *ul, C_SCREEN_XY *lr)
{
    ((view_port *) view_port_p)->clear(*((screen_xy *) ul), *((screen_xy *) lr));
}

void view_port_draw(void *view_port_p, void *cline_p, C_SCREEN_XY *where)
{
    ((view_port *) view_port_p)->draw((cline *) cline_p, *((screen_xy *) where));
}

void view_port_draw_at(void *view_port_p, C_MODEL_POS *pos, C_SCREEN_XY *where)
{
    ((view_port *) view_port_p)->draw(*((model_pos *) pos), *((screen_xy *) where));
}

void view_port_adjust(void *view_port_p, long delta_x, long delta_y)
{
    ((view_port *) view_port_p)->adjust(delta_x, delta_y);
}
                       
void view_port_redraw_at(void *view_port_p, void *model_pos_p1, void *model_pos_p2, void *vpos_p)
{
    ((view_port *) view_port_p)->redraw(*((model_pos *) model_pos_p1), *((model_pos *) model_pos_p2), (view_pos *) vpos_p);
}

void* view_port_xy_2_vpos(void *view_port_p, C_SCREEN_XY *where)
{
    return((void *) ((view_port*) view_port_p)->xy_2_vpos(*((screen_xy *) where)));
}

void *view_port_cline_2_vpos(void *view_port_p, cline *cline_p)
{
    return((void *) ((view_port*) view_port_p)->cline_2_vpos((cline *) cline_p));
}

C_SCREEN_XY *view_port_cline_2_xy(void *view_port_p, cline *cline_p, C_SCREEN_XY* xy_p)
{
    char *new_screen_xy;
    screen_xy xy;

    if ((long) sizeof(screen_xy) != (long) sizeof(C_SCREEN_XY)) {
        printf("view_port_cline_2_xy: sizes of screen_xy and type C_SCREEN_XY not compatible\n");
        return(NULL);
        }
    else {
        xy = ((view_port *) view_port_p)->cline_2_xy((cline *) cline_p);
        new_screen_xy = tml_os_byte_copy((char *) &xy, (char *) xy_p, (long) sizeof(screen_xy));
        return((C_SCREEN_XY *) new_screen_xy);
        }
}

void *view_port_mpos_2_vpos(void *view_port_p, C_MODEL_POS *pos)
{
    return((void *) ((view_port*) view_port_p)->mpos_2_vpos(*((model_pos *) pos)));
}

C_SCREEN_XY *view_port_mpos_2_xy(void *view_port_p, C_MODEL_POS *pos, C_SCREEN_XY *xy_p, void *view_pos_p)
{
    char *new_screen_xy;
    screen_xy xy;

    if ((long) sizeof(screen_xy) != (long) sizeof(C_SCREEN_XY)) {
        printf("view_port_mpos_2_xy: sizes of screen_xy and type C_SCREEN_XY not compatible\n");
        return(NULL);
        }
    else {
        xy = ((view_port *) view_port_p)->mpos_2_xy(*((model_pos *) pos), (view_pos *) view_pos_p);
        new_screen_xy = tml_os_byte_copy((char *) &xy, (char *) xy_p, (long) sizeof(screen_xy));
        return((C_SCREEN_XY *) new_screen_xy);
        }
}

long view_port_is_visible(void *view_port_p, void *phead_p)
{
    return(((view_port *) view_port_p)->is_visible((phead *) phead_p));
}

long view_port_append_string(void *view_port_p, char *in_string, long count, C_MODEL_POS *model_pos_p)
{
    return(((view_port *) view_port_p)->append_string(in_string, count, *((model_pos *) model_pos_p)));
}

long view_port_prepend_string(void *view_port_p, char *in_string, long count, C_MODEL_POS *model_pos_p)
{
    return(((view_port *) view_port_p)->prepend_string(in_string, count, *((model_pos *) model_pos_p)));
}

long view_port_delete_char(void *view_port_p, C_MODEL_POS *model_pos_p)
{
    return(((view_port *) view_port_p)->delete_char(*((model_pos *) model_pos_p)));
}

long view_port_ins_before_cursor(void *view_port_p, char *in_string, long count)
{
    return(((view_port *) view_port_p)->ins_before_cursor(in_string, count));
}

long view_port_ins_after_cursor(void *view_port_p, char *in_string, long count)
{
    return(((view_port *) view_port_p)->ins_after_cursor(in_string, count));
}

long view_port_insert_prop_at_node(void *view_port_p, void *to_node, void *add_prop)
{
    return(((view_port *) view_port_p)->insert_prop((node *)to_node, (property *) add_prop));
}

long view_port_insert_prop_on_range(void *view_port_p, C_MODEL_POS *pos1, C_MODEL_POS *pos2, void *add_prop)
{
    return(((view_port *) view_port_p)->insert_prop(*((model_pos *) pos1), *((model_pos *) pos2), (property *) add_prop));
}

long view_port_delete_prop(void *view_port_p, void *del_prop, void *from_node)
{
    return(((view_port *) view_port_p)->delete_prop((property *) del_prop, (node *) from_node));
}

void *view_port_mark_region(void *view_port_p, C_MODEL_POS *pos1, C_MODEL_POS *pos2)
{
    return(((view_port *) view_port_p)->mark_region(*((model_pos *) pos1), *((model_pos *) pos2)));
}
   


/* FOR CLASS VIEW_POS */

void *view_pos_class_alloc()
{
    return (void *) new view_pos;
}

void view_pos_free(void *view_pos_p)
{
    delete (view_pos *) view_pos_p;
}

void *view_pos_get_prior(void *view_pos_p)
{
    return((void *) ((view_pos *) view_pos_p)->get_prior());
}

void *view_pos_get_next(void *view_pos_p)
{
    return((void *) ((view_pos *) view_pos_p)->get_next());
}

void *view_pos_get_up(void *view_pos_p)
{
    return((void *) ((view_pos *) view_pos_p)->get_up());
}

void *view_pos_get_down(void *view_pos_p)
{
    return((void *) ((view_pos *) view_pos_p)->get_down());
}

void *view_pos_get_view_port(void *view_pos_p)
{
    return((void *) ((view_pos *) view_pos_p)->get_viewport());
}

long view_pos_get_x(void *view_pos_p)
{
    return(((view_pos *) view_pos_p)->get_x());
}

long view_pos_get_y(void *view_pos_p)
{
    return(((view_pos *) view_pos_p)->get_y());
}

C_MODEL_POS* view_pos_get_start(void *view_pos_p, C_MODEL_POS *pos_p)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("view_pos_get_start: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else if (!view_pos_p){
        pos.in_node = 0;
        pos.char_index = -1;
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *)pos_p, (long) sizeof(model_pos));
        return(NULL);
        }
    else {
        pos = ((view_pos *) view_pos_p)->get_start();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

C_MODEL_POS* view_pos_get_end(void *view_pos_p, C_MODEL_POS *pos_p)
{
    char *new_model_pos;
    model_pos pos;

    if ((long) sizeof(model_pos) != (long) sizeof(C_MODEL_POS)) {
        printf("view_pos_get_end: sizes of class model_pos and type C_MODEL_POS are incompatible\n");
        return(NULL);
        }
    else if (!view_pos_p){
        pos.in_node = 0;
        pos.char_index = -1;
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *)pos_p, (long) sizeof(model_pos));
        return(NULL);
        }
    else {
        pos = ((view_pos *) view_pos_p)->get_end();
        new_model_pos = tml_os_byte_copy((char *) &pos, (char *) pos_p, (long) sizeof(model_pos));
        return((C_MODEL_POS *) new_model_pos);
        }
}

void view_pos_set_x(void *view_pos_p, long new_x)
{
    ((view_pos *) view_pos_p)->set_x(new_x);
}

void view_pos_set_y(void *view_pos_p, long new_y)
{
    ((view_pos *) view_pos_p)->set_y(new_y);
}

void view_pos_set_line(void *view_pos_p, void *cline_p)
{
    ((view_pos *) view_pos_p)->set_line((cline *) cline_p);
}

void view_pos_detach(void *view_pos_p)
{
    ((view_pos *) view_pos_p)->detach();
}

void view_pos_attach_port(void *view_pos_p, void *view_port_p)
{
    ((view_pos *) view_pos_p)->attach((view_port *) view_port_p);
}

long view_pos_attach_pos(void *view_pos_p1, void *view_pos_p2, long dir)
{
    return(((view_pos *) view_pos_p1)->attach((view_pos *) view_pos_p2, (vp_dir) dir));
}

void *view_pos_add(void *view_pos_p, long dir)
{
    return((void *) ((view_pos *) view_pos_p)->add((vp_dir) dir));
}

void view_pos_adjust(void *view_pos_p, long x_delta, long y_delta)
{
    ((view_pos *) view_pos_p)->adjust(x_delta, y_delta);
}

long view_pos_contains(void *view_pos_p, C_MODEL_POS *pos)
{
    return(((view_pos *) view_pos_p)->contains(*((model_pos *) pos)));
}




/* UTILITY FUNCTIONS */

void* tml_c_tml_dp_get_dflt_grafsys_type()
{
    return (void*) tml_dp_get_dflt_grafsys_type();
}

void* tml_c_tml_dp_get_dflt_vpftbl()
{
    return (void*) tml_dp_get_dflt_vpftbl();
}

void tml_c_tml_dp_set_dflt_vpftbl(void* ftbl)
{
    tml_dp_set_dflt_vpftbl((vpftbl*) ftbl);
}

void* tml_c_tml_dp_get_x_vpftbl()
{
    return tml_dp_get_x_vpftbl();
}

#ifdef apollo
void* tml_c_tml_dp_get_gpr_vpftbl()
{
    return tml_dp_get_gpr_vpftbl();
}
#endif

GEN_FUNCP tml_c_get_graf_func_from_table(void* ftbl, long index)
{
    return *(((GEN_FUNCP*) ftbl) + index);
}

void *tml_c_pvec_alloc(void *model_p)
{
    return((void *) pvec_alloc((model *) model_p));
}

void *tml_c_prop_alloc(void *model_p)
{
    return((void *) prop_alloc((model *) model_p));
}

void *tml_c_arg_alloc(void *model_p)
{
    return((void *) arg_alloc((model *) model_p));
}

void *tml_c_make_subtree(void *node_p1, void *node_p2)
{
    return((void *) make_subtree((node *) node_p1, (node *) node_p2));
}

void *tml_c_copy_component(C_MODEL_POS *pos1, C_MODEL_POS *pos2, void *model_p)
{
    return((void *) copy_component((model_pos *) pos1, (model_pos *) pos2, (model *) model_p));
}
           
void *tml_c_make_wrap_prop(long wrap_on, model *model_p)
{
    return((void *) make_wrap_prop(wrap_on, (model *) model_p));
}

void *tml_c_make_margin_prop(long left, long right, long top, long bottom, void *model_p)
{
    return((void *) make_margin_prop(left, right, top, bottom, (model *) model_p));
}

void *tml_c_make_inv_prop(long inv_on, model *model_p)
{
    return((void *) make_inv_prop(inv_on, (model *) model_p));
}

void *tml_c_make_font_prop(char *font_name, long name_len, void *model_p)
{
    return((void *) make_font_prop(font_name, name_len, (model *) model_p));
}

void *tml_c_make_font_prop_grafsys(char *font_name, long name_len, void *model_p, void *for_grafsys)
{
    return((void *) make_font_prop(font_name, name_len, (model *) model_p, (grafsys) for_grafsys));
}

void *tml_c_make_page_size_prop(long width, long depth, void *in_model)
{
    return((void *) make_page_size_prop(width, depth, (model *) in_model));
}

void *tml_c_make_tab_prop(long default_val, long num_stops, long *tab_st_list, void* in_model)
{
    return((void *) make_tab_prop((model *) in_model, default_val, num_stops, tab_st_list));
}

void tml_c_connect_prop(void *add_prop, void *to_node)
{
    connect_prop((property *) add_prop, (node *) to_node);
}

void tml_c_gc_linearize(void *model_p)
{
    gc_linearize((model *) model_p);
}

void* tml_c_prune(void *model_p)
{
    return((void *) prune((model *) model_p));
}

void* tml_c_d_read_model_handle(void *file_handle)
{
    return((void *) d_read_model(file_handle));
}

void* tml_c_d_read_model_path(char* pathname, long name_len)
{
    return((void *) d_read_model(pathname, name_len));
}

void* tml_c_d_read_model_open(char* pathname, long name_len, long oflag, long mode, void* *file_handle_p)
{
    return((void *) d_read_model(pathname, name_len, oflag, mode, file_handle_p));
}

void* tml_c_read_uasc_path(char* pathname, long name_len)
{
    return((void *) read_uasc(pathname, name_len));
}

void* tml_c_read_uasc_handle(void *file_handle)
{
    return((void*) read_uasc(file_handle));
}

long tml_c_tml_model_print(void* model_p, C_PRINT_DB *print_params, char* outfilename, long outlen)
{
    return(tml_model_print((model*) model_p, (print_db*) print_params, outfilename, outlen));
}

long tml_c_tml_file_print(char* infilename, long inlen, C_PRINT_DB *print_params, char* outfilename, long outlen)
{
    return(tml_file_print(infilename, inlen, (print_db*) print_params, outfilename, outlen));
}

void tml_c_first_pvec_hist(void* node_p, void* prop_vec_p)
{
    first_pvec_hist((node*) node_p, (prop_vec*) prop_vec_p);
}

void tml_c_first_entry_arg_hist(void* property_p, void* argument_p)
{
    first_entry_arg_hist((property*) property_p, (argument*) argument_p);
}

void tml_c_first_exit_arg_hist(void* property_p, void* argument_p)
{
    first_exit_arg_hist((property*) property_p, (argument*) argument_p);
}


/* OS-DEPENDENT UTILITY FUNCTIONS */
/* (not declared in the public C interface - use tml_os_depends.h) */

void tml_c_tml_os_qsort(char *base, unsigned long nel, unsigned long size, long (*cmp)(void*, void*))
{
    tml_os_qsort(base, nel, size, cmp);
}

void* tml_c_tml_os_open(char* file_name, long name_len, long flags, long mode)
{
    return((void *) tml_os_open(file_name, name_len, flags, mode));
}

long tml_c_tml_os_read(void* file_handle, char* buf, unsigned long nbyte)
{
    return(tml_os_read(file_handle, buf, nbyte));
}

long tml_c_tml_os_write(void* file_handle, char* buf, unsigned long nbyte)
{
    return(tml_os_write(file_handle, buf, nbyte));
}

long tml_c_tml_os_rewind(void* file_handle)
{
    return(tml_os_rewind(file_handle));
}

long tml_c_tml_os_close(void* file_handle)
{
    return(tml_os_close(file_handle));
}

long tml_c_tml_os_ptr_cmp(char* *ptr1, char* *ptr2)
{
    return(tml_os_ptr_cmp(ptr1, ptr2));
}

void* tml_c_tml_os_getpid()
{
    return(tml_os_getpid());
}

void* tml_c_tml_os_ignore_interrupts()
{
    return(tml_os_ignore_interrupts());
}

void tml_c_tml_os_p_sem(void* on_sem)
{
    tml_os_p_sem(on_sem);
}

void* tml_c_tml_os_v_sem(void* proc_id, void* on_sem)
{
    return tml_os_v_sem(proc_id, on_sem);
}

void* tml_c_tml_os_restore_interrupts(void* interrupts)
{
    return(tml_os_restore_interrupts(interrupts));
}

void* tml_c_tml_os_kill_proc(void* proc_id)
{
    return(tml_os_kill_proc(proc_id));
}

#ifdef __cplusplus
}
#endif

