/**************************************************************************************
    tml_c_structs.h
    Declarations of proxies for C++ classes used in
        C interface functions

    (c) Copyright Apollo Computer Inc., 1990 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef tml_c_structs_h
#define tml_c_structs_h

#ifdef __STDC__
#define not_vanilla_C
#endif

#ifdef __cplusplus
#define not_vanilla_C
#endif

#ifdef c_plusplus
#define not_vanilla_C
#endif

#ifdef not_vanilla_C

/* FROM HERE TO "#else", declarations for C++ and ANSI C programmers         */
/******************************************************************************
    Structures shared by C++ and C callers of C interface functions

******************************************************************************/

/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS MODEL_POS */

struct tml_c_model_pos {
    void* in_node;
    long  char_index;
    };
typedef struct tml_c_model_pos C_MODEL_POS;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS SCREEN_XY */
struct tml_c_screen_xy {
    long at_x,
         at_y;
    };
typedef struct tml_c_screen_xy C_SCREEN_XY;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS MODEL_RANGE */
struct tml_c_model_range {
    C_MODEL_POS     start,
                    end;
    };
typedef struct tml_c_model_range C_MODEL_RANGE;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS MARK_LIST */
struct tml_c_mark_list {
    struct tml_c_mark_list *next,
                           *prior;
    C_MODEL_RANGE          *mark_range;
    };
typedef struct tml_c_mark_list C_MARK_LIST;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS XY_COORD */
struct tml_c_xy_coord {
    long x_coord,
         y_coord;
    };
typedef struct tml_c_xy_coord C_XY_COORD;



/* TYPEDEF FOR A POINTER TO THE FUNCTION WHICH DRAWS THE CURSOR */

    typedef long (*tml_c_cursor_draw)(void* in_bitmap, long wherex, long wherey, long cursor_hi, long cursor_wide, long erase);
    typedef tml_c_cursor_draw C_CURSOR_DRAW;



/* TYPEDEF FOR A POINTER TO THE FUNCTION WHICH THE MARK-DRAWING FUNCTION USES TO ACTUALLY DRAW A MARK */

    typedef void (*tml_c_core_m_draw)(void* bitmap_handle, long src_ulx, long src_uly, long src_lrx, long src_lry, long dst_ulx, long dst_uly, long rop);
    typedef tml_c_core_m_draw C_CORE_M_DRAW;



/* TYPEDEF FOR A POINTER TO THE FUNCTION WHICH DRAWS A MARK */

    typedef long (*tml_c_draw_m_draw)(void* in_viewport, C_CORE_M_DRAW core_func, C_MODEL_POS start_draw, C_MODEL_POS end_draw, long erase);
    typedef tml_c_draw_m_draw C_DRAW_M_DRAW;



/* TYPEDEF FOR A POINTER TO A PROPERTY FUNCTION */

    typedef void* (*tml_c_prop_func_pt)(void* arg_vec);
    typedef tml_c_prop_func_pt C_PROP_FUNC_PT;



/* TYPEDEF FOR A POINTER TO A SEARCH FUNCTION */

    typedef long (*tml_c_search_func)(C_MODEL_RANGE search_area, C_MODEL_RANGE* found_str, void* pattern, long forward);
    typedef tml_c_search_func C_SEARCH_FUNC;



/* TYPEDEF FOR A POINTER TO A REGULAR EXPRESSION COMPILER FUNCTION */

    typedef long (*tml_c_compile_func)(char* exp, long len, void* pattern, long pattern_max, long case_sens);
    typedef tml_c_compile_func C_COMPILE_FUNC;



/* TYPEDEF FOR A POINTER TO A SEARCH NAVIGATION FUNCTION */

    typedef char* (*tml_c_search_nav_func)(C_MODEL_POS* ptr);
    typedef tml_c_search_nav_func C_SEARCH_NAV_FUNC;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS SEARCH_ENV */
struct tml_c_search_env {
    struct tml_c_search_env *next;
    C_SEARCH_FUNC search_rtn;
    C_COMPILE_FUNC compile_rtn;
    C_SEARCH_NAV_FUNC next_char_func,
                      prior_char_func,
                      this_char_func;
    };
typedef struct tml_c_search_env C_SEARCH_ENV;



/* STRUCTURE USED TO SPECIFY PRINTING PARAMETERS */
struct print_db {
    long pointsize;
    long columns;
    long orientation;
    };


/* C VERSION OF THAT STRUCTURE */
struct tml_c_print_db {
     long   pointsize;
     long   columns;
     long   page_orientation;
};
typedef struct tml_c_print_db C_PRINT_DB;

#else

/* FROM HERE TO "#endif", declarations for K&R C programmers                 */
/******************************************************************************
    Structures shared by C++ and C callers of C interface functions

******************************************************************************/

/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS MODEL_POS */

struct tml_c_model_pos {
    char *in_node;
    long  char_index;
    };
typedef struct tml_c_model_pos C_MODEL_POS;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS SCREEN_XY */
struct tml_c_screen_xy {
    long at_x,
         at_y;
    };
typedef struct tml_c_screen_xy C_SCREEN_XY;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS MODEL_RANGE */
struct tml_c_model_range {
    C_MODEL_POS     start,
                    end;
    };
typedef struct tml_c_model_range C_MODEL_RANGE;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS MARK_LIST */
struct tml_c_mark_list {
    struct tml_c_mark_list *next,
                           *prior;
    C_MODEL_RANGE          *mark_range;
    };
typedef struct tml_c_mark_list C_MARK_LIST;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS XY_COORD */
struct tml_c_xy_coord {
    long x_coord,
         y_coord;
    };
typedef struct tml_c_xy_coord C_XY_COORD;



/* TYPEDEF FOR A POINTER TO THE FUNCTION WHICH DRAWS THE CURSOR */

        /* ARGS: char *in_bitmap, long wherex, long wherey, long cursor_hi, long cursor_wide, long erase */

    typedef long (*tml_c_cursor_draw)();
    typedef tml_c_cursor_draw C_CURSOR_DRAW;



/* TYPEDEF FOR A POINTER TO THE FUNCTION WHICH THE MARK-DRAWING FUNCTION USES TO ACTUALLY DRAW A MARK */

        /* ARGS: char *bitmap_handle, long src_ulx, long src_uly, long src_lrx, long src_lry, long dst_ulx, long dst_uly, long rop */

    typedef  (*tml_c_core_m_draw)();
    typedef tml_c_core_m_draw C_CORE_M_DRAW;



/* TYPEDEF FOR A POINTER TO THE FUNCTION WHICH DRAWS A MARK */

        /* ARGS: char *in_viewport, C_CORE_M_DRAW core_func, C_MODEL_POS start_draw, C_MODEL_POS end_draw, long erase */

    typedef long (*tml_c_draw_m_draw)();
    typedef tml_c_draw_m_draw C_DRAW_M_DRAW;



/* TYPEDEF FOR A POINTER TO A PROPERTY FUNCTION */

        /* ARGS: char *arg_vec */

    typedef char *(*tml_c_prop_func_pt)();
    typedef tml_c_prop_func_pt C_PROP_FUNC_PT;



/* TYPEDEF FOR A POINTER TO A SEARCH FUNCTION */

        /* ARGS: C_MODEL_RANGE search_area, C_MODEL_RANGE* found_str, char *pattern, long forward */

    typedef long (*tml_c_search_func)();
    typedef tml_c_search_func C_SEARCH_FUNC;



/* TYPEDEF FOR A POINTER TO A REGULAR EXPRESSION COMPILER FUNCTION */

        /* ARGS: char* exp, long len, char *pattern, long pattern_max, long case_sens */

    typedef long (*tml_c_compile_func)();
    typedef tml_c_compile_func C_COMPILE_FUNC;



/* TYPEDEF FOR A POINTER TO A SEARCH NAVIGATION FUNCTION */

        /* ARGS: C_MODEL_POS* ptr */

    typedef char* (*tml_c_search_nav_func)();
    typedef tml_c_search_nav_func C_SEARCH_NAV_FUNC;



/* STRUCTURE USED TO CAPTURE CONTENTS OF A MEMBER OF CLASS SEARCH_ENV */
struct tml_c_search_env {
    struct tml_c_search_env *next;
    C_SEARCH_FUNC search_rtn;
    C_COMPILE_FUNC compile_rtn;
    C_SEARCH_NAV_FUNC next_char_func,
                      prior_char_func,
                      this_char_func;
    };
typedef struct tml_c_search_env C_SEARCH_ENV;



/* STRUCTURE USED TO SPECIFY PRINTING PARAMETERS */
struct print_db {
    long pointsize;
    long columns;
    long orientation;
    };


/* C VERSION OF THAT STRUCTURE */
struct tml_c_print_db {
     long   pointsize;
     long   columns;
     long   page_orientation;
};
typedef struct tml_c_print_db C_PRINT_DB;

#endif

#endif
/* tml_c_structs_h */
