/**************************************************************************************
    tml_dp_gpr_graph.c
    Implementation of the device-dependent service routines for
      gpr-based systems

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <string.h>

#define MAXNAMLEN 1024

#include <apollo/base.h>
#include <apollo/gpr.h>
#include <apollo/ms.h>
#include <apollo/fontn.h>
#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/tml_c_types.h>
#include <tml/tml_os_depends.h>
#include <tml/tml_dp_graph.hxx>
#include <tml/dflt_view_funcs.hxx>

#define FONT_PATH    "/sys/dm/fonts/"
#define min_fontx   1
#define BAD_FONT_NUM    (min_fontx - 1)
#define BAD_FONT_ID    ((long) NULL)
#define FALSE   0

/* service types for calls to gpr_serve_default_font_name() */
#define set_default_table          1
#define set_default_font           2
#define get_default_font           3

/* service types for calls to gpr_serve_font_handle_list() */

#define get_font_handle            1
#define release_font_handle        2
#define unload_fonts               3

/* service types for calls to gpr_serve_bm_info() */

#define make_bitmap_handle         1
#define delete_bitmap_handle       2
#define get_current_bitmap_handle  3
#define set_current_bitmap_handle  4
#define find_bitmap_handle_info    5

/*initializer for default contents of default table*/
#ifdef apollo_global
GEN_FUNCP* gpr_dflt_view_port_function_table;
#else
GEN_FUNCP gpr_dflt_view_port_function_table[] = {
    (GEN_FUNCP) &tml_dp_gpr_get_default_font,
    (GEN_FUNCP) &tml_dp_gpr_get_fhandle,
    (GEN_FUNCP) &tml_dp_gpr_load_font,
    (GEN_FUNCP) &tml_dp_gpr_get_fcharsize,
    (GEN_FUNCP) &tml_dp_gpr_get_fcharnum,
    (GEN_FUNCP) &tml_dp_gpr_get_fchar1,
    (GEN_FUNCP) &tml_dp_gpr_get_fhi,
    (GEN_FUNCP) &tml_dp_gpr_get_fpoint,
    (GEN_FUNCP) &tml_dp_gpr_get_fbline,
    (GEN_FUNCP) &tml_dp_gpr_is_fixed,
    (GEN_FUNCP) &tml_dp_gpr_get_fwidths,
    (GEN_FUNCP) &tml_dp_gpr_get_ficspace,
    (GEN_FUNCP) &tml_dp_gpr_set_ficspace,
    (GEN_FUNCP) &tml_dp_gpr_get_fspace,
    (GEN_FUNCP) &tml_dp_gpr_set_fspace,
    (GEN_FUNCP) &tml_dp_gpr_funload,
    (GEN_FUNCP) &tml_dp_gpr_frelease,
    (GEN_FUNCP) &tml_dp_gpr_make_graphic_handle,
    (GEN_FUNCP) &tml_dp_gpr_delete_graphic_handle,
    (GEN_FUNCP) &tml_dp_gpr_get_hi,
    (GEN_FUNCP) &tml_dp_gpr_get_wide,
    (GEN_FUNCP) &tml_dp_gpr_get_foreg_id,
    (GEN_FUNCP) &tml_dp_gpr_get_backg_id,
    (GEN_FUNCP) &tml_dp_gpr_get_fill_backg_id,
    (GEN_FUNCP) &tml_dp_gpr_get_fill_obj_id,
    (GEN_FUNCP) &tml_dp_gpr_set_foreg_id,
    (GEN_FUNCP) &tml_dp_gpr_set_backg_id,
    (GEN_FUNCP) &tml_dp_gpr_set_fill_backg_id,
    (GEN_FUNCP) &tml_dp_gpr_set_fill_obj_id,
    (GEN_FUNCP) &tml_dp_gpr_set_clip,
    (GEN_FUNCP) &tml_dp_gpr_clear_clip,
    (GEN_FUNCP) &tml_dp_gpr_clear,
    (GEN_FUNCP) &tml_dp_gpr_copy,
    (GEN_FUNCP) &tml_dp_gpr_blt,
    (GEN_FUNCP) &tml_dp_gpr_string_draw,
    (GEN_FUNCP) &tml_dp_gpr_string_size,
    (GEN_FUNCP) &tml_dp_gpr_damage_repair,
    (GEN_FUNCP) &tml_dp_gpr_dflt_cursor_draw,
    (GEN_FUNCP) &tml_dp_gpr_get_grafsys_obj,
    (GEN_FUNCP) &get_gpr_default_font_name,
    (GEN_FUNCP) &set_gpr_default_font_name,
    /* The null member stands in for the grafsys_type member */
    /*   of the vpftbl to which this array is cast in C++    */
    (GEN_FUNCP) NULL
};
#endif

/*function to return gpr default viewport function table */
void* tml_dp_get_gpr_genfuncp_tbl()
{
    return (void*) &gpr_dflt_view_port_function_table[0];
}

#define  NUM_SLOTS      20

/*****************************************************************************
    Structure for font management
*****************************************************************************/

struct fhandle {
    struct fhandle         *next;
    struct fhandle         *prev;
    char                   *font_name;
    short                   name_len;
    char                    loaded;
    char                    replicated;
    short                   font_num;
    short                   repli_font_num;
    short                   is_fixed;
    pinteger                version;
    short                   ref_count;
    fontn_$table_ptr_tver1  font_table;
    };

/*****************************************************************************
    Structure for bitmap management
*****************************************************************************/

struct bm_info {
    struct bm_info         *next;
    struct bm_info         *prior;
    gpr_$bitmap_desc_t     *bitmap;
    char                   *viewport_id;
    long                    size_x;
    long                    size_y;
    long                    foreground_id;  /* text color */
    long                    background_id;  /* character backing rectangle color */
    long                    fill_backg_id;  /* viewport surface color */
    long                    fill_obj_id;    /* color used to fill circles, triangles, rectangles */
    gpr_$raster_op_array_t  current_rops;
    char                   *font_id;
    gpr_$direction_t        current_text_direction;
    long                    drawn;          /* flag prevents infinite loop in initial draw call */
    };

/******************************************************************************
    Globals for font management
******************************************************************************/

/* This global need not be per-process as it is never written by any process */

/* tml_font_names:      default for table of default font names */

    char *tml_font_names[] = {
        "std",
        "std.1280bw",
        "std.19l",
        };

/******************************************************************************
    Structure for points of triangle for default cursor
******************************************************************************/

struct tml_cursor_triangle {
    gpr_$position_t tv1;
    gpr_$position_t tv2;
    gpr_$position_t tv3;
};

/* To make these globals per-process, must be declared with attribute specifier */
/*   for Apollo builds to allow use in global and inlib'd libraries */

/* tml_font_table_size:   size of current table of default font names */
/* tml_font_table:        current table of of default font names */
/* tml_default_font:      the current default font */
/* tml_fhandle_list_head: list of active font handles */
/* tml_bm_info_queue:     list of active bitmap wrappers */
/* tml_current_bitmap:    is usually the current bitmap in gpr sense */
/* tml_cursor_size:       size of cursor */
/* tml_cursor_points:     points of cursor triangle */
/* tml_trash_size:        unused return value, made global to save auto space */

#ifdef apollo
    long                           tml_font_table_size __attribute((section(tml_font_table_size))) = (long) sizeof(tml_font_names)/(long) sizeof(char*);
    char*                         *tml_font_table __attribute((section(tml_font_table))) = tml_font_names;
    char*                          tml_default_font __attribute((section(tml_default_font))) = NULL;
    struct fhandle                *tml_fhandle_list_head __attribute((section(tml_fhandle_list_head))) = NULL;
    struct bm_info                *tml_bm_info_queue __attribute((section(tml_bm_info_queue))) = NULL;
    tml_c_bitmap_id                tml_current_bitmap __attribute((section(tml_current_bitmap))) = NULL;
    gpr_$offset_t                  tml_cursor_size __attribute((section(tml_cursor_size))) = { (gpr_$coordinate_t) 0, (gpr_$coordinate_t) 0 };
    struct tml_cursor_triangle     tml_cursor_points __attribute((section(tml_cursor_points)));
    gpr_$offset_t                  tml_trash_size __attribute((section(tml_trash_size)));

#else
    long                           tml_font_table_size = (long) sizeof(tml_font_names)/(long) sizeof(char*);
    char*                         *tml_font_table = tml_font_names;
    char*                          tml_default_font = 0;
    struct fhandle                *tml_fhandle_list_head = NULL;
    struct bm_info                *tml_bm_info_queue = NULL;
    tml_c_bitmap_id                tml_current_bitmap = NULL;
    gpr_$offset_t                  tml_cursor_size = { (gpr_$coordinate_t) 0, (gpr_$coordinate_t) 0 };
    struct tml_cursor_triangle     tml_cursor_points;
    gpr_$offset_t                  tml_trash_size;
#endif

/******************************************************************************
    Functions to get/set name of default font and to set default font table
******************************************************************************/
/* maintain default font name */
char *gpr_serve_default_font_name(service, length, table, table_size)
long service;
long *length;
char* *table;
long table_size;
{
    long font_is_present();
    char full_path[MAXNAMLEN];

    long i;
    long font_len;

    switch (service) {

    case set_default_table:

        if (NULL == table) {        /* resume use of default table */
            tml_font_table_size = (long) sizeof(tml_font_names)/(long) sizeof(char*);
            tml_font_table = tml_font_names;
            }
        else {                      /* set the table */
            tml_font_table_size = table_size;
            tml_font_table = (char**) tml_c_tml_os_alloc(tml_font_table_size, (unsigned long) sizeof(char*));
            for (i = 0; i < tml_font_table_size; i++) {
                font_len = strlen(table[i]);
                tml_font_table[i] = tml_c_tml_os_alloc(font_len, (unsigned long) sizeof(char));
                tml_font_table[i] = tml_c_tml_os_byte_copy(table[i], tml_font_table[i], font_len);
                *(tml_font_table[i] + font_len) = '\0';
                }
            }
        return(NULL);

    case set_default_font:

        if (NULL == *(table + 0))
            tml_default_font = NULL;
        else {
            tml_default_font = tml_c_tml_os_alloc(*length + 1, (unsigned long) sizeof(char));
            tml_default_font = tml_c_tml_os_byte_copy(table[0], tml_default_font, *length);
            *(tml_default_font + *length) = '\0';
            }
        return(tml_default_font);

    case get_default_font:

        if (NULL == tml_default_font) {
            for (i = 0; i < tml_font_table_size; i++) {
                if (font_is_present(tml_font_table[i], strlen(tml_font_table[i])))
                    break;
                }

            if (i == tml_font_table_size) {
                for (i = 0; i < tml_font_table_size; i++) {
                    strcpy(full_path, FONT_PATH);
                    font_len = strlen(tml_font_table[i]);
                    strncat(full_path, tml_font_table[i], font_len);
                   /* *(full_path + strlen(FONT_PATH) + font_len) = '\0';*/
                    if (font_is_present(full_path, strlen(full_path)))
                        break;
                    }
                if (i == tml_font_table_size)
                    return(NULL);
                else
                    tml_default_font = tml_font_table[i];
                }
            else
                tml_default_font = tml_font_table[i];
            }
        *length = strlen(tml_default_font);
        return(tml_default_font);
    }
}


/* returns 1 if named font is present, 0 otherwise */
long font_is_present(font_name, name_len)
char *font_name;
short name_len;
{
    char          *mapped_at;
    unsigned long  length_mapped;
    status_$t      status;
    unsigned       long size;

    size = (long) sizeof(fontn_$table_tver1);
    mapped_at = ms_$mapl(font_name, name_len, (unsigned long) 0, size, ms_$cowriters, ms_$r, FALSE, &length_mapped, &status);
    if (status_$ok != status.all) {
        return(0);
        }
    else {
        ms_$unmap(((void*) mapped_at), length_mapped, &status);
        return (1);
        }
}


/* set the default font table - returns NULL */
char *set_default_font_table(table, table_size)
char* *table;                           /* array of null-terminated font names */
                                        /*   if NULL, use internal table */
long table_size;                         /* number of entries in table */
{
    return(gpr_serve_default_font_name(set_default_table, NULL, table, table_size));
}
        
/*externally accessable routine to return the default font*/
char* tml_dp_gpr_get_default_font(length)
long *length;
{
    return(gpr_serve_default_font_name(get_default_font, length, NULL, 0));
}
    

/* returns name of default font (allocated from free store) */
/* and length of name in contents of integer pointer argument */
char *get_gpr_default_font_name(length)
long *length;
{
    return(gpr_serve_default_font_name(get_default_font, length, NULL, 0));
}
        

/* set the default font name. Argument length contains length of name */
/* if font_name is NULL, the default is cleared and the next call to get_gpr_default_font_name() */
/*   will reinitialize the default font from the internal table of default fonts */
char *set_gpr_default_font_name(font_name, length)
char *font_name;
long *length;
{
    return(gpr_serve_default_font_name(set_default_font, length, &font_name, 1));
}


/******************************************************************************
    Functions for manipulating font handles
******************************************************************************/
/* allocate and release font handles. */
/* return NULL on failure, struct fhandle * otherwise */
struct fhandle *gpr_serve_font_handle_list(service, font_name, name_len, font_handle)
long service;
char *font_name;
short name_len;
struct fhandle *font_handle;
{

    char *get_font_info();

    struct fhandle *cur_fhandle;
    struct fhandle *next_fhandle;
    char           *cur_font_name;
    char            full_path[MAXNAMLEN]; /* == 512, sys5 definition from dirent.h */

    switch (service) {

        case get_font_handle:

            /* note: depends on tml_c_tml_os_alloc()'s nulling of space it allocates */
                                                   
            cur_fhandle = tml_fhandle_list_head;
        
            /* see if font is already loaded */
        
            while (NULL != cur_fhandle) {
                cur_font_name = cur_fhandle->font_name;
                if ((name_len == cur_fhandle->name_len) && (0 == strncmp(font_name, cur_font_name, name_len))) {
                    if (BAD_FONT_ID == cur_fhandle->font_table) {     /* previously tml_unloaded */
                        if (!(cur_fhandle->font_table = (fontn_$table_ptr_tver1) get_font_info(font_name, name_len, &(cur_fhandle->version)))) {
                            strcpy(full_path, FONT_PATH);
                            strncat(full_path, font_name, name_len);
                            *(full_path + strlen(FONT_PATH) + name_len) = '\0';
                            if (!(cur_fhandle->font_table = (fontn_$table_ptr_tver1) get_font_info(full_path, name_len + strlen(FONT_PATH), &(cur_fhandle->version)))) {
                                cur_fhandle = NULL;
                                }
                            }
                        }
                    cur_fhandle->ref_count++;
                    return(cur_fhandle);
                    }
                cur_fhandle = cur_fhandle->next;
                }
        
            /* font not already loaded */
        
            cur_fhandle = (struct fhandle *) tml_c_tml_os_alloc(1, (unsigned long) sizeof(struct fhandle));
            cur_fhandle->prev = NULL;
            cur_fhandle->next = tml_fhandle_list_head;
            if (tml_fhandle_list_head)
                tml_fhandle_list_head->prev = cur_fhandle;
            tml_fhandle_list_head = cur_fhandle;
            cur_fhandle->font_name = (char *) tml_c_tml_os_alloc(name_len + 1, (unsigned long) sizeof(char));
            strncpy(cur_fhandle->font_name, font_name, name_len);
            cur_fhandle->name_len = name_len;
            cur_fhandle->loaded = 0;
            cur_fhandle->replicated = 0;
            cur_fhandle->font_num = BAD_FONT_NUM;
            cur_fhandle->repli_font_num = BAD_FONT_NUM;
            cur_fhandle->font_table = BAD_FONT_ID;
            cur_fhandle->ref_count = 1;
            if (!(cur_fhandle->font_table = (fontn_$table_ptr_tver1) get_font_info(font_name, name_len, &(cur_fhandle->version)))) {
                strcpy(full_path, FONT_PATH);
                strncat(full_path, font_name, name_len);
                *(full_path + strlen(FONT_PATH) + name_len) = '\0';
                if (!(cur_fhandle->font_table = (fontn_$table_ptr_tver1) get_font_info(full_path, name_len + strlen(FONT_PATH), &(cur_fhandle->version)))) {
                    tml_dp_gpr_frelease((char *) cur_fhandle);
                    cur_fhandle = NULL;
                    }
                }
            return(cur_fhandle);
            
        case release_font_handle:

            font_handle->ref_count--;
            if (!font_handle->ref_count){
                if (NULL != font_handle->prev)
                    font_handle->prev->next = font_handle->next;
                else
                    tml_fhandle_list_head = font_handle->next;
                if (NULL != font_handle->next)
                    font_handle->next->prev = font_handle->prev;
                tml_c_tml_os_free(font_handle->font_name);
                tml_c_tml_os_free((void*) font_handle->font_table);
                tml_c_tml_os_free((void*) font_handle);
                }
            return(NULL);

        case unload_fonts:

            for (next_fhandle = tml_fhandle_list_head;
                 NULL != next_fhandle;
                 next_fhandle = next_fhandle->next) {
                next_fhandle->loaded = 0;
                next_fhandle->replicated = 0;
                next_fhandle->font_num = BAD_FONT_NUM;
                next_fhandle->repli_font_num = BAD_FONT_NUM;
                }
            return(NULL);

        }
}

/* Get a font handle from a font name. Initialize the smd_$font_table_ptr_t field with */
/*    information direct from the font file. Look for the font under the name given, */
/*    and if it is not found, look for it under "/sys/dm/fonts/<name_given>". If still */
/*    not found, return error. If found, map in the first (long) sizeof(smd_$font_table_t) bytes, */
/*    copy them, and unmap the font file */
void *tml_dp_gpr_get_fhandle(font_name, name_len)
    char *font_name;
    long name_len;
{
    return((char *) gpr_serve_font_handle_list(get_font_handle, font_name, name_len, NULL));
}

/* get font information from the font file specified by name and length arguments */
char* get_font_info(font_name, name_len,font_vers_no)
char *font_name;
short name_len;
pinteger *font_vers_no;
{
    char               *mapped_at;
    unsigned long       header_size, table_size, size, length_mapped, num_entries;
    status_$t           status;
    fontn_$table_tver1 *font_table;
    char               *font_info;

    header_size = (long) sizeof(fontn_$table_tver1);
    mapped_at = ms_$mapl(font_name, name_len, (unsigned long) 0, (unsigned long) header_size, ms_$cowriters, ms_$r, FALSE, &length_mapped, &status);
    if (status_$ok != status.all)
        return(NULL);

    /* read the font table header */
    font_table = (fontn_$table_tver1*) mapped_at;

    /* set return parameter to font version number */
    *font_vers_no = font_table->version;

    /* calculate number of entries in index table and maximum number of entries in descriptor table */
    num_entries = font_table->highest_char_code + 1;

    /* calculate file extent to end of maximum possible size of tables */
    /*   bytes_per_char: also gives number of bytes in index table entry per character */
    /*   chars_in_font:  number of nonzero entries in index table; maximum number of entries in descriptor table */
    /*   fontn_$desc:    size of descriptor table entry */
    table_size = (num_entries * font_table->bytes_per_char) + (font_table->chars_in_font * sizeof(fontn_$desc));

    /* get tables */
    size = header_size + table_size;
    ms_$unmap(mapped_at, length_mapped, &status);
    mapped_at = ms_$mapl(font_name, name_len, (unsigned long) 0, (unsigned long) size, ms_$cowriters, ms_$r, FALSE, &length_mapped, &status);

    if (status_$ok != status.all)
        return(NULL);

    font_info = tml_c_tml_os_alloc(1, size);
   
    /* copy font information */
    tml_c_tml_os_byte_copy(mapped_at, (char *) font_info, size);

    ms_$unmap(mapped_at, length_mapped, &status);

    return (font_info);
}


/* load a font - return load status cast as an integer - equivalent to gpr_$load_font and gpr_$replicate_font */
long tml_dp_gpr_load_font(in_font_handle)
void* in_font_handle;
{
    struct fhandle        *font_handle = (struct fhandle*) in_font_handle;
    status_$t              font_status;

    if (0 == font_handle->loaded) {
        gpr_$load_font_file(font_handle->font_name, font_handle->name_len, &(font_handle->font_num), &font_status);
        if (status_$ok == font_status.all){
            font_handle->loaded = -1;
            gpr_$replicate_font((short int) font_handle->font_table, &(font_handle->repli_font_num), &font_status);
            }
        }
    return((long) font_status.all);
}


long tml_dp_gpr_get_fcharsize(in_font_handle)
    void* in_font_handle;
{     
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    return (font_handle->font_table->bytes_per_char);
}


long tml_dp_gpr_get_fcharnum(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;
    unsigned long lo_index;
    unsigned long hi_index;

    if (font_handle->font_table->default_char < (lo_index = font_handle->font_table->lowest_char_code))
        lo_index = font_handle->font_table->default_char;
    if (font_handle->font_table->default_char > (hi_index = font_handle->font_table->highest_char_code))
        hi_index = font_handle->font_table->default_char;
    return(1 + hi_index - lo_index);
}


long tml_dp_gpr_get_fchar1(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    if (font_handle->font_table->lowest_char_code 
      < font_handle->font_table->default_char)
        return(font_handle->font_table->lowest_char_code);
    return font_handle->font_table->default_char;
}


long tml_dp_gpr_get_fhi(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    return(font_handle->font_table->max_height + font_handle->font_table->v_spacing);
}


/* return font point size from font handle */
long tml_dp_gpr_get_fpoint(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    return(font_handle->font_table->point_size);
}


long tml_dp_gpr_get_fbline(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    return(font_handle->font_table->max_up);
}


/* return character width in rasters for fixed fonts, zero for proportional-spaced fonts */
long tml_dp_gpr_is_fixed(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    long *widths;
    long  space_size;
    long  is_fixed;
    char *index_table;
    long  i;
    long  index;

    if (font_handle->font_table->space_type == fontn_$mono_spacing)
        return (font_handle->font_table->space_size);
    else
        return 0;
}


/* return pointer to array of character widths for proportional fonts, NULL otherwise */
short** tml_dp_gpr_get_fwidths(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    register long             i, 
                              j,
                              num_entries, 
                              tbl_index, 
                              dflt_width;
    register short           *widths,
                             *lbearings, 
                             *rbearings;
             short          **tbl_arr;
             long             lo_index,
                              num_slots;
    register fontn_$desc     *desc_tbl;
    register unsigned char   *index_tbl;
    register unsigned short  *index_tbl16;
       
    if (font_handle->font_table->space_type == fontn_$mono_spacing)
        /* monospace font; no width table needed */
        return 0;

    /* lowest character value for which table will have information */
    /*   or default character, whichever is lower */
    lo_index = tml_dp_gpr_get_fchar1(in_font_handle);

    /* number of entries to fill in width table */
    num_entries = tml_dp_gpr_get_fcharnum(in_font_handle);

    if (0 == num_entries) {
        fprintf(stderr,"tml_dp_gpr_get_fwidths: font has no characters, hence no widths\n");
        return NULL;
        }

    if (tml_dp_gpr_is_fixed(font_handle))
        return NULL;

    /* use default width of space character for undescribed characters */
    dflt_width = font_handle->font_table->space_size;

    /* descriptor table -- see /usr/include/apollo/fontn.h */
    desc_tbl = (fontn_$desc*) (((unsigned char*) font_handle->font_table) + font_handle->font_table->desc_table_offset);

    /* array of arrays to be returned */
    tbl_arr = (short**) tml_c_tml_os_alloc(3, (unsigned long) sizeof(short*));

    /* array of widths to be returned for each indexed character */
    num_slots = num_entries;
    if (num_entries & 1)
        ++num_slots;
    tbl_arr[0] = widths    = (short*) tml_c_tml_os_alloc(num_slots, (unsigned long) sizeof(short));
    /* array of left bearings to be returned for each indexed character */
    tbl_arr[1] = lbearings = (short*) tml_c_tml_os_alloc(num_slots, (unsigned long) sizeof(short));
    /* array of right bearings to be returned for each indexed character */
    tbl_arr[2] = rbearings = (short*) tml_c_tml_os_alloc(num_slots, (unsigned long) sizeof(short));

    if (1 == font_handle->font_table->bytes_per_char) {
        /* 8-bit characters */
        index_tbl = ((unsigned char*) font_handle->font_table) + font_handle->font_table->index_table_offset;
        for (i = 0, j = lo_index; i <= font_handle->font_table->highest_char_code; i++, j++) {
            if (tbl_index = *(index_tbl + j)) {
                /* correct faulty font information */
                *(widths + i)    = (desc_tbl + tbl_index - 1)->width;
                if (0 < (desc_tbl + tbl_index - 1)->left)
                    (desc_tbl + tbl_index - 1)->left *= -1;
                *(lbearings + i) = (desc_tbl + tbl_index - 1)->left;
                if (0 > (desc_tbl + tbl_index - 1)->right)
                    (desc_tbl + tbl_index - 1)->right *= -1;
                else if ((desc_tbl + tbl_index - 1)->right < (desc_tbl + tbl_index - 1)->width)
                    (desc_tbl + tbl_index - 1)->right = (desc_tbl + tbl_index - 1)->width;
                *(rbearings + i) = (desc_tbl + tbl_index - 1)->right;
                }
            else {
                /* directionality not a factor in GPR font widths because GPR paints */
                /*   upside-down and backward to render characters in right-to-left text */
                /* therefore set width and rbearing to default width and make lbearing zero */
                /*   when character without descriptor encountered, without regard to */
                /*   text path of characters */
                *(widths + i)    = dflt_width;
                *(lbearings + i) = 0;
                *(rbearings + i) = dflt_width;
                }
            }
        for (; i < num_entries; i++){
            /* directionality not a factor in GPR, which paints upside-down and */
            /*   backward to render characters in right-to-left text */
            /* therefore set width and rbearing to default width and make lbearing zero */
            /*   when character without descriptor encountered */
            *(widths + i)    = dflt_width;
            *(lbearings + i) = 0;
            *(rbearings + i) = dflt_width;
            }
        }
    else {
        /* 16-bit characters */
        index_tbl16 = (unsigned short*) (((unsigned char*) font_handle->font_table) + font_handle->font_table->index_table_offset);
        for (i = 0, j = lo_index; i <= font_handle->font_table->highest_char_code; i++, j++) {
            /* correct faulty font information */
            if (tbl_index = *(index_tbl16 + j)) {
                *(widths + i)    = (desc_tbl + tbl_index - 1)->width;
                if (0 < (desc_tbl + tbl_index - 1)->left)
                    (desc_tbl + tbl_index - 1)->left *= -1;
                *(lbearings + i) = (desc_tbl + tbl_index - 1)->left;
                if (0 > (desc_tbl + tbl_index - 1)->right)
                    (desc_tbl + tbl_index - 1)->right *= -1;
                *(rbearings + i) = (desc_tbl + tbl_index - 1)->right;
                }
            else {
                /* directionality not a factor in GPR, which paints upside-down and */
                /*   backward to render characters in right-to-left text */
                /* therefore set width and rbearing to default width and make lbearing zero */
                /*   when character without descriptor encountered */
                *(widths + i)    = dflt_width;
                *(lbearings + i) = 0;
                *(rbearings + i) = dflt_width;
                }
            }
        for (; i < num_entries; i++){
            /* directionality not a factor in GPR, which paints upside-down and */
            /*   backward to render characters in right-to-left text */
            /* therefore set width and rbearing to default width and make lbearing zero */
            /*   when character without descriptor encountered */
            *(widths + i)    = dflt_width;
            *(lbearings + i) = 0;
            *(rbearings + i) = dflt_width;
            }
        }
    return tbl_arr;
}


long tml_dp_gpr_get_ficspace(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    return(font_handle->font_table->inter_space_dflt);
}


/* set fixed intercharacter space for fixed fonts, nominal letterspacing for proportional fonts */
/* Implements track kerning. */
/* NOTE: For GPR, when this quantity is positive it increases the space */
/*   between characters regardless of text direction because the font */
/*   images for right-to-left text are rendered upside-down and backwards. */
/*   A negative value decreases the space between characters. */
long tml_dp_gpr_set_ficspace(in_font_handle, new_icspace)
    void* in_font_handle;
    long new_icspace;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    long status;

    font_handle->font_table->inter_space_dflt = new_icspace;

    return(new_icspace);
}


/* get the default character, if any, for the font */
long tml_dp_gpr_get_fspace(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    return font_handle->font_table->default_char;
}


/* set the default character for the font */
void tml_dp_gpr_set_fspace(in_font_handle, new_space)
    void* in_font_handle;
    long  new_space;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    font_handle->font_table->default_char = new_space;
}


/* unload a font */
long tml_dp_gpr_funload(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;
    status_$t font_status;

    gpr_$unload_font_file((short) (font_handle->font_table), &font_status);
    if (status_$ok == font_status.all){
        font_handle->loaded = 0;
        font_handle->replicated = 0;
        font_handle->font_num = BAD_FONT_NUM;
        font_handle->repli_font_num = BAD_FONT_NUM;
        font_handle->font_table = BAD_FONT_ID;
        }
    return((long) font_status.all);
}


void tml_dp_gpr_frelease(in_font_handle)
    void* in_font_handle;
{
    struct fhandle *font_handle = (struct fhandle*) in_font_handle;

    gpr_serve_font_handle_list(release_font_handle, NULL, 0, font_handle);
}

/********************************************************************************
    Functions to manage bitmap information
********************************************************************************/

/* find a bitmap in the bitmap store */
/* returns NULL if the bitmap isn't found */
/* otherwise returns a pointer to the struct bm_info associated with the bitmap */
tml_c_bitmap_id find_bitmap_info(bitmap_id, bitmap_info_ptr)
    tml_c_bitmap_id  bitmap_id;
    struct bm_info* *bitmap_info_ptr;
{
    tml_c_bitmap_id gpr_serve_bm_info();

    return(gpr_serve_bm_info(find_bitmap_handle_info, bitmap_id, bitmap_info_ptr, NULL));
}

/* make and track bitmaps */
/* return NULL on failure, a bitmap_id otherwise */
tml_c_bitmap_id gpr_serve_bm_info(service, bitmap_id, bitmap_info_ptr, viewport_id)
    long             service;
    tml_c_bitmap_id  bitmap_id;
    struct bm_info* *bitmap_info_ptr;
    tml_c_view_port  viewport_id;
{
    status_$t           status;
    tml_c_font_tbl      cur_font_tbl;
    char               *font_name;
    short               font_name_len;
    short               current_font;
    gpr_$direction_t    current_dir;
    struct bm_info     *bitmap_info;
    gpr_$mask_32_t      canned_mask = -1;
    gpr_$raster_op_t    canned_rop = gpr_$rop_src;
    unsigned short int  prim_set;
    unsigned short int  prim_bit;
    unsigned long       i;

    switch (service) {

        case make_bitmap_handle:

            /* return tml_c_bitmap_id argument if successful */

            if (NULL == find_bitmap_info(bitmap_id, &bitmap_info)) {
                if (NULL == (bitmap_info = (struct bm_info *) tml_c_tml_os_alloc(1, (unsigned long) sizeof(struct bm_info))))
                    return(NULL);
                bitmap_info->drawn = 0;
                bitmap_info->bitmap = (gpr_$bitmap_desc_t *) tml_c_tml_os_alloc(1, (unsigned long) sizeof(gpr_$bitmap_desc_t));
                if (!bitmap_info->bitmap)
                    return (NULL);
                else
                    *(bitmap_info->bitmap) = *((gpr_$bitmap_desc_t*)bitmap_id);
                bitmap_info->next = tml_bm_info_queue;
                if (tml_bm_info_queue)
                    tml_bm_info_queue->prior = bitmap_info;
                tml_bm_info_queue = bitmap_info;
                bitmap_info->size_x = bitmap_get_wide(bitmap_id);
                bitmap_info->size_y = bitmap_get_hi(bitmap_id);

                bitmap_info->foreground_id = bitmap_get_foreg_id(bitmap_id);
                bitmap_info->background_id = bitmap_get_backg_id(bitmap_id);

                /* gpr_$background (-2) is background value in direct mode when user */
                /*   desires background color to be the background color of the DM */
                /*   window used for direct mode */
                /* a value of 1 is the default foreground when gpr_$background is set */
                /* the gpr calls are safe only in direct and frame mode, */
                /*   but gpr_$background is a possible value only in those modes */
                /* ignore the possibility of gpr_$transparent background; */
                /*   that value doesn't work in TMLib anyway */
                if (gpr_$background == bitmap_info->background_id) {
                    bitmap_info->background_id = gpr_$inq_background(&status);
                    gpr_$set_text_background_value((gpr_$pixel_value_t) bitmap_info->background_id, &status);
                    if (1 == bitmap_get_foreg_id(bitmap_id)) {
                        bitmap_info->foreground_id = gpr_$inq_foreground(&status);
                        gpr_$set_text_value((gpr_$pixel_value_t) bitmap_info->foreground_id, &status);
                        }
                    }

                /* retrieve the raster op primitive set of the destination bitmap */
                /*   (i.e. the set of types of gpr operations that are affected */
                /*   by raster ops */
                gpr_$inq_raster_op_prim_set(&prim_set, &status);
                
                /* make all modes accessible to raster operations */
                /* add all modes to the raster op prim set */
                /*   (the raster op prim set is a Pascal set; the C enum for the */
                /*   gpr operations types is zero-based, not one-based like the */
                /*   Pascal version of the enumeration) */
                for (i = 0, prim_bit = 1; i <= gpr_$rop_fill; i++) {
                    gpr_$raster_op_prim_set(prim_set | prim_bit, &status);
                    prim_bit <<= 1;
                    }

                /* set raster ops active on all planes */
                /* set initial value of rops on all planes to gpr_$rop_src */
                gpr_$set_raster_op_mask(canned_mask, canned_rop, &status);

                /* set fill value for objects (circles, triangles, rectangles) */
                /*   to text background value */
                /* this allows tml_dp_gpr_clear to work */
                /* the value is differently set and then restored by tml_dp_gpr_dflt_cursor_draw */
                gpr_$set_fill_value((gpr_$pixel_value_t) bitmap_info->background_id, &status);
                if (status.all != status_$ok) return;

                /* set fill value for window background */
                /*   to text background value */
                gpr_$set_fill_background_value((gpr_$pixel_value_t) bitmap_info->background_id, &status);
                if (status.all != status_$ok) return;

                /* record fill_background_value and fill_value of bitmap */
                bitmap_info->fill_backg_id = bitmap_get_fill_backg_id(bitmap_id);
                bitmap_info->fill_obj_id = bitmap_get_fill_obj_id(bitmap_id);

                bitmap_info->viewport_id = viewport_id;
                gpr_$inq_raster_ops(bitmap_info->current_rops, &status);
                if (status.all != status_$ok)
                    return(NULL);
                if (NULL == (cur_font_tbl = (char *) view_port_inq_state(viewport_id, tml_font_prop)))
                    bitmap_info->font_id = NULL;
                else {
                    font_name = font_tbl_get_name(cur_font_tbl);
                    font_name_len = font_tbl_get_name_count(cur_font_tbl);
                    bitmap_info->font_id = tml_dp_gpr_get_fhandle(font_name, font_name_len);
                    gpr_$inq_text(&current_font, &(bitmap_info->current_text_direction), &status);
                    if (status.all != status_$ok)
                        return(NULL);
                    }
                }
            bitmap_info->viewport_id = viewport_id;
            return((tml_c_bitmap_id) bitmap_info->bitmap);

        case delete_bitmap_handle:

            /* return tml_c_bitmap_id argument if successful */

            if (NULL == find_bitmap_info(bitmap_id, &bitmap_info))
                return(NULL);
            else {
                if (bitmap_info->prior)
                    bitmap_info->prior->next = bitmap_info->next;
                else
                    tml_bm_info_queue = bitmap_info->next;
                if (bitmap_info->next)
                    bitmap_info->next->prior = bitmap_info->prior;
                if (bitmap_info->bitmap)
                    tml_c_tml_os_free((void*) bitmap_info->bitmap);
                tml_c_tml_os_free((void*) bitmap_info);
                return(bitmap_id);
                }

        case get_current_bitmap_handle:

            /* return current_bitmap if successful */

            return(tml_current_bitmap);

        case set_current_bitmap_handle:

            /* only bitmap_id's returned from tml_dp_gpr_make_graphic_handle are eligible */
            /* returns tml_c_bitmap_id argument if successful */

            if (NULL == find_bitmap_info(bitmap_id, &bitmap_info))
                return(NULL);
            else {
                gpr_$set_bitmap(*((gpr_$bitmap_desc_t*) bitmap_id), &status);
                if (status.all != status_$ok)
                    return(NULL);
                tml_current_bitmap = bitmap_id;
                return(tml_current_bitmap);
                }

        case find_bitmap_handle_info:

            /* set contents of bitmap_info_ptr to address of struct bm_info for the bitmap */
            /*   if there is one; otherwise null bitmap_info_ptr's contents and return NULL */

            for (bitmap_info = tml_bm_info_queue; bitmap_info; bitmap_info = bitmap_info->next) {
                if (*(bitmap_info->bitmap) == (*(gpr_$bitmap_desc_t *) bitmap_id)) {
                    *bitmap_info_ptr = bitmap_info;
                    return(bitmap_id);
                    }
            }
            *bitmap_info_ptr = NULL;
            return(NULL);

        } /* end switch */
}

/*refresh a bitmap*/
tml_dp_gpr_refresh(unobs, pos_change)
    boolean unobs,
            pos_change;
{
    struct bm_info     *bitmap_info;
    gpr_$bitmap_desc_t  this_bitmap;
    status_$t           status;
    tml_c_view_port     refresh_viewport;
    gpr_$offset_t       bm_size;

    gpr_$inq_bitmap(&this_bitmap, &status);
    if (status.all != status_$ok) return;
                                                    
    if (NULL == gpr_serve_bm_info(find_bitmap_handle_info, &this_bitmap, &bitmap_info, NULL))
        return;
                        
    refresh_viewport = bitmap_info->viewport_id;
    view_port_redraw(refresh_viewport);
}


/* make a bitmap/viewport coordination structure (struct bm_info) */
tml_c_bitmap_id tml_dp_gpr_make_graphic_handle(bitmap_id, viewport_id)
    tml_c_bitmap_id bitmap_id;
    tml_c_view_port viewport_id;
{
    status_$t           status;
    gpr_$bitmap_desc_t  old_bitmap,
                        new_bitmap;
    gpr_$rwin_pr_t      refresh_rtn,
                        no_rtn;
    tml_c_bitmap_id     return_bitmap;

    new_bitmap = *((gpr_$bitmap_desc_t*) bitmap_id);
    gpr_$inq_bitmap(&old_bitmap, &status);
    if (status.all != status_$ok)
        old_bitmap = NULL;
    if (old_bitmap != new_bitmap) {
        gpr_$set_bitmap(new_bitmap, &status);
        if (status.all != status_$ok)
            return;
        }
    no_rtn = 0;
    refresh_rtn = &tml_dp_gpr_refresh;
    gpr_$set_refresh_entry(refresh_rtn, no_rtn, &status);
    return_bitmap = gpr_serve_bm_info(make_bitmap_handle, bitmap_id, NULL, viewport_id);
    if (old_bitmap && (old_bitmap != new_bitmap))
        gpr_$set_bitmap(old_bitmap, &status);
    return return_bitmap;
    }

/* delete a bitmap/viewport coordination structure (struct bm_info) */
void tml_dp_gpr_delete_graphic_handle(bitmap_id)
    tml_c_bitmap_id bitmap_id;
{
    gpr_serve_font_handle_list(unload_fonts, NULL, 0, NULL);
    gpr_serve_bm_info(delete_bitmap_handle, bitmap_id, NULL, NULL);
}

/* return the bitmap_info for the current bitmap */
tml_c_bitmap_id tml_dp_gpr_get_current_bitmap()
{
    return(gpr_serve_bm_info(get_current_bitmap_handle, 0, NULL, NULL));
}


/********************************************************************************
    Functions to get information about a bitmap.
    The bitmap_get_* functions use gpr calls to access the bitmap itself and
        should be used only when no bitmap_info structure is associated with
        the bitmap, although they are safe to use when such a structure has
        been associated with the bitmap, and that structure will be updated.
        The corresponding tml_dp_gpr_get_* functions access the bitmap_info
        associated with an argument bitmap_handle. The corresponding
        bitmap_info_get_* functions return fields of an argument bitmap_info.
********************************************************************************/

/* use the current tml bitmap_id to derive a current gpr bitmap_id */
gpr_$bitmap_desc_t tml_dp_gpr_get_gpr_bitmap()
{
    tml_c_bitmap_id current_bitmap_id;

    if (NULL == (current_bitmap_id = tml_dp_gpr_get_current_bitmap()))
        return 0;
    return *((gpr_$bitmap_desc_t*) current_bitmap_id);
}

/*get the height of a bitmap from gpr*/
long bitmap_get_hi(bitmap_handle)
    gpr_$bitmap_desc_t    *bitmap_handle;
{
    status_$t           status;
    gpr_$offset_t       window_size;
    gpr_$rgb_plane_t    hi_plane;
    gpr_$bitmap_desc_t  current_bitmap;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$inq_bitmap_dimensions(*bitmap_handle, &window_size, &hi_plane, &status);
    if (status.all != status_$ok) return -1;
    return window_size.y_size;
}

/*get the height of a bitmap from its bitmap_info*/
long tml_dp_gpr_get_hi(bitmap_handle)
    tml_c_bitmap_id bitmap_handle;
{
    return(bitmap_get_hi(bitmap_handle));
}

/*get the width of a bitmap from gpr*/
long bitmap_get_wide(bitmap_handle)
    gpr_$bitmap_desc_t    *bitmap_handle;
{
    status_$t           status;
    gpr_$offset_t       window_size;
    gpr_$rgb_plane_t    hi_plane;
    gpr_$bitmap_desc_t  current_bitmap;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }
    gpr_$inq_bitmap_dimensions(*bitmap_handle, &window_size, &hi_plane, &status);
    if (status.all != status_$ok) return -1;
    return window_size.x_size;
}

/*get the width of a bitmap from its bitmap_info*/
long tml_dp_gpr_get_wide(bitmap_handle)
    tml_c_bitmap_id bitmap_handle;
{
    return(bitmap_get_wide(bitmap_handle));
}

/*get the foreground (text drawing) color of a bitmap from gpr*/
long bitmap_get_foreg_id(bitmap_handle)
    gpr_$bitmap_desc_t    *bitmap_handle;
{
    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    gpr_$pixel_value_t  foreground, background;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$inq_text_values(&foreground, &background, &status);
    if (status.all != status_$ok) return -1;

    return foreground;
}

/*get the foreground (text drawing) color of a bitmap from its bitmap_info*/
long tml_dp_gpr_get_foreg_id(bitmap_handle)
    tml_c_bitmap_id bitmap_handle;
{
    struct bm_info *bitmap_info;

    if (NULL == find_bitmap_info(bitmap_handle, &bitmap_info))
        return(bitmap_get_foreg_id(bitmap_handle));
    else
        return(bitmap_info->foreground_id);
}

/*get the (text) background color of a bitmap from gpr*/
long bitmap_get_backg_id(bitmap_handle)
    gpr_$bitmap_desc_t    *bitmap_handle;
{
    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    gpr_$pixel_value_t  foreground, background;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$inq_text_values(&foreground, &background, &status);
    if (status.all != status_$ok) return -1;

    return background;
}

/*get the (text) background color of a bitmap from its bitmap_info*/
long tml_dp_gpr_get_backg_id(bitmap_handle)
    tml_c_bitmap_id bitmap_handle;
{
    struct bm_info *bitmap_info;

    if (NULL == find_bitmap_info(bitmap_handle, &bitmap_info))
        return(bitmap_get_backg_id(bitmap_handle));
    else
        return(bitmap_info->background_id);
}

/* get the background color for the viewport from gpr (i.e. not the color of */
/*   the text, not the color of the rectangle enclosing each text character, */
/*   but the color of the "surface" on which the text characters and their */
/*   enclosing rectangles are overlaid) */
long bitmap_get_fill_backg_id(bitmap_handle)
    gpr_$bitmap_desc_t   *bitmap_handle;
{
    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    gpr_$pixel_value_t  fill_backg_id;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle) {
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$inq_fill_background_value(&fill_backg_id, &status);
    if (status.all != status_$ok) return -1;

    return fill_backg_id;
}

/* get the background color for a viewport from its bitmap's bitmap_info */
long tml_dp_gpr_get_fill_backg_id(bitmap_handle)
    tml_c_bitmap_id bitmap_handle;
{
    struct bm_info *bitmap_info;

    if (NULL == find_bitmap_info(bitmap_handle, &bitmap_info))
        return(bitmap_get_fill_backg_id(bitmap_handle));
    else
        return(bitmap_info->fill_backg_id);
}

/* get the area fill color for a viewport's bitmap from gpr */
/*   (the color used to fill circles, triangles, and rectangles) */
long bitmap_get_fill_obj_id(bitmap_handle)
    gpr_$bitmap_desc_t   *bitmap_handle;
{
    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    gpr_$pixel_value_t  fill_obj_id;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle) {
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$inq_fill_value(&fill_obj_id, &status);
    if (status.all != status_$ok) return -1;

    return fill_obj_id;
}

long tml_dp_gpr_get_fill_obj_id(bitmap_handle)
    tml_c_bitmap_id bitmap_handle;
{
    struct bm_info *bitmap_info;

    if (NULL == find_bitmap_info(bitmap_handle, &bitmap_info))
        return(bitmap_get_fill_obj_id(bitmap_handle));
    else
        return(bitmap_info->fill_obj_id);
}

/********************************************************************************
    Functions to set bitmap information.
********************************************************************************/

/*set the foreground (text drawing) color for a bitmap*/
void tml_dp_gpr_set_foreg_id(in_bitmap_handle, in_new_fg_id)
    void* in_bitmap_handle;
    long in_new_fg_id;
{
    gpr_$bitmap_desc_t *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;
    gpr_$pixel_value_t  new_fg_id = (gpr_$pixel_value_t) in_new_fg_id;

    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$set_text_value(new_fg_id, &status);
    if (NULL != find_bitmap_info(bitmap_handle, &bitmap_info))
        bitmap_info->foreground_id = new_fg_id;
}

/*set the background value*/
void tml_dp_gpr_set_backg_id(in_bitmap_handle, in_new_bg_id)
    void* in_bitmap_handle;
    long in_new_bg_id;
{
    gpr_$bitmap_desc_t *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;
    gpr_$pixel_value_t  new_bg_id = (gpr_$pixel_value_t) in_new_bg_id;

    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$set_text_background_value(new_bg_id, &status);
    if (NULL != find_bitmap_info(bitmap_handle, &bitmap_info))
        bitmap_info->background_id = new_bg_id;
}
                                                             
/* set the background fill color for the viewport (i.e. not the color of the text, */
/*   not the color of the rectangle enclosing each text character, but the */
/*   color of the "surface" on which the text characters and their enclosing */
/*   rectangles are overlaid */
void tml_dp_gpr_set_fill_backg_id(in_bitmap_handle, in_new_fill_bg_id)
    void* in_bitmap_handle;
    long in_new_fill_bg_id;
{
    gpr_$bitmap_desc_t *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;
    gpr_$pixel_value_t  new_fill_bg_id = (gpr_$pixel_value_t) in_new_fill_bg_id;

    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$set_fill_background_value(new_fill_bg_id, &status);
    if (NULL != find_bitmap_info(bitmap_handle, &bitmap_info))
        bitmap_info->fill_backg_id = new_fill_bg_id;
}

/* set the object fill color for the viewport (i.e. not the color of the text, */
/*   not the color of the rectangle enclosing each text character, but the */
/*   color of the "surface" on which the text characters and their enclosing */
/*   rectangles are overlaid */
void tml_dp_gpr_set_fill_obj_id(in_bitmap_handle, in_new_fill_obj_id)
    void* in_bitmap_handle;
    long in_new_fill_obj_id;
{
    gpr_$bitmap_desc_t *bitmap_handle = in_bitmap_handle;
    gpr_$pixel_value_t  new_fill_obj_id = (gpr_$pixel_value_t) in_new_fill_obj_id;

    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    struct bm_info     *bitmap_info;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }

    gpr_$set_fill_value(new_fill_obj_id, &status);
    if (NULL != find_bitmap_info(bitmap_handle, &bitmap_info))
        bitmap_info->fill_obj_id = new_fill_obj_id;
}

/*set and activate a clipping window in a bitmap*/
void tml_dp_gpr_set_clip(in_bitmap_handle, clip_ulx, clip_uly, clip_lrx, clip_lry)
    void*   in_bitmap_handle;
    long    clip_ulx,
            clip_uly,
            clip_lrx,
            clip_lry;
{
    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    gpr_$window_t       clip_window;
    struct bm_info     *bitmap_info;

    gpr_$bitmap_desc_t    *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }
                                                  
    clip_window.window_base.x_coord = clip_ulx;
    clip_window.window_base.y_coord = clip_uly;
    clip_window.window_size.x_size = clip_lrx - clip_ulx;
    clip_window.window_size.y_size = clip_lry - clip_uly;

    gpr_$set_clip_window(clip_window, &status);
    if (status.all != status_$ok) return;
    gpr_$set_clipping_active(true, &status);
}

/*deactivate clipping*/
void tml_dp_gpr_clear_clip(in_bitmap_handle)
    void* in_bitmap_handle;
{
    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    struct bm_info     *bitmap_info;

    gpr_$bitmap_desc_t *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL);
        }
                                                  
    gpr_$set_clipping_active(false, &status);
}

/********************************************************************************
    Functions that draw things.
********************************************************************************/

/*clear a rectangular area of the bitmap*/
void tml_dp_gpr_clear(in_bitmap_handle, ulx, uly, lrx, lry)
    void* in_bitmap_handle;
    long    ulx, uly, lrx, lry;
{
    status_$t               status;
    gpr_$bitmap_desc_t      current_bitmap;
    gpr_$window_t           clear_rect;
    gpr_$pixel_value_t      fill_value, background;
    gpr_$mask_32_t          canned_mask = -1;
    gpr_$raster_op_t        canned_rop = gpr_$rop_src;
    boolean                 unobsc;
    struct bm_info         *bitmap_info;
    short                   returned_slots;
    gpr_$window_t           vis_list[NUM_SLOTS];
    register long           i;
    unsigned short int      prim_set;
    unsigned short int      prim_bit;
    gpr_$bitmap_desc_t      dm_curs_pat;
    gpr_$raster_op_array_t  dm_curs_raster_op;
    boolean                 dm_curs_active;
    gpr_$position_t         dm_curs_position;
    gpr_$position_t         dm_curs_origin;
    status_$t               dm_curs_status;

    gpr_$bitmap_desc_t    *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        current_bitmap = *((gpr_$bitmap_desc_t*) gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL));
        }
                         
    if (!current_bitmap) {                  /* no bitmap_info for the bitmap */
        background = bitmap_get_backg_id(bitmap_handle);
        if (status.all != status_$ok) return;
        gpr_$inq_fill_value(&fill_value, &status);
        if (status.all != status_$ok) return;
        }
    else {
        find_bitmap_info(bitmap_handle, &bitmap_info);
        background = bitmap_info->background_id;
        fill_value = bitmap_info->fill_obj_id;
        }

    if (fill_value != background) {

        /* NOTE: RESET OF RASTER OP PRIMITIVE SET */
        /* MUST PRECEDE RESET OF FILL VALUE */
        /* OR FILL VALUE WILL NOT BE RESET */
        
        /* remember the raster op primitive set of the destination bitmap */
        /*   (i.e. the set of types of gpr operations that are affected */
        /*   by raster ops */
        gpr_$inq_raster_op_prim_set(&prim_set, &status);
        
        /* add gpr_$rop_fill to the raster op prim set */
        /*   (the raster op prim set is a Pascal set; the C enum for the */
        /*   gpr operations types is zero-based, not one-based like */
        /*   Pascal version of the enumeration) */
        for (i = 0, prim_bit = 1; i < gpr_$rop_fill; i++)
            prim_bit <<= 1;
        gpr_$raster_op_prim_set(prim_set | prim_bit, &status);

        /* set raster operations to copy for all planes */
        gpr_$set_raster_op_mask(canned_mask, canned_rop, &status);

        gpr_$set_fill_value(background, &status);
        if (status.all != status_$ok) return;
        }

    clear_rect.window_base.x_coord = ulx;
    clear_rect.window_base.y_coord = 1 + uly;
    clear_rect.window_size.x_size = lrx - ulx;
    clear_rect.window_size.y_size = 1 + lry - uly;

    /* permit acquisition of display when window obscured */
    gpr_$set_obscured_opt(gpr_$ok_if_obs, &status);

    /* get cursor information */
    gpr_$inq_cursor(&dm_curs_pat, dm_curs_raster_op, &dm_curs_active, 
                    &dm_curs_position, &dm_curs_origin, &dm_curs_status);

    /* set DM cursor inactive during blit */
    gpr_$set_cursor_active(false, &status);

    unobsc = gpr_$acquire_display(&status);
    if (status.all != status_$ok){
        gpr_$release_display(&status);
        return;
        }
    if (unobsc)
        gpr_$rectangle(clear_rect, &status);
    else {
        gpr_$inq_vis_list((short)NUM_SLOTS, &returned_slots, vis_list, &status);
        if (returned_slots > NUM_SLOTS)
            returned_slots = NUM_SLOTS;
        gpr_$set_clipping_active(true, &status);
        for (i = 0; i < returned_slots; i++) {
            gpr_$set_clip_window(vis_list[i], &status);
            gpr_$rectangle(clear_rect, &status);
            }
        gpr_$set_clipping_active(false, &status);
        }

    gpr_$set_fill_value(fill_value, &status);
    if (status.all != status_$ok) return;

    /* reset activity of DM cursor to value before blt */
    gpr_$set_cursor_active(dm_curs_active, &status);

    gpr_$release_display(&status);
    
}

/*copy a portion of the indicated bitmap from one location to another*/
void tml_dp_gpr_copy(in_bitmap_handle, src_ulx, src_uly, src_lrx, src_lry, dst_ulx, dst_uly)
    void* in_bitmap_handle;
    long    src_ulx,
            src_uly,
            src_lrx,
            src_lry,
            dst_ulx,
            dst_uly;
{
    status_$t               status;
    gpr_$bitmap_desc_t      current_bitmap;
    gpr_$window_t           copy_rect;
    gpr_$position_t         destination;
    gpr_$raster_op_array_t  current_rops;
    gpr_$mask_32_t          canned_mask = -1;
    gpr_$raster_op_t        canned_rop = gpr_$rop_src;
    short                   i;
    boolean                 unobsc;
    struct bm_info         *bitmap_info;
    short                   returned_slots;
    gpr_$window_t           vis_list[NUM_SLOTS];
    register long           j;
    gpr_$bitmap_desc_t      dm_curs_pat;
    gpr_$raster_op_array_t  dm_curs_raster_op;
    boolean                 dm_curs_active;
    gpr_$position_t         dm_curs_position;
    gpr_$position_t         dm_curs_origin;
    status_$t               dm_curs_status;

    gpr_$bitmap_desc_t    *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        current_bitmap = *((gpr_$bitmap_desc_t*) gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL));
        }
              
    /*make sure that we've got the right raster-op*/
    if (!current_bitmap) {            /* no bitmap_info for bitmap_handle */
        gpr_$inq_raster_ops(current_rops, &status);
        if (status.all != status_$ok) return;
        }
    else
        find_bitmap_info(bitmap_handle, &bitmap_info);

    /* set raster operations to copy for all planes */
    gpr_$set_raster_op_mask(canned_mask, canned_rop, &status);

    copy_rect.window_base.x_coord = src_ulx;
    copy_rect.window_base.y_coord = src_uly;
    copy_rect.window_size.x_size = src_lrx - src_ulx;
    copy_rect.window_size.y_size = src_lry - src_uly;
    destination.x_coord = dst_ulx;
    destination.y_coord = dst_uly;
    
    /* get cursor information */
    gpr_$inq_cursor(&dm_curs_pat, dm_curs_raster_op, &dm_curs_active, 
                    &dm_curs_position, &dm_curs_origin, &dm_curs_status);

    /* set DM cursor inactive during blit */
    gpr_$set_cursor_active(false, &status);

    /* permit acquisition of display when window obscured */
    gpr_$set_obscured_opt(gpr_$ok_if_obs, &status);

    unobsc = gpr_$acquire_display(&status);
    if (status.all != status_$ok){
        gpr_$release_display(&status);
        return;
        }
    if (unobsc)
        gpr_$pixel_blt(*bitmap_handle, copy_rect, destination, &status);
    else {
        gpr_$inq_vis_list((short)NUM_SLOTS, &returned_slots, vis_list, &status);
        if (returned_slots > NUM_SLOTS)
            returned_slots = NUM_SLOTS;
        gpr_$set_clipping_active(true, &status);
        for (j = 0; j < returned_slots; j++) {
            gpr_$set_clip_window(vis_list[j], &status);
            gpr_$pixel_blt(*bitmap_handle, copy_rect, destination, &status);
            }
        gpr_$set_clipping_active(false, &status);
        }

    /* reset activity of DM cursor to value before blt */
    gpr_$set_cursor_active(dm_curs_active, &status);

    gpr_$release_display(&status);
}

/*blit a portion of the bitmap to another position, using the indicated rop*/
/*when all six long arguments are 0, set the rop */
void tml_dp_gpr_blt(in_bitmap_handle, src_ulx, src_uly, src_lrx, src_lry, dst_ulx, dst_uly, rop_id)
    void*    in_bitmap_handle;
    long    src_ulx,
            src_uly,
            src_lrx,
            src_lry,
            dst_ulx,
            dst_uly,
            rop_id;
{
    status_$t               status;
    gpr_$bitmap_desc_t      current_bitmap;
    gpr_$window_t           copy_rect;
    gpr_$position_t         destination;
    gpr_$raster_op_array_t  current_rops;
    gpr_$mask_32_t          canned_mask = -1;
    short                   i;
    boolean                 unobsc;
    struct bm_info         *bitmap_info;
    short                   returned_slots;
    gpr_$window_t           vis_list[NUM_SLOTS];
    register long           j;
    gpr_$bitmap_desc_t      dm_curs_pat;
    gpr_$raster_op_array_t  dm_curs_raster_op;
    boolean                 dm_curs_active;
    gpr_$position_t         dm_curs_position;
    gpr_$position_t         dm_curs_origin;
    status_$t               dm_curs_status;

    gpr_$bitmap_desc_t    *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        current_bitmap = *((gpr_$bitmap_desc_t*) gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL));
        }
              
    /*make sure that we've got the right raster-op*/
    if (status.all != status_$ok) return;
    if (!current_bitmap) {                  /* no bitmap_info for bitmap_handle */
        gpr_$inq_raster_ops(current_rops, &status);
        if (status.all != status_$ok) return;
        }
    else
        find_bitmap_info(bitmap_handle, &bitmap_info);

    /* set raster operations to argument value for all planes */
    gpr_$set_raster_op_mask(canned_mask, (gpr_$raster_op_t) rop_id, &status);

    if (src_ulx | src_uly | src_lrx | src_lry | dst_ulx | dst_uly) {
        copy_rect.window_base.x_coord = src_ulx;
        copy_rect.window_base.y_coord = src_uly;
        copy_rect.window_size.x_size = src_lrx - src_ulx;
        copy_rect.window_size.y_size = src_lry - src_uly;
        destination.x_coord = dst_ulx;
        destination.y_coord = dst_uly;
        
        /* get cursor information */
        gpr_$inq_cursor(&dm_curs_pat, dm_curs_raster_op, &dm_curs_active, 
                        &dm_curs_position, &dm_curs_origin, &dm_curs_status);
        
        /* set DM cursor inactive during blit */
        gpr_$set_cursor_active(false, &status);
        
        /* permit acquisition of display when window obscured */
        gpr_$set_obscured_opt(gpr_$ok_if_obs, &status);
        
        unobsc = gpr_$acquire_display(&status);
        if (status.all != status_$ok){
            gpr_$release_display(&status);
            return;
            }
        if (unobsc)
            gpr_$pixel_blt(*bitmap_handle, copy_rect, destination, &status);
        else {
            gpr_$inq_vis_list((short) NUM_SLOTS, &returned_slots, vis_list, &status);
            if (returned_slots > NUM_SLOTS)
                returned_slots = NUM_SLOTS;
            gpr_$set_clipping_active(true, &status);
            for (j = 0; j < returned_slots; j++) {
                gpr_$set_clip_window(vis_list[j], &status);
                gpr_$pixel_blt(*bitmap_handle, copy_rect, destination, &status);
                }
            gpr_$set_clipping_active(false, &status);
            }
        
        /* reset activity of DM cursor to value before blt */
        gpr_$set_cursor_active(dm_curs_active, &status);
        
        gpr_$release_display(&status);
        }
}

/* draw a sting into the bitmap at the indicated positon in the indicated font */
/* return the length of the string */
/* use a transparent background to prevent interference with overlapping text, */
/*   as in e.g. italics or Arabic script. */
long tml_dp_gpr_string_draw(in_bitmap_handle, string, st_count, wherex, wherey, in_use_font)
    void* in_bitmap_handle;
    char* string;
    long  st_count,
          wherex,
          wherey;
    void* in_use_font;
{
    status_$t               status;
    gpr_$bitmap_desc_t      current_bitmap;
    short                   current_font;
    struct fhandle*         current_fhandle;
    gpr_$direction_t        current_dir;
    gpr_$coordinate_t       newx, newy;
    gpr_$pixel_value_t      background;
    boolean                 unobsc;
    struct bm_info         *bitmap_info;
    short                   returned_slots;
    gpr_$window_t           vis_list[NUM_SLOTS];
    register long           i;
    gpr_$bitmap_desc_t      dm_curs_pat;
    gpr_$raster_op_array_t  dm_curs_raster_op;
    boolean                 dm_curs_active;
    gpr_$position_t         dm_curs_position;
    gpr_$position_t         dm_curs_origin;
    status_$t               dm_curs_status;
    tml_c_view_pos          top_view_pos;
    tml_c_format            in_format;
    tml_c_phead             on_phead;
    tml_c_cline             on_cline;
    C_MODEL_POS*            start_pos;
    C_SCREEN_XY*            start_xy;
    tml_c_node              at_node;
    unsigned char*          at_char;
    tml_c_model             in_model;
    csm_c_char_set          in_char_set;
    unsigned long           text_dir;
    gpr_$bitmap_desc_t *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;
    struct fhandle     *use_font = (struct fhandle*) in_use_font;

    /*first, make sure that the font is ok, and that there is a string to draw*/
    if (!use_font || !st_count || !string)
        return -1;

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    /* derive text painting direction */
    find_bitmap_info(bitmap_handle, &bitmap_info);
    start_pos = c_model_pos_alloc();
    start_xy = c_screen_xy_alloc();
    start_xy->at_x = wherex;
    start_xy->at_y = wherey;
    top_view_pos = view_port_get_top_line(bitmap_info->viewport_id);
    if (top_view_pos) {
        if (on_cline = view_pos_get_cline(top_view_pos))
            view_port_xy_2_mpos(bitmap_info->viewport_id, start_xy, start_pos);
        }
    if ((0 == top_view_pos) || (0 == on_cline) || (0 == start_pos->in_node)) {
        in_format = view_port_get_format(bitmap_info->viewport_id);
        on_phead = format_get_first_page(in_format);
        phead_get_start(on_phead, start_pos);
        }
    at_node = start_pos->in_node;
    at_char = (unsigned char*) node_get_contents(at_node);
    at_char += start_pos->char_index;
    in_model = node_get_model(at_node);
    in_char_set = (csm_c_char_set) model_inq_state_at_node(in_model, at_node, tml_char_set_prop);
    text_dir = char_set_get_dir_to_char(in_char_set, at_char);
    switch (text_dir) {
        case 0:
            gpr_$set_text_path(gpr_$left, &status);
            break;
        case 1:
            gpr_$set_text_path(gpr_$right, &status);
            break;
        case 2:
            gpr_$set_text_path(gpr_$up, &status);
            break;
        case 3:
            gpr_$set_text_path(gpr_$down, &status);
            break;
        default:
            gpr_$set_text_path(gpr_$right, &status);
            break;
        }
    c_model_pos_free(start_pos);
    c_screen_xy_free(start_xy);

    if (current_bitmap != *bitmap_handle){
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        current_bitmap = *((gpr_$bitmap_desc_t*) gpr_serve_bm_info(set_current_bitmap_handle, bitmap_handle, NULL, NULL));
        }
              
    /*if the font isn't loaded, load it*/
    if (!use_font->font_num) {
        gpr_$load_font_file(use_font->font_name, use_font->name_len, &use_font->font_num, &status);
        if (status.all != status_$ok) return -1;
        gpr_$set_text_font(use_font->font_num, &status);
        if (status.all != status_$ok) return -1;
        }
    else {      /*check to make sure that the font requested is the current font*/
        if ( ! current_bitmap)
            gpr_$inq_text(&current_font, &current_dir, &status);
        else {
            current_fhandle = (struct fhandle *) bitmap_info->font_id;
            if (current_fhandle) {
                current_font = current_fhandle->font_num;
                current_dir = bitmap_info->current_text_direction;
                }
            else
                gpr_$inq_text(&current_font, &current_dir, &status);
            }
        if ((status.all != status_$ok) || (current_font != use_font->font_num))
            gpr_$set_text_font(use_font->font_num, &status);
        if (status.all != status_$ok) return -1;
        }

    gpr_$move((gpr_$coordinate_t) wherex, (gpr_$coordinate_t) wherey, &status);
    if (status.all != status_$ok) return -1;

    /* get cursor information */
    gpr_$inq_cursor(&dm_curs_pat, dm_curs_raster_op, &dm_curs_active, 
                    &dm_curs_position, &dm_curs_origin, &dm_curs_status);

    /* set DM cursor inactive during blit */
    gpr_$set_cursor_active(false, &status);

    /* permit acquisition of display when window obscured */
    gpr_$set_obscured_opt(gpr_$ok_if_obs, &status);

    /* remember text background value */
    background = gpr_$inq_background(&status);

    /* make text background color transparent */
    gpr_$set_text_background_value(gpr_$transparent, &status);

    unobsc = gpr_$acquire_display(&status);
    if (status.all != status_$ok){
        gpr_$release_display(&status);
        return -1;
        }
    if (unobsc) {
        if (1 == use_font->font_table->bytes_per_char)
            gpr_$text(string, (short) st_count, &status);
        else
            gpr_$text16((unsigned short*) string, (short) st_count, &status);
        }
    else {
        gpr_$inq_vis_list((short) NUM_SLOTS, &returned_slots, vis_list, &status);
        if (returned_slots > NUM_SLOTS)
            returned_slots = NUM_SLOTS;
        gpr_$set_clipping_active(true, &status);
        for (i = 0; i < returned_slots; i++) {
            gpr_$set_clip_window(vis_list[i], &status);
            gpr_$move((gpr_$coordinate_t) wherex, (gpr_$coordinate_t) wherey, &status);
            if (1 == use_font->font_table->bytes_per_char)
                gpr_$text(string, (short) st_count, &status);
            else
                gpr_$text16((unsigned short*) string, (short) st_count, &status);
            }
        gpr_$set_clipping_active(false, &status);
        }

    /* reset text background color to previous value */
    gpr_$set_text_background_value(background, &status);

    /* reset activity of DM cursor to value before blt */
    gpr_$set_cursor_active(dm_curs_active, &status);

    gpr_$release_display(&status);
    if (status.all != status_$ok)
        return -1;

    gpr_$inq_cp(&newx, &newy, &status);
    if (status.all != status_$ok) return -1;

    return (newx - (gpr_$coordinate_t) wherex);
}

/*return the size of a string in the indicated font*/
long tml_dp_gpr_string_size(string, st_count, in_use_font)
    char* string;
    long  st_count;
    void* in_use_font;
{
    register long                   i,
                                    width;
    register ushort                 icspace;
    register ushort                 char_index;
    register fontn_$table_ptr_tver1 font_table;
             short*                 index_table;
             fontn_$desc*           desc_table;
             status_$t              status;

    struct fhandle *use_font = (struct fhandle*) in_use_font;

    font_table = use_font->font_table;
    /*take care of the easy case if we can*/
    if (use_font->is_fixed)
        return ((use_font->is_fixed + font_table->inter_space_dflt) * st_count);

    /*it's proportional--do it the hard way*/
    icspace = font_table->inter_space_dflt;
    index_table = (short*) (font_table + font_table->index_table_offset);
    desc_table  = (fontn_$desc*) (font_table + font_table->desc_table_offset);
    for (i = 0, width = 0; i < st_count; i++){
        char_index = (ushort) (index_table[*(string + i)]);
        if (char_index)
            width += desc_table[char_index].width + icspace;
        else
            width += font_table->space_size + icspace;
        }
    return width;
}

/*repair any damage in the specified viewport*/
void tml_dp_gpr_damage_repair(in_bitmap_handle)
    void* in_bitmap_handle;
{
    status_$t           status;
    gpr_$bitmap_desc_t  current_bitmap;
    struct bm_info     *bitmap_info;
    tml_c_view_port     refresh_viewport;

    gpr_$bitmap_desc_t *bitmap_handle = (gpr_$bitmap_desc_t*) in_bitmap_handle;

    extern void tml_dp_init_draw(tml_c_view_port refresh_viewport);

    current_bitmap = tml_dp_gpr_get_gpr_bitmap();

    if (current_bitmap != *bitmap_handle) {
        gpr_$set_bitmap(*bitmap_handle, &status);
        if (status.all != status_$ok) return;
        }

    /* permit acquisition of display when window obscured */
    gpr_$set_obscured_opt(gpr_$ok_if_obs, &status);

    if (NULL == gpr_serve_bm_info(find_bitmap_handle_info, bitmap_handle, &bitmap_info, NULL))
        return;

    /* get viewport to refresh */
    refresh_viewport = bitmap_info->viewport_id;

    /* perform initial draw if required */
    if (0 == bitmap_info->drawn 
      && (NULL == view_port_xy_2_cline(refresh_viewport, &zero_zero))) {
        bitmap_info->drawn = -1;
        tml_dp_init_draw(refresh_viewport);
        return;
        }

    gpr_$acquire_display(&status);
    gpr_$release_display(&status);

}

/*takes a bitmap_id (obtained from viewport->get_handle()) and returns a void*
    that points to the graphics-system-dependent information contained within it,
    in this case a gpr_$bitmap_desc_t */
/*a noop for gpr*/
void* tml_dp_gpr_get_grafsys_obj(bitmap_handle)
    void* bitmap_handle;
{
    return bitmap_handle;
}

/*default cursor drawing routine--invert a triangle of specified height and width, */
/* with the back edge (lr) on the indicated x,y coordinate in the viewport*/
long tml_dp_gpr_dflt_cursor_draw(in_draw_to_bitmap, wherex, wherey, height, width, erase)
    void* in_draw_to_bitmap;
    long  wherex, wherey,
          height, width,
          erase;
{
    gpr_$bitmap_desc_t*    draw_to_bitmap = (gpr_$bitmap_desc_t*) in_draw_to_bitmap;

    long                   i;
    gpr_$bitmap_desc_t     current_bitmap;
    gpr_$rgb_plane_t       bm_planes;
    gpr_$pixel_value_t     fg_index;
    gpr_$pixel_value_t     bg_index;
    gpr_$pixel_value_t     flip_index;
    gpr_$pixel_value_t     fill_index;
    status_$t              status;
    gpr_$raster_op_array_t current_rops;
    unsigned short int     prim_set;
    unsigned short int     prim_bit;
    boolean                unobsc;
    short                  returned_slots;
    gpr_$window_list_t     vis_list;
    gpr_$bitmap_desc_t     dm_curs_pat;
    gpr_$raster_op_array_t dm_curs_raster_op;
    boolean                dm_curs_active;
    gpr_$position_t        dm_curs_position;
    gpr_$position_t        dm_curs_origin;
    status_$t              dm_curs_status;

    /* For this function to work properly the bitmap associated */
    /*   with the draw_to_bitmap argument must have as many planes */
    /*   as the display allows */

    /* remember the current bitmap */
    gpr_$inq_bitmap(&current_bitmap, &status);

    /* make current the bitmap on which the cursor is to be drawn */
    gpr_$set_bitmap(*draw_to_bitmap, &status);
    if (status.all != status_$ok) {
        /* restore original bitmap */
        gpr_$set_bitmap(current_bitmap, &status);
        return 0;
        }

    /* retrieve number of planes */
    gpr_$inq_bitmap_dimensions(*draw_to_bitmap, &tml_trash_size, &bm_planes, &status);

    /* obtain text foreground and background colors */
    gpr_$inq_text_values(&fg_index, &bg_index, &status);

    /* replace default value for background with actual pixel value */
    /* gpr_$background (-2) is background value in direct mode when user */
    /*   desires background color to be the background color of the DM */
    /*   window used for direct mode */
    /* a value of 1 is the default foreground when gpr_$background is set */
    /* the gpr calls are safe only in direct and frame mode, */
    /*   but gpr_$background is a possible value only in those modes */
    /* ignore the possibility of gpr_$transparent background; */
    /*   that value doesn't work in TMLib anyway */
    if (gpr_$background == bg_index) {
        bg_index = gpr_$inq_background(&status);
        if (1 == fg_index)
            fg_index = gpr_$inq_foreground(&status);
        }

    /* get phony color index that flips foreground and background color */
    /*   indices when ropped over them with XOR */
    flip_index = fg_index ^ bg_index;

    /* initialize cursor parameter globals if necessary */
    /* or if present and requested cursor size or position parameters differ, */
    /*   recalculate triangle points and reset to requested values */
    if (((0 == tml_cursor_size.x_size) && (0 == tml_cursor_size.y_size))
      || ((wherex != tml_cursor_points.tv1.x_coord)
      ||  (wherey != tml_cursor_points.tv1.y_coord)
      ||  (width  != tml_cursor_size.x_size)
      ||  (height != tml_cursor_size.y_size))) {
        tml_cursor_size.x_size = (gpr_$coordinate_t) width;
        tml_cursor_size.y_size = (gpr_$coordinate_t) height;
        tml_cursor_points.tv1.x_coord = (gpr_$coordinate_t) wherex;
        tml_cursor_points.tv1.y_coord = (gpr_$coordinate_t) wherey;
        tml_cursor_points.tv2.x_coord = (gpr_$coordinate_t) (wherex - (width / 2));
        tml_cursor_points.tv2.y_coord = (gpr_$coordinate_t) (wherey + height);
        tml_cursor_points.tv3.x_coord = (gpr_$coordinate_t) (wherex + (width / 2));
        tml_cursor_points.tv3.y_coord = (gpr_$coordinate_t) (wherey + height);
        }

    /* permit acquisition of display when window obscured */
    gpr_$set_obscured_opt(gpr_$ok_if_obs, &status);

    /* acquire display prior to writing cursor */
    unobsc = gpr_$acquire_display(&status);

    /* remember the rops of the destination bitmap */
    gpr_$inq_raster_ops(current_rops, &status);

    /* remember the fill value of the destination bitmap */
    gpr_$inq_fill_value(&fill_index, &status);

    /* NOTE: RESET OF RASTER OP PRIMITIVE SET */
    /* MUST PRECEDE RESET OF FILL VALUE */
    /* OR FILL VALUE WILL NOT BE RESET */

    /* remember the raster op primitive set of the destination bitmap */
    /*   (i.e. the set of gpr operations types that are affected */
    /*   by raster ops */
    gpr_$inq_raster_op_prim_set(&prim_set, &status);

    /* add gpr_$rop_fill to the raster op prim set */
    /*   (the raster op prim set is a Pascal set; the C enum for the */
    /*   gpr operations types is zero-based, not one-based like */
    /*   Pascal version of the enumeration) */
    for (i = 0, prim_bit = 1; i < gpr_$rop_fill; i++)
        prim_bit <<= 1;
    gpr_$raster_op_prim_set(prim_set | prim_bit, &status);

    /* set the fill index to the flipping value */
    gpr_$set_fill_value(flip_index, &status);

    /* set the rops to SOURCE XOR DESTINATION */
    for (i = 0; i <= bm_planes; i++)
        gpr_$set_raster_op((gpr_$rgb_plane_t) i, gpr_$rop_src_xor_dst, &status);

    /* get cursor information */
    gpr_$inq_cursor(&dm_curs_pat, dm_curs_raster_op, &dm_curs_active, 
                    &dm_curs_position, &dm_curs_origin, &dm_curs_status);

    /* set DM cursor inactive during blit */
    gpr_$set_cursor_active(false, &status);

    /* flip foreground and background values on the bitmap within the cursor triangle */
    if (unobsc)
        gpr_$triangle(tml_cursor_points.tv1, tml_cursor_points.tv2, tml_cursor_points.tv3, &status);
    else {
        gpr_$inq_vis_list((short)NUM_SLOTS, &returned_slots, vis_list, &status);
        if (returned_slots > NUM_SLOTS)
            returned_slots = NUM_SLOTS;
        gpr_$set_clipping_active(true, &status);
        for (i = 0; i < returned_slots; i++) {
            gpr_$set_clip_window(vis_list[i], &status);
            gpr_$triangle(tml_cursor_points.tv1, tml_cursor_points.tv2, tml_cursor_points.tv3, &status);
            }
        gpr_$set_clipping_active(false, &status);
        }

    /* reset activity of DM cursor to value before blt */
    gpr_$set_cursor_active(dm_curs_active, &status);

    /* restore the fill value of the destination bitmap */
    gpr_$set_fill_value(fill_index, &status);

    /* restore the raster_op_prim_set of the destination bitmap */
    gpr_$raster_op_prim_set(prim_set, &status);

    /* restore the rops of the destination bitmap */
    for (i = 0; i <= bm_planes; i++)
        gpr_$set_raster_op((gpr_$rgb_plane_t) i, current_rops[i], &status);

    /* restore current bitmap */
    if (current_bitmap != *draw_to_bitmap)
        gpr_$set_bitmap(current_bitmap, &status);

    /* release display after drawing cursor on bitmap */
    gpr_$release_display(&status);

    if (erase)
        return 0;
    else
        return 1;
}

