/**************************************************************************************
    tml_functions.hxx
    header file for the function definitions and assorted type declarations
      for the portable version of the text management library.

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#ifndef tml_functions_hxx
#define tml_functions_hxx

#ifndef tml_objects_hxx
#include <tml/tml_objects.hxx>
#endif

#include <tml/tml_cxx2.0_linkage_decls.hxx>

/**************************************************************************
    Declarations for print functions
**************************************************************************/

enum orientation {portrait, landscape};

/**************************************************************************
    Modelling component functions using the traits

***************************************************************************/

extern node* node_alloc(long node_size, model* in_model = 0);

extern prop_vec* pvec_alloc(model* to_model = 0);
 
extern property* prop_alloc(model* to_model = 0);
 
extern argument* arg_alloc(model* to_model = 0);
                                                                                                                       
extern node* make_subtree(node *start_node, node *end_node);

extern node* copy_component(model_pos *start, model_pos *end, model *to_model);
                  
extern property* make_wrap_prop(long wrap_on, model *in_model);

extern property* make_margin_prop(long left,
                                  long right,
                                  long top,
                                  long bottom,
                                  model *in_model);

extern property* make_inv_prop(long inv_on, model *in_model);
  
extern property* make_font_prop(char *font_name, long name_len, model *in_model, grafsys for_grafsys = 0);

extern property* make_page_size_prop(long width, long depth, model *in_model);

extern property* make_header_prop(model *in_model,
                                  node  *left_hdr,
                                  node  *center_hdr,
                                  node  *right_hdr);

extern property* make_footer_prop(model *in_model,
                                  node  *left_ftr,
                                  node  *center_ftr,
                                  node  *right_ftr);

extern property* make_form_feed_prop(long form_feed_on, model *in_model);

extern property* make_char_set_prop(char_set* cur_char_set, model* in_model);

extern property* make_union_char_set_prop(union_char_set* cur_union, model* in_model);

extern property* make_tab_prop(model *in_mod, long default_val, long num_stops, long *tab_st_list);

extern void connect_prop(property *add_prop, node *to_node);
                  
overload d_read_model;

extern model* d_read_model(void* file_handle);

extern model* d_read_model(char* pathname, long name_len);

extern model* d_read_model(char* pathname, long name_len, long oflag, long mode, void* *file_handle_p);

overload read_uasc;

extern model* read_uasc(void *file_handle);

extern model* read_uasc(char* pathname, long name_len);

extern void gc_linearize(model* old_model);

extern model* prune(model* old_model);

extern long tml_model_print(model* this_model, print_db* print_params, char* outfilename, long outlen);

extern long tml_file_print(char* infilename, long inlen, print_db* print_params, char* outfilename, long outlen);
                     
extern void* tml_make_fhandle(void *file_id, long for_file_type);

extern void tml_free_fhandle(void *fhandle);


/*functions for the two supported search environments--tml_aegis_ functions support
    aegis regular expression pattern matching, while tml_unix_ functions support
    unix regular expression pattern matching. The tml_unix_ functions are loaded as
    the default search environment*/

/*declarations of the aegis regular expression matching stuff*/
extern char *tml_aegis_compile_re(char *compile_st,long compile_len,void *out_pat,long out_pat_max,long case_sens=1);
extern long tml_aegis_match_re(char *search_st, long search_len, void *use_pat);
extern long tml_aegis_min_match(void *pattern);

/*declarations of the unix regular expression matching stuff*/
extern char *tml_unix_compile_re(char *compile_st,long compile_len,void *out_pat,long out_pat_max,long case_sens=1);
extern long tml_unix_match_re(char *search_st, long search_len, void *use_pat);
extern long tml_unix_min_match(void *pattern);

/* get default union_char_set */
extern union_char_set* get_dflt_union_char_set();

/*character-set-dependent string functions using default character set for model*/
extern char* strcat_for_model (model* in_model, char* s1, char* s2);
extern char* strncat_for_model(model* in_model, char* s1, char* s2, long n);
extern long  strcmp_for_model (model* in_model, char* s1, char* s2);
extern long  strncmp_for_model(model* in_model, char* s1, char* s2, long n);
extern char* strcpy_for_model (model* in_model, char* s1, char* s2);
extern char* strncpy_for_model(model* in_model, char* s1, char* s2, long n);
extern long  strlen_for_model (model* in_model, char*);
extern char* index_for_model  (model* in_model, char* s, char* c);
extern char* rindex_for_model (model* in_model, char* s, char* c);

/*character-set-dependent test for null*/
extern long  is_null_for_model(model* in_model, char* c);

/*character-set-dependent COLLATING CHARACTER string comparisons*/
extern long char_arr_cmp_for_model(model* in_model, char* s1, char* s2);
extern long char_arr_ncmp_for_model(model* in_model, char* s1, char* s2, unsigned long n);

/*declarations of character-set-dependent string functions using character set in effect at a node*/
extern char* strcat_for_node (node* at_node, char* s1, char* s2);
extern char* strncat_for_node(node* at_node, char* s1, char* s2, long n);
extern long  strcmp_for_node (node* at_node, char* s1, char* s2);
extern long  strncmp_for_node(node* at_node, char* s1, char* s2, long n);
extern char* strcpy_for_node (node* at_node, char* s1, char* s2);
extern char* strncpy_for_node(node* at_node, char* s1, char* s2, long n);
extern long  strlen_for_node (node* at_node, char*);
extern char* index_for_node  (node* at_node, char* s, char* c);
extern char* rindex_for_node (node* at_node, char* s, char* c);

/*character-set-dependent test for null*/
extern long  is_null_for_node(node* at_node, char* c);

/*character-set-dependent COLLATING CHARACTER string comparisons*/
extern long char_arr_cmp_for_node(node* at_node, char* s1, char* s2);
extern long char_arr_ncmp_for_node(node* at_node, char* s1, char* s2, unsigned long n);

/*character-set-independent COLLATING CHARACTER string comparisons*/
extern long char_arr_cmp(csm_character* c1, csm_character* c2);
extern long char_arr_ncmp(csm_character* c1, csm_character* c2, unsigned long n);

/* create union_char_sets from binary file indicated by file handle, using rfp to read the file */
extern union_char_set* read_union_char_sets(void *file_handle, gc_hash_tbl*& hash_tbl, tml_ext_type*& types, endian*& their_endian);

/* create union_char_sets from human-readable file indicated by file handle, using rfp to read the file */
extern union_char_set* read_config_file_union_char_sets(void *file_handle);

/* set or reset a character set pointer to be used for csm_cmp_for_qsort() */
extern void set_csm_char_set_for_qsort(char_set* new_char_set);

/* string comparison function to be used with UNIX qsort() */
/*   when character sets are involved */
extern int csm_cmp_for_qsort(char** sp1, char** sp2);

/* update history when a call to prop_vec::detach detaches the only */
/*   remaining prop_vec in a vector. */
/* Call this function when a call to prop_vec::detach() returns 0. */
/* Argument "on_node" is the node whose properties member contained */
/*   only the one prop_vec "removed". */
extern void first_pvec_hist(node* on_node, prop_vec* removed);

/* update history when a call to argument::detach detaches the only */
/*   remaining argument in a property object's entry argument  vector. */
/* Call this function when a call to argument::detach() on the entry */
/*   argument vector of a property returns 0. */
/* Argument "on_prop" is the property whose entry_arg member contained */
/*   only the one argument "removed". */
extern void first_entry_arg_hist(property* on_prop, argument* removed);

/* update history when a call to argument::detach detaches the only */
/*   remaining argument in a property object's exit argument vector. */
/* Call this function when a call to argument::detach() on the exit */
/*   argument vector of a property returns 0. */
/* Argument "on_prop" is the property whose exit_arg member contained */
/*   only the one argument "removed". */
extern void first_exit_arg_hist(property* on_prop, argument* removed);

/* return the viewport function table for XWindows */
extern vpftbl* tml_dp_get_x_vpftbl();

#ifdef apollo
/* return the viewport function table for Apollo GPR */
extern vpftbl* tml_dp_get_gpr_vpftbl();
#endif

/* return the type identifier for the default graphic system */
extern grafsys  tml_dp_get_dflt_grafsys_type();

/* get the default graphics system */
extern vpftbl*  tml_dp_get_dflt_vpftbl();

/* set the default graphics system */
extern void     tml_dp_set_dflt_vpftbl(vpftbl* new_table);

#endif 
// tml_functions_hxx

