/**************************************************************************************
    File tml_magic.cxx
    Utility functions for TML magic numbers

    (c) Copyright Apollo Computer Inc., 1988 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/endian.hxx>


/******************************************************************************/
/*                                                                            */
/*  base class endian member functions                                        */
/*                                                                            */
/******************************************************************************/


HALF_PTR       endian::my_half_ptr(HALF_PTR their_half_ptr)
{
    if (sizeof(HALF_PTR) == sizeof(unsigned short))
        return((HALF_PTR) my_short((unsigned short) their_half_ptr));
    else if (sizeof(HALF_PTR) == sizeof(unsigned long))
        return((HALF_PTR) my_long((unsigned long) their_half_ptr));
    else
        return(their_half_ptr);
}


unsigned long* endian::long_arr_shuf (unsigned long* inbuf, long n)
{
    long i;

    for (i = 0; i < n; i++)
        *(inbuf + i) = my_long(*(inbuf + i));
    return(inbuf);
}


unsigned short* endian::short_arr_shuf(unsigned short* inbuf, long n)
{
    long i;

    for (i = 0; i < n; i++)
        *(inbuf + i) = my_short(*(inbuf + i));
    return(inbuf);
}

/******************************************************************************/
/*                                                                            */
/*  member functions for derived endian classes used in restoring files       */
/*     written by heterogeneous architectures                                 */
/*                                                                            */
/******************************************************************************/


endian_0123::endian_0123()
{
    s_mask_arr     = (unsigned long*) (new unsigned long [2]);
    l_mask_arr     = (unsigned long*) (new unsigned long [4]);
    s_shift_arr    = (unsigned long*) (new unsigned long [2]);
    l_shift_arr    = (unsigned long*) (new unsigned long [4]);

    s_mask_arr[0]  = (unsigned long) 0x0000FF00;
    s_mask_arr[1]  = (unsigned long) 0x000000FF;

    s_shift_arr[0] = 8;
    s_shift_arr[1] = 0;

    l_mask_arr[0]  = (unsigned long) 0xFF000000;
    l_mask_arr[1]  = (unsigned long) 0x00FF0000;
    l_mask_arr[2]  = (unsigned long) 0x0000FF00;
    l_mask_arr[3]  = (unsigned long) 0x000000FF;

    l_shift_arr[0] = 24;
    l_shift_arr[1] = 16;
    l_shift_arr[2] =  8;
    l_shift_arr[3] =  0;
}

endian_1032::endian_1032()
{
    s_mask_arr     = (unsigned long*) (new unsigned long [2]);
    l_mask_arr     = (unsigned long*) (new unsigned long [4]);
    s_shift_arr    = (unsigned long*) (new unsigned long [2]);
    l_shift_arr    = (unsigned long*) (new unsigned long [4]);

    s_mask_arr[0]  = (unsigned long) 0x000000FF;
    s_mask_arr[1]  = (unsigned long) 0x0000FF00;

    s_shift_arr[0] = 0;
    s_shift_arr[1] = 8;

    l_mask_arr[0]  = (unsigned long) 0x00FF0000;
    l_mask_arr[1]  = (unsigned long) 0xFF000000;
    l_mask_arr[2]  = (unsigned long) 0x000000FF;
    l_mask_arr[3]  = (unsigned long) 0x0000FF00;

    l_shift_arr[0] = 16;
    l_shift_arr[1] = 24;
    l_shift_arr[2] =  0;
    l_shift_arr[3] =  8;
}

endian_2301::endian_2301()
{
    s_mask_arr     = (unsigned long*) (new unsigned long [2]);
    l_mask_arr     = (unsigned long*) (new unsigned long [4]);
    s_shift_arr    = (unsigned long*) (new unsigned long [2]);
    l_shift_arr    = (unsigned long*) (new unsigned long [4]);

    s_mask_arr[0]  = (unsigned long) 0x0000FF00;
    s_mask_arr[1]  = (unsigned long) 0x000000FF;

    s_shift_arr[0] = 8;
    s_shift_arr[1] = 0;

    l_mask_arr[0]  = (unsigned long) 0x0000FF00;
    l_mask_arr[1]  = (unsigned long) 0x000000FF;
    l_mask_arr[2]  = (unsigned long) 0xFF000000;
    l_mask_arr[3]  = (unsigned long) 0x00FF0000;

    l_shift_arr[0] =  8;
    l_shift_arr[1] =  0;
    l_shift_arr[2] = 24;
    l_shift_arr[3] = 16;
}

endian_3210::endian_3210()
{
    s_mask_arr     = (unsigned long*) (new unsigned long [2]);
    l_mask_arr     = (unsigned long*) (new unsigned long [4]);
    s_shift_arr    = (unsigned long*) (new unsigned long [2]);
    l_shift_arr    = (unsigned long*) (new unsigned long [4]);

    s_mask_arr[0]  = (unsigned long) 0x000000FF;
    s_mask_arr[1]  = (unsigned long) 0x0000FF00;

    s_shift_arr[0] = 0;
    s_shift_arr[1] = 8;

    l_mask_arr[0]  = (unsigned long) 0x000000FF;
    l_mask_arr[1]  = (unsigned long) 0x0000FF00;
    l_mask_arr[2]  = (unsigned long) 0x00FF0000;
    l_mask_arr[3]  = (unsigned long) 0xFF000000;

    l_shift_arr[0] =  0;
    l_shift_arr[1] =  8;
    l_shift_arr[2] = 16;
    l_shift_arr[3] = 24;
}

unsigned long   endian_3210::my_long(unsigned long their_u_long )
{
    return(their_u_long);
}

unsigned short  endian_3210::my_short(unsigned short their_u_short)
{
    return(their_u_short);
}

HALF_PTR        endian_3210::my_half_ptr(HALF_PTR their_half_ptr)
{
    return(their_half_ptr);
}

unsigned long*  endian_3210::long_arr_shuf(unsigned long* inbuf, long n)
{
    return(inbuf);
}

unsigned short* endian_3210::short_arr_shuf(unsigned short* inbuf, long n)
{
    return(inbuf);
}

